// Copyright 2022 xensik. All rights reserved.
//
// Use of this source code is governed by a GNU GPLv3 license
// that can be found in the LICENSE file.

#include "stdafx.hpp"
#include "iw5.hpp"

#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4244)
#endif

namespace xsk::gsc::iw5
{

std::unordered_map<std::uint8_t, std::string_view> opcode_map;
std::unordered_map<std::uint16_t, std::string_view> function_map;
std::unordered_map<std::uint16_t, std::string_view> method_map;
std::unordered_map<std::uint16_t, std::string_view> file_map;
std::unordered_map<std::uint16_t, std::string_view> token_map;
std::unordered_map<std::string_view, std::uint8_t> opcode_map_rev;
std::unordered_map<std::string_view, std::uint16_t> function_map_rev;
std::unordered_map<std::string_view, std::uint16_t> method_map_rev;
std::unordered_map<std::string_view, std::uint16_t> file_map_rev;
std::unordered_map<std::string, std::uint16_t> token_map_rev;
std::unordered_map<std::string, std::vector<std::uint8_t>> files;
read_cb_type read_callback = nullptr;
std::set<std::string> string_map;

void resolver::init(read_cb_type callback)
{
    read_callback = callback;
}

void resolver::cleanup()
{
    files.clear();
}

auto resolver::opcode_id(const std::string& name) -> std::uint8_t
{
    const auto itr = opcode_map_rev.find(name);

    if (itr != opcode_map_rev.end())
    {
        return itr->second;
    }

    throw error(utils::string::va("Couldn't resolve opcode id for name '%s'!", name.data()));
}

auto resolver::opcode_name(std::uint8_t id) -> std::string
{
    const auto itr = opcode_map.find(id);

    if (itr != opcode_map.end())
    {
        return std::string(itr->second);
    }

    throw error(utils::string::va("Couldn't resolve opcode name for id '0x%hhX'!", id));
}

auto resolver::function_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_func_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(6), nullptr, 16));
    }

    const auto itr = function_map_rev.find(name);

    if (itr != function_map_rev.end())
    {
        return itr->second;
    }

    throw error(utils::string::va("Couldn't resolve builtin function id for name '%s'!", name.data()));
}

auto resolver::function_name(std::uint16_t id) -> std::string
{
    const auto itr = function_map.find(id);

    if (itr != function_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_func_%04X", id);
}

auto resolver::method_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_meth_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(6), nullptr, 16));
    }

    const auto itr = method_map_rev.find(name);

    if (itr != method_map_rev.end())
    {
        return itr->second;
    }

    throw error(utils::string::va("Couldn't resolve builtin method id for name '%s'!", name.data()));
}

auto resolver::method_name(std::uint16_t id) -> std::string
{
    const auto itr = method_map.find(id);

    if (itr != method_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_meth_%04X", id);
}

auto resolver::file_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_id_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(4), nullptr, 16));
    }

    const auto itr = file_map_rev.find(name);

    if (itr != file_map_rev.end())
    {
        return itr->second;
    }

    return 0;
}

auto resolver::file_name(std::uint16_t id) -> std::string
{
    const auto itr = file_map.find(id);

    if (itr != file_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_id_%04X", id);
}

auto resolver::token_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_id_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(4), nullptr, 16));
    }

    const auto itr = token_map_rev.find(name);

    if (itr != token_map_rev.end())
    {
        return itr->second;
    }

    return 0;
}

auto resolver::token_name(std::uint16_t id) -> std::string
{
    const auto itr = token_map.find(id);

    if (itr != token_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_id_%04X", id);
}

auto resolver::find_function(const std::string& name) -> bool
{
    if (name.starts_with("_func_")) return true;

    const auto itr = function_map_rev.find(name);

    if (itr != function_map_rev.end())
    {
        return true;
    }

    return false;
}

auto resolver::find_method(const std::string& name) -> bool
{
    if (name.starts_with("_meth_")) return true;

    const auto itr = method_map_rev.find(name);

    if (itr != method_map_rev.end())
    {
        return true;
    }

    return false;
}

void resolver::add_function(const std::string& name, std::uint16_t id)
{
    const auto itr = function_map_rev.find(name);

    if (itr != function_map_rev.end())
    {
        throw error(utils::string::va("builtin function '%s' already defined.", name.data()));
    }

    const auto str = string_map.find(name);

    if (str != string_map.end())
    {
        function_map.insert({ id, *str });
        function_map_rev.insert({ *str, id });
    }
    else
    {
        auto ins = string_map.insert(name);

        if (ins.second)
        {
            function_map.insert({ id, *ins.first });
            function_map_rev.insert({ *ins.first, id });
        }
    }
}

void resolver::add_method(const std::string& name, std::uint16_t id)
{
    const auto itr = method_map_rev.find(name);

    if (itr != method_map_rev.end())
    {
        throw error(utils::string::va("builtin method '%s' already defined.", name.data()));
    }

    const auto str = string_map.find(name);

    if (str != string_map.end())
    {
        method_map.insert({ id, *str });
        method_map_rev.insert({ *str, id });
    }
    else
    {
        auto ins = string_map.insert(name);

        if (ins.second)
        {
            method_map.insert({ id, *ins.first });
            method_map_rev.insert({ *ins.first, id });
        }
    }
}

auto resolver::make_token(std::string_view str) -> std::string
{
    if (str.starts_with("_id_") || str.starts_with("_func_") || str.starts_with("_meth_"))
    {
        return std::string(str);
    }

    auto data = std::string(str.begin(), str.end());

    for (std::size_t i = 0; i < data.size(); i++)
    {
        data[i] = static_cast<char>(std::tolower(static_cast<unsigned char>(str[i])));
        if (data[i] == '\\') data[i] = '/';
    }

    return data;
}

auto resolver::file_data(const std::string& name) -> std::tuple<const std::string*, char*, size_t>
{
    const auto& itr = files.find(name);

    if (itr != files.end())
    {
        return { &itr->first ,reinterpret_cast<char*>(itr->second.data()), itr->second.size() };
    }

    auto data = read_callback(name);

    const auto& res = files.insert({ name, std::move(data)});

    if (res.second)
    {
        return { &res.first->first, reinterpret_cast<char*>(res.first->second.data()), res.first->second.size() };
    }

    throw error("couldn't open gsc file '" + name + "'");
}

std::set<std::string_view> paths
{
    "aitype"sv,
    "animscripts"sv,
    "character"sv,
    "codescripts"sv,
    "common_scripts"sv,
    "maps"sv,
    "mptype"sv,
    "vehicle_scripts"sv,
    "xmodelalias"sv,
    "animscripts/civilian"sv,
    "animscripts/dog"sv,
    "animscripts/hummer_turret"sv,
    "animscripts/riotshield"sv,
    "animscripts/saw"sv,
    "animscripts/scripted"sv,
    "animscripts/technical"sv,
    "animscripts/traverse"sv,
    "maps/animated_models"sv,
    "maps/cinematic_setups"sv,
    "maps/createart"sv,
    "maps/createfx"sv,
    "maps/gametypes"sv,
    "maps/mp"sv,
    "maps/mp/gametypes"sv,
    "maps/mp/killstreaks"sv,
    "maps/mp/perks"sv,
};

auto resolver::fs_to_game_path(const std::filesystem::path& file) -> std::filesystem::path
{
    auto result = std::filesystem::path();
    auto root = false;

    for (auto& entry : file)
    {
        if (!root && paths.contains(entry.string()))
        {
            result = entry;
            root = true;
        }
        else if (paths.contains(result.string()))
        {
            result /= entry;
        }
    }

    return result.empty() ? file : result;
}

const std::array<std::pair<std::uint8_t, const char*>, 153> opcode_list
{{
    { 0x00, "END" },
    { 0x01, "RETN" },
    { 0x02, "GET_BYTE" },
    { 0x03, "GET_NBYTE" },
    { 0x04, "GET_USHORT" },
    { 0x05, "GET_NUSHORT" },
    { 0x06, "GET_INT" },
    { 0x07, "GET_BUILTIN_FUNC" },
    { 0x08, "GET_BUILTIN_METHOD" },
    { 0x09, "GET_FLOAT" },
    { 0x0A, "GET_STRING" },
    { 0x0B, "GET_UNDEFINED" },
    { 0x0C, "GET_ZERO" },
    { 0x0D, "WAITTILLFRAMEEND" },
    { 0x0E, "CREATE_LOCAL_VARIABLE" },
    { 0x0F, "REMOVE_LOCAL_VARIABLES" },
    { 0x10, "EVAL_LOCAL_VARIABLE_CACHED0" },
    { 0x11, "EVAL_LOCAL_VARIABLE_CACHED1" },
    { 0x12, "EVAL_LOCAL_VARIABLE_CACHED2" },
    { 0x13, "EVAL_LOCAL_VARIABLE_CACHED3" },
    { 0x14, "EVAL_LOCAL_VARIABLE_CACHED4" },
    { 0x15, "EVAL_LOCAL_VARIABLE_CACHED5" },
    { 0x16, "EVAL_LOCAL_VARIABLE_CACHED" },
    { 0x17, "EVAL_LOCAL_ARRAY_CACHED" },
    { 0x18, "EVAL_ARRAY" },
    { 0x1A, "EVAL_NEW_LOCAL_ARRAY_REF_CACHED0" },
    { 0x19, "EVAL_LOCAL_ARRAY_REF_CACHED0" },
    { 0x1B, "EVAL_LOCAL_ARRAY_REF_CACHED" },
    { 0x1C, "EVAL_ARRAY_REF" },
    { 0x1D, "CLEAR_ARRAY" },
    { 0x1E, "EMPTY_ARRAY" },
    { 0x1F, "ADD_ARRAY" },
    { 0x20, "PRE_CALL" },
    { 0x21, "CALL_LOCAL_FUNC2" },
    { 0x22, "CALL_LOCAL_FUNC" },
    { 0x23, "CALL_LOCAL_METHOD" },
    { 0x24, "CALL_LOCAL_FUNC_THREAD" },
    { 0x25, "CALL_LOCAL_FUNC_CHILD_THREAD" },
    { 0x26, "CALL_LOCAL_METHOD_THREAD" },
    { 0x27, "CALL_LOCAL_METHOD_CHILD_THREAD" },
    { 0x28, "CALL_FAR_FUNC2" },
    { 0x29, "CALL_FAR_FUNC" },
    { 0x2A, "CALL_FAR_METHOD" },
    { 0x2B, "CALL_FAR_FUNC_THREAD" },
    { 0x2C, "CALL_FAR_FUNC_CHILD_THREAD" },
    { 0x2D, "CALL_FAR_METHOD_THEAD" },
    { 0x2E, "CALL_FAR_METHOD_CHILD_THEAD" },
    { 0x2F, "CALL_FUNC_POINTER" },
    { 0x30, "CALL_METHOD_POINTER" },
    { 0x31, "CALL_FUNC_THREAD_POINTER" },
    { 0x32, "CALL_FUNC_CHILD_THREAD_POINTER" },
    { 0x33, "CALL_METHOD_THREAD_POINTER" },
    { 0x34, "CALL_METHOD_CHILD_THREAD_POINTER" },
    { 0x35, "CALL_BUILTIN_FUNC_POINTER" },
    { 0x36, "CALL_BUILTIN_METHOD_POINTER" },
    { 0x37, "GET_ISTRING" },
    { 0x38, "GET_VECTOR" },
    { 0x39, "GET_LEVEL_OBJ" },
    { 0x3A, "GET_ANIM_OBJ" },
    { 0x3B, "GET_SELF" },
    { 0x3C, "GET_THISTHREAD" },
    { 0x3D, "GET_LEVEL" },
    { 0x3E, "GET_GAME" },
    { 0x3F, "GET_ANIM" },
    { 0x40, "GET_ANIMATION" },
    { 0x41, "GET_GAME_REF" },
    { 0x42, "INC" },
    { 0x43, "DEC" },
    { 0x44, "BIT_OR" },
    { 0x45, "JMP_EXPR_FALSE" },
    { 0x46, "BIT_EXOR" },
    { 0x47, "BIT_AND" },
    { 0x48, "EQUALITY" },
    { 0x49, "INEQUALITY" },
    { 0x4A, "LESS" },
    { 0x4B, "GREATER" },
    { 0x4C, "JMP_EXPR_TRUE" },
    { 0x4D, "LESSEQUAL" },
    { 0x4E, "JMP_BACK" },
    { 0x4F, "WAITTILLMATCH2" },
    { 0x50, "WAITTILL" },
    { 0x51, "NOTIFY" },
    { 0x52, "ENDON" },
    { 0x53, "VOIDCODEPOS" },
    { 0x54, "SWITCH" },
    { 0x55, "ENDSWITCH" },
    { 0x56, "VECTOR" },
    { 0x57, "JMP_FALSE" },
    { 0x58, "GREATEREQUAL" },
    { 0x59, "SHIFT_LEFT" },
    { 0x5A, "SHIFT_RIGHT" },
    { 0x5B, "PLUS" },
    { 0x5C, "JMP" },
    { 0x5D, "MINUS" },
    { 0x5E, "MULT" },
    { 0x5F, "DIV" },
    { 0x60, "MOD" },
    { 0x61, "JMP_TRUE" },
    { 0x62, "SIZE" },
    { 0x63, "WAITTILLMATCH" },
    { 0x64, "GET_LOCAL_FUNC" },
    { 0x65, "GET_FAR_FUNC" },
    { 0x66, "GET_SELF_OBJ" },
    { 0x67, "EVAL_LEVEL_FIELD_VARIABLE" },
    { 0x68, "EVAL_ANIM_FIELD_VARIABLE" },
    { 0x69, "EVAL_SELF_FIELD_VARIABLE" },
    { 0x6A, "EVAL_FIELD_VARIABLE" },
    { 0x6B, "EVAL_LEVEL_FIELD_VARIABLE_REF" },
    { 0x6C, "EVAL_ANIM_FIELD_VARIABLE_REF" },
    { 0x6D, "EVAL_SELF_FIELD_VARIABLE_REF" },
    { 0x6E, "EVAL_FIELD_VARIABLE_REF" },
    { 0x6F, "CLEAR_FIELD_VARIABLE" },
    { 0x70, "SAFE_CREATE_VARIABLE_FIELD_CACHED" },
    { 0x71, "SAFE_SET_VARIABLE_FIELD_CACHED0" },
    { 0x72, "SAFE_SET_VARIABLE_FIELD_CACHED" },
    { 0x73, "SAFE_SET_WAITTILL_VARIABLE_FIELD_CACHED" },
    { 0x74, "GET_ANIMTREE" },
    { 0x75, "CLEAR_PARAMS" },
    { 0x76, "CHECK_CLEAR_PARAMS" },
    { 0x77, "EVAL_LOCAL_VARIABLE_REF_CACHED0" },
    { 0x78, "EVAL_NEW_LOCAL_VARIABLE_REF_CACHED0" },
    { 0x79, "EVAL_LOCAL_VARIABLE_REF_CACHED" },
    { 0x7A, "SET_LEVEL_FIELD_VARIABLE_FIELD" },
    { 0x7B, "SET_VARIABLE_FIELD" },
    { 0x7C, "CLEAR_VARIABLE_FIELD" },
    { 0x7D, "SET_ANIM_FIELD_VARIABLE_FIELD" },
    { 0x7E, "SET_SELF_FIELD_VARIABLE_FIELD" },
    { 0x7F, "SET_LOCAL_VARIABLE_FIELD_CACHED0" },
    { 0x80, "SET_NEW_LOCAL_VARIABLE_FIELD_CACHED0" },
    { 0x81, "SET_LOCAL_VARIABLE_FIELD_CACHED" },
    { 0x82, "CLEAR_LOCAL_VARIABLE_FIELD_CACHED" },
    { 0x83, "CLEAR_LOCAL_VARIABLE_FIELD_CACHED0" },
    { 0x84, "CALL_BUILTIN_FUNC_0" },
    { 0x85, "CALL_BUILTIN_FUNC_1" },
    { 0x86, "CALL_BUILTIN_FUNC_2" },
    { 0x87, "CALL_BUILTIN_FUNC_3" },
    { 0x88, "CALL_BUILTIN_FUNC_4" },
    { 0x89, "CALL_BUILTIN_FUNC_5" },
    { 0x8A, "CALL_BUILTIN_FUNC" },
    { 0x8B, "CALL_BUILTIN_METHOD_0" },
    { 0x8C, "CALL_BUILTIN_METHOD_1" },
    { 0x8D, "CALL_BUILTIN_METHOD_2" },
    { 0x8E, "CALL_BUILTIN_METHOD_3" },
    { 0x8F, "CALL_BUILTIN_METHOD_4" },
    { 0x90, "CALL_BUILTIN_METHOD_5" },
    { 0x91, "CALL_BUILTIN_METHOD" },
    { 0x92, "WAIT" },
    { 0x93, "DEC_TOP" },
    { 0x94, "CAST_FIELD_OBJ" },
    { 0x95, "EVAL_LOCAL_VARIABLE_OBJECT_CACHED" },
    { 0x96, "CAST_BOOL" },
    { 0x97, "BOOL_NOT" },
    { 0x98, "BOOL_COMPLEMENT" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 455> function_list
{{
    { 0x001, "precacheturret" },
    { 0x002, "getweaponarray" },
    { 0x003, "createprintchannel" },
    { 0x004, "updategamerprofileall" },
    { 0x005, "clearlocalizedstrings" },
    { 0x006, "setphysicsgravitydir" },
    { 0x007, "gettimescale" },
    { 0x008, "settimescale" },
    { 0x009, "setslowmotionview" },
    { 0x00A, "forcesharedammo" },
    { 0x00B, "refreshhudcompass" },
    { 0x00C, "refreshhudammocounter" },
    { 0x00D, "notifyoncommand" },
    { 0x00E, "setprintchannel" },
    { 0x00F, "print" },
    { 0x010, "println" },
    { 0x011, "print3d" },
    { 0x012, "line" },
    { 0x013, "spawnturret" },
    { 0x014, "canspawnturret" },
    { 0x015, "assert" },
    { 0x016, "pausecinematicingame" },
    { 0x017, "drawcompassfriendlies" },
    { 0x018, "bulletspread" },
    { 0x019, "bullettracer" },
    { 0x01A, "badplace_delete" },
    { 0x01B, "badplace_cylinder" },
    { 0x01C, "badplace_arc" },
    { 0x01D, "badplace_brush" },
    { 0x01E, "clearallcorpses" },
    { 0x01F, "setturretnode" },
    { 0x020, "unsetturretnode" },
    { 0x021, "setnodepriority" },
    { 0x022, "isnodeoccupied" },
    { 0x023, "setdebugorigin" },
    { 0x024, "setdebugangles" },
    { 0x025, "updategamerprofile" },
    { 0x026, "assertex" },
    { 0x027, "assertmsg" },
    { 0x028, "isdefined" },
    { 0x029, "isstring" },
    { 0x02A, "setdvar" },
    { 0x02B, "setdynamicdvar" },
    { 0x02C, "setdvarifuninitialized" },
    { 0x02D, "setdevdvar" },
    { 0x02E, "setdevdvarifuninitialized" },
    { 0x02F, "getdvar" },
    { 0x030, "getdvarint" },
    { 0x031, "getdvarfloat" },
    { 0x032, "getdvarvector" },
    { 0x033, "gettime" },
    { 0x034, "getentbynum" },
    { 0x035, "getweaponmodel" },
    { 0x036, "getculldist" },
    { 0x037, "sethalfresparticles" },
    { 0x038, "getmapsunlight" },
    { 0x039, "setsunlight" },
    { 0x03A, "resetsunlight" },
    { 0x03B, "getmapsundirection" },
    { 0x03C, "getmapsunangles" },
    { 0x03D, "setsundirection" },
    { 0x03E, "lerpsundirection" },
    { 0x03F, "lerpsunangles" },
    { 0x040, "resetsundirection" },
    { 0x041, "enableforcedsunshadows" },
    { 0x042, "enableforcednosunshadows" },
    { 0x043, "disableforcedsunshadows" },
    { 0x044, "enableouterspacemodellighting" },
    { 0x045, "disableouterspacemodellighting" },
    { 0x046, "remapstage" },
    { 0x047, "changelevel" },
    { 0x048, "missionsuccess" },
    { 0x049, "missionfailed" },
    { 0x04A, "cinematic" },
    { 0x04B, "cinematicingame" },
    { 0x04C, "cinematicingamesync" },
    { 0x04D, "cinematicingameloop" },
    { 0x04E, "cinematicingameloopresident" },
    { 0x04F, "iscinematicplaying" },
    { 0x050, "stopcinematicingame" },
    { 0x051, "getweaponhidetags" },
    { 0x052, "getanimlength" },
    { 0x053, "animhasnotetrack" },
    { 0x054, "getnotetracktimes" },
    { 0x055, "spawn" },
    { 0x056, "spawnloopsound" },
    { 0x057, "spawnloopingsound" },
    { 0x058, "bullettrace" },
    { 0x059, "target_setmaxsize" },
    { 0x05A, "target_setcolor" },
    { 0x05B, "target_setdelay" },
    { 0x05C, "getstartorigin" },
    { 0x05D, "getstartangles" },
    { 0x05E, "getcycleoriginoffset" },
    { 0x05F, "magicgrenade" },
    { 0x060, "magicgrenademanual" },
    { 0x061, "setblur" },
    { 0x062, "musicplay" },
    { 0x063, "musicstop" },
    { 0x064, "soundfade" },
    { 0x065, "soundsettimescalefactor" },
    { 0x066, "soundresettimescale" },
    { 0x067, "setocclusionpreset" },
    { 0x068, "levelsoundfade" },
    { 0x069, "precachenightvisioncodeassets" },
    { 0x06A, "precachedigitaldistortcodeassets" },
    { 0x06B, "precacheminimapsentrycodeassets" },
    { 0x06C, "savegame" },
    { 0x06D, "issavesuccessful" },
    { 0x06E, "issaverecentlyloaded" },
    { 0x06F, "savegamenocommit" },
    { 0x070, "commitsave" },
    { 0x071, "commitwouldbevalid" },
    { 0x072, "getfxvisibility" },
    { 0x073, "setculldist" },
    { 0x074, "bullettracepassed" },
    { 0x075, "sighttracepassed" },
    { 0x076, "physicstrace" },
    { 0x077, "playerphysicstrace" },
    { 0x078, "getgroundposition" },
    { 0x079, "getmovedelta" },
    { 0x07A, "getangledelta" },
    { 0x07B, "getnorthyaw" },
    { 0x07C, "getcommandfromkey" },
    { 0x07D, "getsticksconfig" },
    { 0x07E, "weaponfightdist" },
    { 0x07F, "weaponmaxdist" },
    { 0x080, "isturretactive" },
    { 0x081, "target_alloc" },
    { 0x082, "target_flush" },
    { 0x083, "target_set" },
    { 0x084, "target_remove" },
    { 0x085, "target_setshader" },
    { 0x086, "target_setoffscreenshader" },
    { 0x087, "target_isinrect" },
    { 0x088, "target_isincircle" },
    { 0x089, "target_startreticlelockon" },
    { 0x08A, "target_clearreticlelockon" },
    { 0x08B, "target_getarray" },
    { 0x08C, "target_istarget" },
    { 0x08D, "target_setattackmode" },
    { 0x08E, "target_setjavelinonly" },
    { 0x08F, "target_hidefromplayer" },
    { 0x090, "target_showtoplayer" },
    { 0x091, "target_setscaledrendermode" },
    { 0x092, "target_drawcornersonly" },
    { 0x093, "target_drawsquare" },
    { 0x094, "target_drawsingle" },
    { 0x095, "target_setminsize" },
    { 0x096, "setnorthyaw" },
    { 0x097, "setslowmotion" },
    { 0x098, "randomint" },
    { 0x099, "randomfloat" },
    { 0x09A, "randomintrange" },
    { 0x09B, "randomfloatrange" },
    { 0x09C, "sin" },
    { 0x09D, "cos" },
    { 0x09E, "tan" },
    { 0x09F, "asin" },
    { 0x0A0, "acos" },
    { 0x0A1, "atan" },
    { 0x0A2, "int" },
    { 0x0A3, "float" },
    { 0x0A4, "abs" },
    { 0x0A5, "min" },
    { 0x0A6, "objective_additionalcurrent" },
    { 0x0A7, "objective_ring" },
    { 0x0A8, "objective_setpointertextoverride" },
    { 0x0A9, "getnode" },
    { 0x0AA, "getnodearray" },
    { 0x0AB, "getallnodes" },
    { 0x0AC, "getnodesinradius" },
    { 0x0AD, "getnodesinradiussorted" },
    { 0x0AE, "getclosestnodeinsight" },
    { 0x0AF, "getreflectionlocs" },
    { 0x0B0, "getreflectionreferencelocs" },
    { 0x0B1, "getvehicletracksegment" },
    { 0x0B2, "getvehicletracksegmentarray" },
    { 0x0B3, "getallvehicletracksegments" },
    { 0x0B4, "isarray" },
    { 0x0B5, "isai" },
    { 0x0B6, "issentient" },
    { 0x0B7, "isgodmode" },
    { 0x0B8, "getdebugdvar" },
    { 0x0B9, "getdebugdvarint" },
    { 0x0BA, "getdebugdvarfloat" },
    { 0x0BB, "setsaveddvar" },
    { 0x0BC, "getfreeaicount" },
    { 0x0BD, "getaicount" },
    { 0x0BE, "getaiarray" },
    { 0x0BF, "getaispeciesarray" },
    { 0x0C0, "getspawnerarray" },
    { 0x0C1, "getcorpsearray" },
    { 0x0C2, "getspawnerteamarray" },
    { 0x0C3, "getweaponclipmodel" },
    { 0x0C4, "getbrushmodelcenter" },
    { 0x0C5, "getkeybinding" },
    { 0x0C6, "max" },
    { 0x0C7, "floor" },
    { 0x0C8, "ceil" },
    { 0x0C9, "exp" },
    { 0x0CA, "log" },
    { 0x0CB, "sqrt" },
    { 0x0CC, "squared" },
    { 0x0CD, "clamp" },
    { 0x0CE, "angleclamp" },
    { 0x0CF, "angleclamp180" },
    { 0x0D0, "vectorfromlinetopoint" },
    { 0x0D1, "pointonsegmentnearesttopoint" },
    { 0x0D2, "distance" },
    { 0x0D3, "distance2d" },
    { 0x0D4, "distancesquared" },
    { 0x0D5, "length" },
    { 0x0D6, "lengthsquared" },
    { 0x0D7, "closer" },
    { 0x0D8, "vectordot" },
    { 0x0D9, "visionsetthermal" },
    { 0x0DA, "visionsetpain" },
    { 0x0DB, "endlobby" },
    { 0x0DC, "setac130ambience" },
    { 0x0DD, "getmapcustom" },
    { 0x0DE, "updateskill" },
    { 0x0DF, "spawnsighttrace" },
    { 0x0E0, "incrementcounter" },
    { 0x0E1, "getcountertotal" },
    { 0x0E2, "getlevelticks" },
    { 0x0E3, "perlinnoise2d" },
    { 0x0E4, "calcrockingangles" },
    { 0x0E5, "sethudlighting" },
    { 0x0E6, "reconevent" },
    { 0x0E7, "reconspatialevent" },
    { 0x0E8, "setsunflareposition" },
    { 0x0E9, "createthreatbiasgroup" },
    { 0x0EA, "threatbiasgroupexists" },
    { 0x0EB, "getthreatbias" },
    { 0x0EC, "setthreatbias" },
    { 0x0ED, "setthreatbiasagainstall" },
    { 0x0EE, "setignoremegroup" },
    { 0x0EF, "isenemyteam" },
    { 0x0F0, "objective_additionalentity" },
    { 0x0F1, "objective_state_nomessage" },
    { 0x0F2, "objective_string" },
    { 0x0F3, "objective_string_nomessage" },
    { 0x0F4, "objective_additionalposition" },
    { 0x0F5, "objective_current_nomessage" },
    { 0x0F6, "vectornormalize" },
    { 0x0F7, "vectortoangles" },
    { 0x0F8, "vectortoyaw" },
    { 0x0F9, "vectorlerp" },
    { 0x0FA, "anglestoup" },
    { 0x0FB, "anglestoright" },
    { 0x0FC, "anglestoforward" },
    { 0x0FD, "combineangles" },
    { 0x0FE, "transformmove" },
    { 0x0FF, "issubstr" },
    { 0x100, "isendstr" },
    { 0x101, "getsubstr" },
    { 0x102, "tolower" },
    { 0x103, "strtok" },
    { 0x104, "stricmp" },
    { 0x105, "ambientplay" },
    { 0x106, "getuavstrengthmax" },
    { 0x107, "getuavstrengthlevelneutral" },
    { 0x108, "getuavstrengthlevelshowenemyfastsweep" },
    { 0x109, "getuavstrengthlevelshowenemydirectional" },
    { 0x10A, "blockteamradar" },
    { 0x10B, "unblockteamradar" },
    { 0x10C, "isteamradarblocked" },
    { 0x10D, "getassignedteam" },
    { 0x10E, "setmatchdata" },
    { 0x10F, "getmatchdata" },
    { 0x110, "sendmatchdata" },
    { 0x111, "clearmatchdata" },
    { 0x112, "setmatchdatadef" },
    { 0x113, "setmatchclientip" },
    { 0x114, "setmatchdataid" },
    { 0x115, "setclientmatchdata" },
    { 0x116, "getclientmatchdata" },
    { 0x117, "setclientmatchdatadef" },
    { 0x118, "sendclientmatchdata" },
    { 0x119, "getbuildversion" },
    { 0x11A, "getbuildnumber" },
    { 0x11B, "getsystemtime" },
    { 0x11C, "getmatchrulesdata" },
    { 0x11D, "isusingmatchrulesdata" },
    { 0x11E, "kick" },
    { 0x11F, "issplitscreen" },
    { 0x120, "setmapcenter" },
    { 0x121, "setgameendtime" },
    { 0x122, "visionsetnaked" },
    { 0x123, "visionsetnight" },
    { 0x124, "visionsetmissilecam" },
    { 0x125, "ambientstop" },
    { 0x126, "precachemodel" },
    { 0x127, "precacheshellshock" },
    { 0x128, "precacheitem" },
    { 0x129, "precacheshader" },
    { 0x12A, "precachestring" },
    { 0x12B, "precachemenu" },
    { 0x12C, "precacherumble" },
    { 0x12D, "precachelocationselector" },
    { 0x12E, "precacheleaderboards" },
    { 0x12F, "loadfx" },
    { 0x130, "playfx" },
    { 0x131, "playfxontag" },
    { 0x132, "stopfxontag" },
    { 0x133, "playloopedfx" },
    { 0x134, "spawnfx" },
    { 0x135, "triggerfx" },
    { 0x136, "playfxontagforclients" },
    { 0x137, "setwinningteam" },
    { 0x138, "announcement" },
    { 0x139, "clientannouncement" },
    { 0x13A, "getteamscore" },
    { 0x13B, "setteamscore" },
    { 0x13C, "setclientnamemode" },
    { 0x13D, "updateclientnames" },
    { 0x13E, "getteamplayersalive" },
    { 0x13F, "logprint" },
    { 0x140, "worldentnumber" },
    { 0x141, "obituary" },
    { 0x142, "positionwouldtelefrag" },
    { 0x143, "canspawn" },
    { 0x144, "getstarttime" },
    { 0x145, "precachestatusicon" },
    { 0x146, "precacheheadicon" },
    { 0x147, "precacheminimapicon" },
    { 0x148, "precachempanim" },
    { 0x149, "map_restart" },
    { 0x14A, "exitlevel" },
    { 0x14B, "addtestclient" },
    { 0x14C, "makedvarserverinfo" },
    { 0x14D, "setarchive" },
    { 0x14E, "allclientsprint" },
    { 0x14F, "clientprint" },
    { 0x150, "mapexists" },
    { 0x151, "isvalidgametype" },
    { 0x152, "matchend" },
    { 0x153, "setplayerteamrank" },
    { 0x154, "endparty" },
    { 0x155, "setteamradar" },
    { 0x156, "getteamradar" },
    { 0x157, "setteamradarstrength" },
    { 0x158, "getteamradarstrength" },
    { 0x159, "getuavstrengthmin" },
    { 0x15A, "physicsexplosionsphere" },
    { 0x15B, "physicsexplosioncylinder" },
    { 0x15C, "physicsjolt" },
    { 0x15D, "physicsjitter" },
    { 0x15E, "setexpfog" },
    { 0x15F, "isexplosivedamagemod" },
    { 0x160, "radiusdamage" },
    { 0x161, "setplayerignoreradiusdamage" },
    { 0x162, "glassradiusdamage" },
    { 0x163, "earthquake" },
    { 0x164, "getnumparts" },
    { 0x165, "objective_onentity" },
    { 0x166, "objective_team" },
    { 0x167, "objective_player" },
    { 0x168, "objective_playerteam" },
    { 0x169, "objective_playerenemyteam" },
    { 0x16A, "iprintln" },
    { 0x16B, "iprintlnbold" },
    { 0x16C, "logstring" },
    { 0x16D, "getent" },
    { 0x16E, "getentarray" },
    { 0x16F, "spawnplane" },
    { 0x170, "spawnstruct" },
    { 0x171, "spawnhelicopter" },
    { 0x172, "isalive" },
    { 0x173, "isspawner" },
    { 0x174, "missile_createattractorent" },
    { 0x175, "missile_createattractororigin" },
    { 0x176, "missile_createrepulsorent" },
    { 0x177, "missile_createrepulsororigin" },
    { 0x178, "missile_deleteattractor" },
    { 0x179, "playsoundatpos" },
    { 0x17A, "newhudelem" },
    { 0x17B, "newclienthudelem" },
    { 0x17C, "newteamhudelem" },
    { 0x17D, "resettimeout" },
    { 0x17E, "precachefxteamthermal" },
    { 0x17F, "isplayer" },
    { 0x180, "isplayernumber" },
    { 0x181, "setwinningplayer" },
    { 0x182, "getpartname" },
    { 0x183, "weaponfiretime" },
    { 0x184, "weaponclipsize" },
    { 0x185, "weaponisauto" },
    { 0x186, "weaponissemiauto" },
    { 0x187, "weaponisboltaction" },
    { 0x188, "weaponinheritsperks" },
    { 0x189, "weaponburstcount" },
    { 0x18A, "weapontype" },
    { 0x18B, "weaponclass" },
    { 0x18C, "getnextarraykey" },
    { 0x18D, "sortbydistance" },
    { 0x18E, "tablelookup" },
    { 0x18F, "tablelookupbyrow" },
    { 0x190, "tablelookupistring" },
    { 0x191, "tablelookupistringbyrow" },
    { 0x192, "tablelookuprownum" },
    { 0x193, "getmissileowner" },
    { 0x194, "magicbullet" },
    { 0x195, "getweaponflashtagname" },
    { 0x196, "averagepoint" },
    { 0x197, "averagenormal" },
    { 0x198, "vehicle_getspawnerarray" },
    { 0x199, "playrumbleonposition" },
    { 0x19A, "playrumblelooponposition" },
    { 0x19B, "stopallrumbles" },
    { 0x19C, "soundexists" },
    { 0x19D, "openfile" },
    { 0x19E, "closefile" },
    { 0x19F, "fprintln" },
    { 0x1A0, "fprintfields" },
    { 0x1A1, "freadln" },
    { 0x1A2, "fgetarg" },
    { 0x1A3, "setminimap" },
    { 0x1A4, "setthermalbodymaterial" },
    { 0x1A5, "getarraykeys" },
    { 0x1A6, "getfirstarraykey" },
    { 0x1A7, "getglass" },
    { 0x1A8, "getglassarray" },
    { 0x1A9, "getglassorigin" },
    { 0x1AA, "isglassdestroyed" },
    { 0x1AB, "destroyglass" },
    { 0x1AC, "deleteglass" },
    { 0x1AD, "getentchannelscount" },
    { 0x1AE, "getentchannelname" },
    { 0x1AF, "objective_add" },
    { 0x1B0, "objective_delete" },
    { 0x1B1, "objective_state" },
    { 0x1B2, "objective_icon" },
    { 0x1B3, "objective_position" },
    { 0x1B4, "objective_current" },
    { 0x1B5, "weaponinventorytype" },
    { 0x1B6, "weaponstartammo" },
    { 0x1B7, "weaponmaxammo" },
    { 0x1B8, "weaponaltweaponname" },
    { 0x1B9, "isweaponcliponly" },
    { 0x1BA, "isweapondetonationtimed" },
    { 0x1BB, "weaponhasthermalscope" },
    { 0x1BC, "getvehiclenode" },
    { 0x1BD, "getvehiclenodearray" },
    { 0x1BE, "getallvehiclenodes" },
    { 0x1BF, "getnumvehicles" },
    { 0x1C0, "precachevehicle" },
    { 0x1C1, "spawnvehicle" },
    { 0x1C2, "vehicle_getarray" },
    { 0x1C3, "gettimesincelastpaused" },
    { 0x1C4, "setlasermaterial" },
    { 0x1C5, "precachefxontag" },
    { 0x1C6, "precachetag" },
    { 0x1C7, "precachesound" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 780> method_list
{{
    { 0x8000, "thermaldrawdisable" },
    { 0x8001, "setturretdismountorg" },
    { 0x8002, "setdamagestate" },
    { 0x8003, "playsoundtoteam" },
    { 0x8004, "playsoundtoplayer" },
    { 0x8005, "playerhide" },
    { 0x8006, "showtoplayer" },
    { 0x8007, "enableplayeruse" },
    { 0x8008, "disableplayeruse" },
    { 0x8009, "makescrambler" },
    { 0x800A, "makeportableradar" },
    { 0x800B, "maketrophysystem" },
    { 0x800C, "placespawnpoint" },
    { 0x800D, "setteamfortrigger" },
    { 0x800E, "clientclaimtrigger" },
    { 0x800F, "clientreleasetrigger" },
    { 0x8010, "releaseclaimedtrigger" },
    { 0x8011, "isusingonlinedataoffline" },
    { 0x8012, "getrestedtime" },
    { 0x8013, "sendleaderboards" },
    { 0x8014, "isonladder" },
    { 0x8015, "getcorpseanim" },
    { 0x8016, "playerforcedeathanim" },
    { 0x8017, "attach" },
    { 0x8018, "attachshieldmodel" },
    { 0x8019, "getlightfovinner" },
    { 0x801A, "getlightfovouter" },
    { 0x801B, "setlightfovrange" },
    { 0x801C, "getlightexponent" },
    { 0x801D, "setlightexponent" },
    { 0x801E, "startragdoll" },
    { 0x801F, "startragdollfromimpact" },
    { 0x8020, "logstring" },
    { 0x8021, "laserhidefromclient" },
    { 0x8022, "stopsoundchannel" },
    { 0x8023, "thermaldrawenable" },
    { 0x8024, "detach" },
    { 0x8025, "detachshieldmodel" },
    { 0x8026, "moveshieldmodel" },
    { 0x8027, "detachall" },
    { 0x8028, "getattachsize" },
    { 0x8029, "getattachmodelname" },
    { 0x802A, "getattachtagname" },
    { 0x802B, "setturretcanaidetach" },
    { 0x802C, "setturretfov" },
    { 0x802D, "lerpfov" },
    { 0x802E, "getvalidcoverpeekouts" },
    { 0x802F, "gethighestnodestance" },
    { 0x8030, "doesnodeallowstance" },
    { 0x8031, "getgunangles" },
    { 0x8032, "magicgrenade" },
    { 0x8033, "magicgrenademanual" },
    { 0x8034, "setfriendlychain" },
    { 0x8035, "getentnum" },
    { 0x8036, "launch" },
    { 0x8037, "setsoundblend" },
    { 0x8038, "makefakeai" },
    { 0x8039, "spawndrone" },
    { 0x803A, "setcorpseremovetimer" },
    { 0x803B, "setlookattext" },
    { 0x803C, "setspawnerteam" },
    { 0x803D, "addaieventlistener" },
    { 0x803E, "removeaieventlistener" },
    { 0x803F, "getlightcolor" },
    { 0x8040, "setlightcolor" },
    { 0x8041, "getlightradius" },
    { 0x8042, "setlightradius" },
    { 0x8043, "getattachignorecollision" },
    { 0x8044, "hidepart" },
    { 0x8045, "hidepart_allinstances" },
    { 0x8046, "hideallparts" },
    { 0x8047, "showpart" },
    { 0x8048, "showallparts" },
    { 0x8049, "linkto" },
    { 0x804A, "linktoblendtotag" },
    { 0x804B, "unlink" },
    { 0x804C, "setnormalhealth" },
    { 0x804D, "dodamage" },
    { 0x804E, "kill" },
    { 0x804F, "show" },
    { 0x8050, "hide" },
    { 0x8051, "showonclient" },
    { 0x8052, "hideonclient" },
    { 0x8053, "laserforceon" },
    { 0x8054, "laserforceoff" },
    { 0x8055, "disconnectpaths" },
    { 0x8056, "connectpaths" },
    { 0x8057, "startusingheroonlylighting" },
    { 0x8058, "stopusingheroonlylighting" },
    { 0x8059, "startusinglessfrequentlighting" },
    { 0x805A, "stopusinglessfrequentlighting" },
    { 0x805B, "setthermalfog" },
    { 0x805C, "setnightvisionfog" },
    { 0x805D, "clearthermalfog" },
    { 0x805E, "clearnightvisionfog" },
    { 0x805F, "digitaldistortsetparams" },
    { 0x8060, "setmode" },
    { 0x8061, "getmode" },
    { 0x8062, "setturretignoregoals" },
    { 0x8063, "islinked" },
    { 0x8064, "enablelinkto" },
    { 0x8065, "playsoundatviewheight" },
    { 0x8066, "prefetchsound" },
    { 0x8067, "setpitch" },
    { 0x8068, "scalepitch" },
    { 0x8069, "setvolume" },
    { 0x806A, "scalevolume" },
    { 0x806B, "setspeakermapmonotostereo" },
    { 0x806C, "setspeakermapmonoto51" },
    { 0x806D, "setdistributed2dsound" },
    { 0x806E, "playsoundasmaster" },
    { 0x806F, "playloopsound" },
    { 0x8070, "eqon" },
    { 0x8071, "eqoff" },
    { 0x8072, "haseq" },
    { 0x8073, "iswaitingonsound" },
    { 0x8074, "getnormalhealth" },
    { 0x8075, "playerlinkto" },
    { 0x8076, "playerlinktodelta" },
    { 0x8077, "playerlinkweaponviewtodelta" },
    { 0x8078, "playerlinktoabsolute" },
    { 0x8079, "playerlinktoblend" },
    { 0x807A, "playerlinkedoffsetenable" },
    { 0x807B, "setwaypointedgestyle_secondaryarrow" },
    { 0x807C, "setwaypointiconoffscreenonly" },
    { 0x807D, "fadeovertime" },
    { 0x807E, "scaleovertime" },
    { 0x807F, "moveovertime" },
    { 0x8080, "reset" },
    { 0x8081, "destroy" },
    { 0x8082, "setpulsefx" },
    { 0x8083, "setplayernamestring" },
    { 0x8084, "changefontscaleovertime" },
    { 0x8085, "startignoringspotlight" },
    { 0x8086, "stopignoringspotlight" },
    { 0x8087, "dontcastshadows" },
    { 0x8088, "castshadows" },
    { 0x8089, "setstablemissile" },
    { 0x808A, "playersetgroundreferenceent" },
    { 0x808B, "dontinterpolate" },
    { 0x808C, "dospawn" },
    { 0x808D, "stalingradspawn" },
    { 0x808E, "getorigin" },
    { 0x808F, "getcentroid" },
    { 0x8090, "getshootatpos" },
    { 0x8091, "getdebugeye" },
    { 0x8092, "useby" },
    { 0x8093, "playsound" },
    { 0x8094, "playerlinkedoffsetdisable" },
    { 0x8095, "playerlinkedsetviewznear" },
    { 0x8096, "playerlinkedsetusebaseangleforviewclamp" },
    { 0x8097, "lerpviewangleclamp" },
    { 0x8098, "setviewangleresistance" },
    { 0x8099, "geteye" },
    { 0x809A, "istouching" },
    { 0x809B, "stoploopsound" },
    { 0x809C, "stopsounds" },
    { 0x809D, "playrumbleonentity" },
    { 0x809E, "playrumblelooponentity" },
    { 0x809F, "stoprumble" },
    { 0x80A0, "delete" },
    { 0x80A1, "setmodel" },
    { 0x80A2, "laseron" },
    { 0x80A3, "laseroff" },
    { 0x80A4, "laseraltviewon" },
    { 0x80A5, "laseraltviewoff" },
    { 0x80A6, "thermalvisionon" },
    { 0x80A7, "thermalvisionoff" },
    { 0x80A8, "thermalvisionfofoverlayon" },
    { 0x80A9, "thermalvisionfofoverlayoff" },
    { 0x80AA, "autospotoverlayon" },
    { 0x80AB, "autospotoverlayoff" },
    { 0x80AC, "setcontents" },
    { 0x80AD, "makeusable" },
    { 0x80AE, "makeunusable" },
    { 0x80AF, "setwhizbyprobabilities" },
    { 0x80B0, "visionsetnakedforplayer_lerp" },
    { 0x80B1, "setwaitnode" },
    { 0x80B2, "returnplayercontrol" },
    { 0x80B3, "vehphys_starttrack" },
    { 0x80B4, "vehphys_clearautodisable" },
    { 0x80B5, "vehicleusealtblendedaudio" },
    { 0x80B6, "settext" },
    { 0x80B7, "clearalltextafterhudelem" },
    { 0x80B8, "setshader" },
    { 0x80B9, "settargetent" },
    { 0x80BA, "cleartargetent" },
    { 0x80BB, "settimer" },
    { 0x80BC, "settimerup" },
    { 0x80BD, "settimerstatic" },
    { 0x80BE, "settenthstimer" },
    { 0x80BF, "settenthstimerup" },
    { 0x80C0, "settenthstimerstatic" },
    { 0x80C1, "setclock" },
    { 0x80C2, "setclockup" },
    { 0x80C3, "setvalue" },
    { 0x80C4, "setwaypoint" },
    { 0x80C5, "setwaypointedgestyle_rotatingicon" },
    { 0x80C6, "setcursorhint" },
    { 0x80C7, "sethintstring" },
    { 0x80C8, "forceusehinton" },
    { 0x80C9, "forceusehintoff" },
    { 0x80CA, "makesoft" },
    { 0x80CB, "makehard" },
    { 0x80CC, "willneverchange" },
    { 0x80CD, "startfiring" },
    { 0x80CE, "stopfiring" },
    { 0x80CF, "isfiringturret" },
    { 0x80D0, "startbarrelspin" },
    { 0x80D1, "stopbarrelspin" },
    { 0x80D2, "getbarrelspinrate" },
    { 0x80D3, "remotecontrolturret" },
    { 0x80D4, "remotecontrolturretoff" },
    { 0x80D5, "shootturret" },
    { 0x80D6, "getturretowner" },
    { 0x80D7, "enabledeathshield" },
    { 0x80D8, "nightvisiongogglesforceon" },
    { 0x80D9, "nightvisiongogglesforceoff" },
    { 0x80DA, "enableinvulnerability" },
    { 0x80DB, "disableinvulnerability" },
    { 0x80DC, "enablebreaching" },
    { 0x80DD, "disablebreaching" },
    { 0x80DE, "forceviewmodelanimation" },
    { 0x80DF, "disableturretdismount" },
    { 0x80E0, "enableturretdismount" },
    { 0x80E1, "uploadscore" },
    { 0x80E2, "uploadtime" },
    { 0x80E3, "uploadleaderboards" },
    { 0x80E4, "giveachievement" },
    { 0x80E5, "hidehud" },
    { 0x80E6, "showhud" },
    { 0x80E7, "mountvehicle" },
    { 0x80E8, "dismountvehicle" },
    { 0x80E9, "enableslowaim" },
    { 0x80EA, "disableslowaim" },
    { 0x80EB, "usehintsinvehicle" },
    { 0x80EC, "vehicleattackbuttonpressed" },
    { 0x80ED, "setwhizbyoffset" },
    { 0x80EE, "setsentryowner" },
    { 0x80EF, "setsentrycarrier" },
    { 0x80F0, "setturretminimapvisible" },
    { 0x80F1, "settargetentity" },
    { 0x80F2, "snaptotargetentity" },
    { 0x80F3, "cleartargetentity" },
    { 0x80F4, "getturrettarget" },
    { 0x80F5, "setplayerspread" },
    { 0x80F6, "setaispread" },
    { 0x80F7, "setsuppressiontime" },
    { 0x80F8, "setflaggedanimknobrestart" },
    { 0x80F9, "setflaggedanimknoblimitedrestart" },
    { 0x80FA, "setflaggedanimknoball" },
    { 0x80FB, "setflaggedanimknoballrestart" },
    { 0x80FC, "setflaggedanim" },
    { 0x80FD, "setflaggedanimlimited" },
    { 0x80FE, "setflaggedanimrestart" },
    { 0x80FF, "setflaggedanimlimitedrestart" },
    { 0x8100, "useanimtree" },
    { 0x8101, "stopuseanimtree" },
    { 0x8102, "setanimtime" },
    { 0x8103, "showviewmodel" },
    { 0x8104, "hideviewmodel" },
    { 0x8105, "allowstand" },
    { 0x8106, "allowcrouch" },
    { 0x8107, "allowprone" },
    { 0x8108, "allowlean" },
    { 0x8109, "setocclusion" },
    { 0x810A, "deactivateocclusion" },
    { 0x810B, "isocclusionenabled" },
    { 0x810C, "iseqenabled" },
    { 0x810D, "seteq" },
    { 0x810E, "seteqbands" },
    { 0x810F, "deactivateeq" },
    { 0x8110, "seteqlerp" },
    { 0x8111, "islookingat" },
    { 0x8112, "isthrowinggrenade" },
    { 0x8113, "isfiring" },
    { 0x8114, "ismeleeing" },
    { 0x8115, "setautopickup" },
    { 0x8116, "allowmelee" },
    { 0x8117, "allowfire" },
    { 0x8118, "enablehealthshield" },
    { 0x8119, "setconvergencetime" },
    { 0x811A, "setconvergenceheightpercent" },
    { 0x811B, "setturretteam" },
    { 0x811C, "maketurretsolid" },
    { 0x811D, "maketurretoperable" },
    { 0x811E, "maketurretinoperable" },
    { 0x811F, "makeentitysentient" },
    { 0x8120, "freeentitysentient" },
    { 0x8121, "isindoor" },
    { 0x8122, "getdroptofloorposition" },
    { 0x8123, "isbadguy" },
    { 0x8124, "animscripted" },
    { 0x8125, "animscriptedthirdperson" },
    { 0x8126, "animrelative" },
    { 0x8127, "stopanimscripted" },
    { 0x8128, "clearanim" },
    { 0x8129, "setanimknob" },
    { 0x812A, "setanimknoblimited" },
    { 0x812B, "setanimknobrestart" },
    { 0x812C, "setanimknoblimitedrestart" },
    { 0x812D, "setanimknoball" },
    { 0x812E, "setanimknoballlimited" },
    { 0x812F, "setanimknoballrestart" },
    { 0x8130, "setanimknoballlimitedrestart" },
    { 0x8131, "setanim" },
    { 0x8132, "setanimlimited" },
    { 0x8133, "setanimrestart" },
    { 0x8134, "setanimlimitedrestart" },
    { 0x8135, "getanimtime" },
    { 0x8136, "getanimweight" },
    { 0x8137, "getanimassettype" },
    { 0x8138, "setflaggedanimknob" },
    { 0x8139, "setflaggedanimknoblimited" },
    { 0x813A, "setturretaccuracy" },
    { 0x813B, "setrightarc" },
    { 0x813C, "setleftarc" },
    { 0x813D, "settoparc" },
    { 0x813E, "setbottomarc" },
    { 0x813F, "setautorotationdelay" },
    { 0x8140, "setdefaultdroppitch" },
    { 0x8141, "restoredefaultdroppitch" },
    { 0x8142, "turretfiredisable" },
    { 0x8143, "getfixednodesafevolume" },
    { 0x8144, "clearfixednodesafevolume" },
    { 0x8145, "isingoal" },
    { 0x8146, "setruntopos" },
    { 0x8147, "nearnode" },
    { 0x8148, "nearclaimnode" },
    { 0x8149, "nearclaimnodeandangle" },
    { 0x814A, "atdangerousnode" },
    { 0x814B, "getenemyinfo" },
    { 0x814C, "clearenemy" },
    { 0x814D, "setentitytarget" },
    { 0x814E, "clearentitytarget" },
    { 0x814F, "setpotentialthreat" },
    { 0x8150, "clearpotentialthreat" },
    { 0x8151, "setflashbanged" },
    { 0x8152, "setengagementmindist" },
    { 0x8153, "setengagementmaxdist" },
    { 0x8154, "isknownenemyinradius" },
    { 0x8155, "isknownenemyinvolume" },
    { 0x8156, "settalktospecies" },
    { 0x8157, "laseralton" },
    { 0x8158, "laseraltoff" },
    { 0x8159, "invisiblenotsolid" },
    { 0x815A, "visiblesolid" },
    { 0x815B, "setdefaultaimlimits" },
    { 0x815C, "initriotshieldhealth" },
    { 0x815D, "getenemysqdist" },
    { 0x815E, "getclosestenemysqdist" },
    { 0x815F, "setthreatbiasgroup" },
    { 0x8160, "getthreatbiasgroup" },
    { 0x8161, "turretfireenable" },
    { 0x8162, "setturretmodechangewait" },
    { 0x8163, "usetriggerrequirelookat" },
    { 0x8164, "getstance" },
    { 0x8165, "setstance" },
    { 0x8166, "itemweaponsetammo" },
    { 0x8167, "getammocount" },
    { 0x8168, "gettagorigin" },
    { 0x8169, "gettagangles" },
    { 0x816A, "shellshock" },
    { 0x816B, "stunplayer" },
    { 0x816C, "stopshellshock" },
    { 0x816D, "fadeoutshellshock" },
    { 0x816E, "setdepthoffield" },
    { 0x816F, "setviewmodeldepthoffield" },
    { 0x8170, "setmotionblurmovescale" },
    { 0x8171, "pickupgrenade" },
    { 0x8172, "useturret" },
    { 0x8173, "stopuseturret" },
    { 0x8174, "canuseturret" },
    { 0x8175, "traversemode" },
    { 0x8176, "animmode" },
    { 0x8177, "orientmode" },
    { 0x8178, "getmotionangle" },
    { 0x8179, "shouldfacemotion" },
    { 0x817A, "getanglestolikelyenemypath" },
    { 0x817B, "setturretanim" },
    { 0x817C, "getturret" },
    { 0x817D, "getgroundenttype" },
    { 0x817E, "animcustom" },
    { 0x817F, "isinscriptedstate" },
    { 0x8180, "canattackenemynode" },
    { 0x8181, "getnegotiationstartnode" },
    { 0x8182, "getnegotiationendnode" },
    { 0x8183, "getdoorpathnode" },
    { 0x8184, "comparenodedirtopathdir" },
    { 0x8185, "checkprone" },
    { 0x8186, "pushplayer" },
    { 0x8187, "checkgrenadethrowpos" },
    { 0x8188, "setgoalnode" },
    { 0x8189, "setgoalpos" },
    { 0x818A, "setgoalentity" },
    { 0x818B, "setgoalvolume" },
    { 0x818C, "setgoalvolumeauto" },
    { 0x818D, "getgoalvolume" },
    { 0x818E, "cleargoalvolume" },
    { 0x818F, "setfixednodesafevolume" },
    { 0x8190, "setmotionblurturnscale" },
    { 0x8191, "setmotionblurzoomscale" },
    { 0x8192, "viewkick" },
    { 0x8193, "localtoworldcoords" },
    { 0x8194, "getentitynumber" },
    { 0x8195, "getentityvelocity" },
    { 0x8196, "enablegrenadetouchdamage" },
    { 0x8197, "disablegrenadetouchdamage" },
    { 0x8198, "enableaimassist" },
    { 0x8199, "setlookatyawlimits" },
    { 0x819A, "stoplookat" },
    { 0x819B, "getmuzzlepos" },
    { 0x819C, "getmuzzleangle" },
    { 0x819D, "getmuzzlesideoffsetpos" },
    { 0x819E, "getaimangle" },
    { 0x819F, "canshoot" },
    { 0x81A0, "canshootenemy" },
    { 0x81A1, "cansee" },
    { 0x81A2, "seerecently" },
    { 0x81A3, "lastknowntime" },
    { 0x81A4, "lastknownpos" },
    { 0x81A5, "dropweapon" },
    { 0x81A6, "maymovetopoint" },
    { 0x81A7, "maymovefrompointtopoint" },
    { 0x81A8, "teleport" },
    { 0x81A9, "forceteleport" },
    { 0x81AA, "safeteleport" },
    { 0x81AB, "withinapproxpathdist" },
    { 0x81AC, "ispathdirect" },
    { 0x81AD, "allowedstances" },
    { 0x81AE, "isstanceallowed" },
    { 0x81AF, "issuppressionwaiting" },
    { 0x81B0, "issuppressed" },
    { 0x81B1, "ismovesuppressed" },
    { 0x81B2, "isgrenadepossafe" },
    { 0x81B3, "checkgrenadethrow" },
    { 0x81B4, "checkgrenadelaunch" },
    { 0x81B5, "checkgrenadelaunchpos" },
    { 0x81B6, "throwgrenade" },
    { 0x81B7, "disableaimassist" },
    { 0x81B8, "radiusdamage" },
    { 0x81B9, "detonate" },
    { 0x81BA, "damageconetrace" },
    { 0x81BB, "sightconetrace" },
    { 0x81BC, "missile_settargetent" },
    { 0x81BD, "missile_settargetpos" },
    { 0x81BE, "missile_cleartarget" },
    { 0x81BF, "missile_setflightmodedirect" },
    { 0x81C0, "missile_setflightmodetop" },
    { 0x81C1, "getlightintensity" },
    { 0x81C2, "setlightintensity" },
    { 0x81C3, "isragdoll" },
    { 0x81C4, "setmovespeedscale" },
    { 0x81C5, "cameralinkto" },
    { 0x81C6, "cameraunlink" },
    { 0x81C7, "startcoverarrival" },
    { 0x81C8, "starttraversearrival" },
    { 0x81C9, "checkcoverexitposwithpath" },
    { 0x81CA, "shoot" },
    { 0x81CB, "shootblank" },
    { 0x81CC, "melee" },
    { 0x81CD, "updateplayersightaccuracy" },
    { 0x81CE, "findshufflecovernode" },
    { 0x81CF, "findnearbycovernode" },
    { 0x81D0, "findcovernode" },
    { 0x81D1, "findbestcovernode" },
    { 0x81D2, "getcovernode" },
    { 0x81D3, "usecovernode" },
    { 0x81D4, "iscovervalidagainstenemy" },
    { 0x81D5, "reacquirestep" },
    { 0x81D6, "findreacquiredirectpath" },
    { 0x81D7, "trimpathtoattack" },
    { 0x81D8, "reacquiremove" },
    { 0x81D9, "findreacquireproximatepath" },
    { 0x81DA, "flagenemyunattackable" },
    { 0x81DB, "enterprone" },
    { 0x81DC, "exitprone" },
    { 0x81DD, "setproneanimnodes" },
    { 0x81DE, "updateprone" },
    { 0x81DF, "clearpitchorient" },
    { 0x81E0, "setlookatanimnodes" },
    { 0x81E1, "setlookat" },
    { 0x81E2, "setlookatentity" },
    { 0x81E3, "controlslinkto" },
    { 0x81E4, "controlsunlink" },
    { 0x81E5, "makevehiclesolidcapsule" },
    { 0x81E6, "makevehiclesolidsphere" },
    { 0x81E7, "makevehiclesolid" },
    { 0x81E8, "remotecontrolvehicle" },
    { 0x81E9, "remotecontrolvehicleoff" },
    { 0x81EA, "isfiringvehicleturret" },
    { 0x81EB, "drivevehicleandcontrolturret" },
    { 0x81EC, "drivevehicleandcontrolturretoff" },
    { 0x81ED, "getplayersetting" },
    { 0x81EE, "getlocalplayerprofiledata" },
    { 0x81EF, "setlocalplayerprofiledata" },
    { 0x81F0, "remotecamerasoundscapeon" },
    { 0x81F1, "remotecamerasoundscapeoff" },
    { 0x81F2, "radarjamon" },
    { 0x81F3, "radarjamoff" },
    { 0x81F4, "setmotiontrackervisible" },
    { 0x81F5, "getmotiontrackervisible" },
    { 0x81F6, "worldpointinreticle_circle" },
    { 0x81F7, "getpointinbounds" },
    { 0x81F8, "transfermarkstonewscriptmodel" },
    { 0x81F9, "setwatersheeting" },
    { 0x81FA, "setweaponhudiconoverride" },
    { 0x81FB, "getweaponhudiconoverride" },
    { 0x81FC, "setempjammed" },
    { 0x81FD, "playersetexpfog" },
    { 0x81FE, "isitemunlocked" },
    { 0x81FF, "getplayerdata" },
    { 0x8200, "vehicleturretcontroloff" },
    { 0x8201, "isturretready" },
    { 0x8202, "vehicledriveto" },
    { 0x8203, "dospawn" },
    { 0x8204, "isphysveh" },
    { 0x8205, "phys_crash" },
    { 0x8206, "phys_launch" },
    { 0x8207, "phys_disablecrashing" },
    { 0x8208, "phys_enablecrashing" },
    { 0x8209, "phys_setspeed" },
    { 0x820A, "phys_setconveyerbelt" },
    { 0x820B, "freehelicopter" },
    { 0x820C, "playerlinkedturretanglesenable" },
    { 0x820D, "playerlinkedturretanglesdisable" },
    { 0x820E, "playersetstreamorigin" },
    { 0x820F, "playerclearstreamorigin" },
    { 0x8210, "nightvisionviewon" },
    { 0x8211, "nightvisionviewoff" },
    { 0x8212, "painvisionon" },
    { 0x8213, "painvisionoff" },
    { 0x8214, "getplayerintelisfound" },
    { 0x8215, "setplayerintelfound" },
    { 0x8216, "newpip" },
    { 0x8217, "sethuddynlight" },
    { 0x8218, "startscriptedanim" },
    { 0x8219, "startcoverbehavior" },
    { 0x821A, "setplayerdata" },
    { 0x821B, "trackerupdate" },
    { 0x821C, "pingplayer" },
    { 0x821D, "buttonpressed" },
    { 0x821E, "sayall" },
    { 0x821F, "sayteam" },
    { 0x8220, "showscoreboard" },
    { 0x8221, "setspawnweapon" },
    { 0x8222, "dropitem" },
    { 0x8223, "dropscavengerbag" },
    { 0x8224, "setjitterparams" },
    { 0x8225, "sethoverparams" },
    { 0x8226, "joltbody" },
    { 0x8227, "freevehicle" },
    { 0x8228, "getwheelsurface" },
    { 0x8229, "getvehicleowner" },
    { 0x822A, "setvehiclelookattext" },
    { 0x822B, "setvehicleteam" },
    { 0x822C, "setneargoalnotifydist" },
    { 0x822D, "setvehgoalpos" },
    { 0x822E, "setgoalyaw" },
    { 0x822F, "cleargoalyaw" },
    { 0x8230, "settargetyaw" },
    { 0x8231, "cleartargetyaw" },
    { 0x8232, "vehicle_helisetai" },
    { 0x8233, "setturrettargetvec" },
    { 0x8234, "setturrettargetent" },
    { 0x8235, "clearturrettarget" },
    { 0x8236, "vehicle_canturrettargetpoint" },
    { 0x8237, "setlookatent" },
    { 0x8238, "clearlookatent" },
    { 0x8239, "setvehweapon" },
    { 0x823A, "fireweapon" },
    { 0x823B, "vehicleturretcontrolon" },
    { 0x823C, "finishplayerdamage" },
    { 0x823D, "suicide" },
    { 0x823E, "closeingamemenu" },
    { 0x823F, "iprintln" },
    { 0x8240, "iprintlnbold" },
    { 0x8241, "spawn" },
    { 0x8242, "setentertime" },
    { 0x8243, "cloneplayer" },
    { 0x8244, "istalking" },
    { 0x8245, "allowspectateteam" },
    { 0x8246, "getguid" },
    { 0x8247, "physicslaunchserver" },
    { 0x8248, "physicslaunchserveritem" },
    { 0x8249, "clonebrushmodeltoscriptmodel" },
    { 0x824A, "scriptmodelplayanim" },
    { 0x824B, "scriptmodelclearanim" },
    { 0x824C, "vehicle_teleport" },
    { 0x824D, "attachpath" },
    { 0x824E, "getattachpos" },
    { 0x824F, "startpath" },
    { 0x8250, "setswitchnode" },
    { 0x8251, "setwaitspeed" },
    { 0x8252, "vehicle_finishdamage" },
    { 0x8253, "vehicle_setspeed" },
    { 0x8254, "vehicle_setspeedimmediate" },
    { 0x8255, "vehicle_rotateyaw" },
    { 0x8256, "vehicle_getspeed" },
    { 0x8257, "vehicle_getvelocity" },
    { 0x8258, "vehicle_getbodyvelocity" },
    { 0x8259, "vehicle_getsteering" },
    { 0x825A, "vehicle_getthrottle" },
    { 0x825B, "vehicle_turnengineoff" },
    { 0x825C, "vehicle_turnengineon" },
    { 0x825D, "getgoalspeedmph" },
    { 0x825E, "setacceleration" },
    { 0x825F, "setdeceleration" },
    { 0x8260, "resumespeed" },
    { 0x8261, "setyawspeed" },
    { 0x8262, "setyawspeedbyname" },
    { 0x8263, "setmaxpitchroll" },
    { 0x8264, "setairresistance" },
    { 0x8265, "setturningability" },
    { 0x8266, "getxuid" },
    { 0x8267, "ishost" },
    { 0x8268, "getspectatingplayer" },
    { 0x8269, "predictstreampos" },
    { 0x826A, "updatescores" },
    { 0x826B, "updatedmscores" },
    { 0x826C, "setrank" },
    { 0x826D, "setcardtitle" },
    { 0x826E, "weaponlocknoclearance" },
    { 0x826F, "visionsyncwithplayer" },
    { 0x8270, "showhudsplash" },
    { 0x8271, "setperk" },
    { 0x8272, "hasperk" },
    { 0x8273, "clearperks" },
    { 0x8274, "unsetperk" },
    { 0x8275, "noclip" },
    { 0x8276, "ufo" },
    { 0x8277, "moveto" },
    { 0x8278, "movex" },
    { 0x8279, "movey" },
    { 0x827A, "movez" },
    { 0x827B, "movegravity" },
    { 0x827C, "moveslide" },
    { 0x827D, "stopmoveslide" },
    { 0x827E, "rotateto" },
    { 0x827F, "rotatepitch" },
    { 0x8280, "rotateyaw" },
    { 0x8281, "rotateroll" },
    { 0x8282, "addpitch" },
    { 0x8283, "addyaw" },
    { 0x8284, "addroll" },
    { 0x8285, "vibrate" },
    { 0x8286, "rotatevelocity" },
    { 0x8287, "solid" },
    { 0x8288, "notsolid" },
    { 0x8289, "setcandamage" },
    { 0x828A, "setcanradiusdamage" },
    { 0x828B, "physicslaunchclient" },
    { 0x828C, "setcardicon" },
    { 0x828D, "setcardnameplate" },
    { 0x828E, "setcarddisplayslot" },
    { 0x828F, "kc_regweaponforfxremoval" },
    { 0x8290, "laststandrevive" },
    { 0x8291, "setspectatedefaults" },
    { 0x8292, "getthirdpersoncrosshairoffset" },
    { 0x8293, "disableweaponpickup" },
    { 0x8294, "enableweaponpickup" },
    { 0x8295, "issplitscreenplayer" },
    { 0x8296, "getweaponslistoffhands" },
    { 0x8297, "getweaponslistitems" },
    { 0x8298, "getweaponslistexclusives" },
    { 0x8299, "getweaponslist" },
    { 0x829A, "canplayerplacesentry" },
    { 0x829B, "canplayerplacetank" },
    { 0x829C, "visionsetnakedforplayer" },
    { 0x829D, "visionsetnightforplayer" },
    { 0x829E, "visionsetmissilecamforplayer" },
    { 0x829F, "visionsetthermalforplayer" },
    { 0x82A0, "visionsetpainforplayer" },
    { 0x82A1, "setblurforplayer" },
    { 0x82A2, "getplayerweaponmodel" },
    { 0x82A3, "getplayerknifemodel" },
    { 0x82A4, "updateplayermodelwithweapons" },
    { 0x82A5, "notifyonplayercommand" },
    { 0x82A6, "canmantle" },
    { 0x82A7, "forcemantle" },
    { 0x82A8, "ismantling" },
    { 0x82A9, "playfx" },
    { 0x82AA, "player_recoilscaleon" },
    { 0x82AB, "player_recoilscaleoff" },
    { 0x82AC, "weaponlockstart" },
    { 0x82AD, "weaponlockfinalize" },
    { 0x82AE, "weaponlockfree" },
    { 0x82AF, "weaponlocktargettooclose" },
    { 0x82B0, "issplitscreenplayerprimary" },
    { 0x82B1, "getviewmodel" },
    { 0x82B2, "fragbuttonpressed" },
    { 0x82B3, "secondaryoffhandbuttonpressed" },
    { 0x82B4, "getcurrentweaponclipammo" },
    { 0x82B5, "setvelocity" },
    { 0x82B6, "getplayerviewheight" },
    { 0x82B7, "getnormalizedmovement" },
    { 0x82B8, "setchannelvolumes" },
    { 0x82B9, "deactivatechannelvolumes" },
    { 0x82BA, "playlocalsound" },
    { 0x82BB, "stoplocalsound" },
    { 0x82BC, "setweaponammoclip" },
    { 0x82BD, "setweaponammostock" },
    { 0x82BE, "getweaponammoclip" },
    { 0x82BF, "getweaponammostock" },
    { 0x82C0, "anyammoforweaponmodes" },
    { 0x82C1, "setclientdvar" },
    { 0x82C2, "setclientdvars" },
    { 0x82C3, "allowads" },
    { 0x82C4, "allowjump" },
    { 0x82C5, "allowsprint" },
    { 0x82C6, "setspreadoverride" },
    { 0x82C7, "resetspreadoverride" },
    { 0x82C8, "setaimspreadmovementscale" },
    { 0x82C9, "setactionslot" },
    { 0x82CA, "setviewkickscale" },
    { 0x82CB, "getviewkickscale" },
    { 0x82CC, "getweaponslistall" },
    { 0x82CD, "getweaponslistprimaries" },
    { 0x82CE, "getnormalizedcameramovement" },
    { 0x82CF, "giveweapon" },
    { 0x82D0, "takeweapon" },
    { 0x82D1, "takeallweapons" },
    { 0x82D2, "getcurrentweapon" },
    { 0x82D3, "getcurrentprimaryweapon" },
    { 0x82D4, "getcurrentoffhand" },
    { 0x82D5, "hasweapon" },
    { 0x82D6, "switchtoweapon" },
    { 0x82D7, "switchtoweaponimmediate" },
    { 0x82D8, "switchtooffhand" },
    { 0x82D9, "setoffhandsecondaryclass" },
    { 0x82DA, "getoffhandsecondaryclass" },
    { 0x82DB, "beginlocationselection" },
    { 0x82DC, "endlocationselection" },
    { 0x82DD, "disableweapons" },
    { 0x82DE, "enableweapons" },
    { 0x82DF, "disableoffhandweapons" },
    { 0x82E0, "enableoffhandweapons" },
    { 0x82E1, "disableweaponswitch" },
    { 0x82E2, "enableweaponswitch" },
    { 0x82E3, "openpopupmenu" },
    { 0x82E4, "openpopupmenunomouse" },
    { 0x82E5, "closepopupmenu" },
    { 0x82E6, "openmenu" },
    { 0x82E7, "closemenu" },
//  { 0x82E8, "_meth_82E8" },
    { 0x82E9, "freezecontrols" },
    { 0x82EA, "disableusability" },
    { 0x82EB, "enableusability" },
    { 0x82EC, "setwhizbyspreads" },
    { 0x82ED, "setwhizbyradii" },
    { 0x82EE, "setreverb" },
    { 0x82EF, "deactivatereverb" },
    { 0x82F0, "setvolmod" },
    { 0x82F1, "setchannelvolume" },
    { 0x82F2, "givestartammo" },
    { 0x82F3, "givemaxammo" },
    { 0x82F4, "getfractionstartammo" },
    { 0x82F5, "getfractionmaxammo" },
    { 0x82F6, "isdualwielding" },
    { 0x82F7, "isreloading" },
    { 0x82F8, "isswitchingweapon" },
    { 0x82F9, "setorigin" },
    { 0x82FA, "getvelocity" },
    { 0x82FB, "setplayerangles" },
    { 0x82FC, "getplayerangles" },
    { 0x82FD, "usebuttonpressed" },
    { 0x82FE, "attackbuttonpressed" },
    { 0x82FF, "adsbuttonpressed" },
    { 0x8300, "meleebuttonpressed" },
    { 0x8301, "playerads" },
    { 0x8302, "isonground" },
    { 0x8303, "isusingturret" },
    { 0x8304, "setviewmodel" },
    { 0x8305, "setoffhandprimaryclass" },
    { 0x8306, "getoffhandprimaryclass" },
    { 0x8307, "startac130" },
    { 0x8308, "stopac130" },
    { 0x8309, "enablemousesteer" },
    { 0x830A, "setscriptmoverkillcam" },
    { 0x830B, "setmapnamestring" },
    { 0x830C, "setgametypestring" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 592> file_list
{{
    { 29, "maps/mp/gametypes/_tweakables" },
    { 30, "common_scripts/utility" },
    { 31, "common_scripts/_createfxmenu" },
    { 32, "common_scripts/_fx" },
    { 65, "maps/_utility" },
    { 66, "maps/_mgturret" },
    { 67, "maps/_bcs_location_trigs" },
    { 68, "maps/_anim" },
    { 69, "maps/_gameskill" },
    { 70, "common_scripts/_destructible" },
    { 94, "common_scripts/_destructible_types" },
    { 95, "maps/_vehicle" },
    { 96, "maps/_mg_penetration" },
    { 97, "maps/_rank" },
    { 98, "maps/_hud_util" },
    { 99, "maps/_hud" },
    { 100, "maps/_missions" },
    { 101, "maps/_colors" },
    { 102, "maps/_spawner" },
    { 103, "maps/_audio" },
    { 104, "maps/_audio_stream_manager" },
    { 133, "maps/_audio_dynamic_ambi" },
    { 134, "maps/_audio_reverb" },
    { 135, "maps/_audio_mix_manager" },
    { 136, "maps/_audio_presets_vehicles" },
    { 137, "maps/_specialops" },
    { 138, "maps/_lights" },
    { 139, "maps/_audio_zone_manager" },
    { 176, "maps/_audio_music" },
    { 177, "maps/_audio_whizby" },
    { 178, "maps/_audio_vehicles" },
    { 179, "maps/_specialops_code" },
    { 180, "maps/_specialops_battlechatter" },
    { 181, "maps/_endmission" },
    { 182, "maps/_utility_code" },
    { 183, "maps/_load" },
    { 184, "maps/_quotes" },
    { 185, "maps/_ambient" },
    { 196, "character/character_hero_europe_price_cc" },
    { 224, "maps/_coop" },
    { 225, "common_scripts/_artcommon" },
    { 226, "maps/_arcademode" },
    { 227, "maps/_damagefeedback" },
    { 228, "maps/_laststand" },
    { 229, "maps/_player_stats" },
    { 230, "maps/_art" },
    { 234, "character/character_tank_crew_a" },
    { 235, "character/character_tank_crew_b" },
    { 264, "maps/_noder" },
    { 265, "common_scripts/_painter" },
    { 266, "maps/_createfx" },
    { 267, "maps/_global_fx" },
    { 268, "maps/_detonategrenades" },
    { 269, "maps/_names" },
    { 270, "maps/_autosave" },
    { 271, "maps/_debug" },
    { 272, "maps/_loadout" },
    { 273, "common_scripts/_elevator" },
    { 314, "common_scripts/_pipes" },
    { 315, "common_scripts/_dynamic_world" },
    { 316, "maps/_introscreen" },
    { 317, "maps/_shutter" },
    { 318, "maps/_escalator" },
    { 319, "maps/_friendlyfire" },
    { 320, "maps/_interactive_objects" },
    { 321, "maps/_intelligence" },
    { 322, "maps/_animatedmodels" },
    { 323, "maps/_fx" },
    { 324, "codescripts/character" },
    { 325, "maps/_compass" },
    { 326, "maps/_hiding_door" },
    { 330, "character/character_opforce_henchmen_lmg_a" },
    { 331, "character/character_opforce_henchmen_lmg_b" },
    { 332, "character/character_hero_europe_price_a" },
    { 358, "maps/_drone" },
    { 359, "maps/_patrol" },
    { 360, "maps/_vehicle_aianim" },
    { 361, "maps/_helicopter_ai" },
    { 362, "maps/_helicopter_globals" },
    { 363, "vehicle_scripts/_attack_heli" },
//  maps/vehiclenames??
    { 365, "maps/_treadfx" },
    { 366, "maps/mp/_utility" },
    { 367, "maps/mp/gametypes/_rank" },
    { 368, "maps/mp/gametypes/_persistence" },
    { 375, "character/character_hero_europe_price_aa" },
    { 396, "maps/_dshk_player_rescue" },
    { 400, "maps/mp/gametypes/_gamelogic" },
    { 401, "maps/mp/killstreaks/_killstreaks" },
    { 402, "maps/mp/gametypes/_missions" },
    { 403, "maps/mp/gametypes/_hud_message" },
    { 404, "characters/mp_character_ally_ghillie_desert" },
    { 405, "characters/mp_character_op_ghillie_desert" },
    { 406, "character/mp_character_ally_ghillie_arctic" },
    { 407, "character/mp_character_op_ghillie_arctic" },
    { 408, "character/mp_character_ally_ghillie_urban" },
    { 409, "character/mp_character_op_ghillie_urban" },
    { 410, "character/mp_character_ally_ghillie_forest" },
    { 411, "character/mp_character_op_ghillie_forest" },
    { 412, "character/mp_character_op_ghillie_militia" },
    { 413, "xmodelalias/alias_delta_elite_heads" },
    { 450, "xmodelalias/alias_delta_elite_heads_longsleeves" },
    { 451, "character/mp_character_delta_elite_assault_aa" },
    { 452, "character/mp_character_delta_elite_assault_ab" },
    { 453, "character/mp_character_delta_elite_assault_ba" },
    { 454, "character/mp_character_delta_elite_assault_bb" },
    { 455, "character/mp_character_delta_elite_lmg_a" },
    { 456, "character/mp_character_delta_elite_lmg_b" },
    { 457, "character/mp_character_delta_elite_smg_a" },
    { 458, "character/mp_character_delta_elite_smg_b" },
    { 488, "character/mp_character_delta_elite_shotgun_a" },
    { 489, "character/mp_character_delta_elite_sniper" },
    { 490, "xmodelalias/alias_sas_heads" },
    { 491, "character/mp_character_sas_urban_assault" },
    { 501, "character/character_hero_europe_soap_injured" },
    { 512, "common_scripts/_createfx" },
// ...
    { 528, "maps/so_survival_mp_paris_precache" },
// ...
    { 532, "character/mp_character_sas_urban_lmg" },
    { 533, "character/mp_character_sas_urban_shotgun" },
    { 534, "character/mp_character_sas_urban_smg" },
    { 535, "character/mp_character_sas_urban_sniper" },
    { 536, "character/mp_character_gign_paris_assault" },
    { 537, "character/mp_character_gign_paris_lmg" },
    { 538, "character/mp_character_gign_paris_shotgun" },
    { 539, "character/mp_character_gign_paris_smg" },
    { 540, "character/mp_character_gign_paris_riot" },
    { 541, "xmodelalias/alias_pmc_africa_heads" },
    { 542, "character/mp_character_pmc_africa_assault_a" },
    { 543, "character/mp_character_pmc_africa_assault_aa" },
    { 544, "character/character_mp_ally_juggernaut" },
// ...
    { 545, "character/mp_character_pmc_africa_lmg_a" },
    { 546, "character/mp_character_pmc_africa_lmg_aa" },
    { 547, "character/mp_character_pmc_africa_smg_aa" },
    { 548, "character/mp_character_pmc_africa_shotgun_a" },
// ...
    { 566, "maps/castle_fx" },
    { 567, "maps/castle_precache" },
// ...
    { 591, "character/mp_character_pmc_africa_sniper" },
    { 592, "character/mp_character_opforce_air_assault" },
    { 593, "character/mp_character_opforce_air_lmg" },
    { 594, "character/mp_character_opforce_air_shotgun" },
    { 595, "character/mp_character_opforce_air_smg" },
    { 596, "character/mp_character_opforce_air_sniper" },
    { 597, "character/character_mp_opforce_juggernaut" },
    { 598, "xmodelalias/alias_russian_military_arctic_heads" },
    { 599, "character/mp_character_opforce_snow_assault" },
    { 600, "character/mp_character_opforce_snow_lmg" },
// ...
    { 610, "maps/so_survival_mp_plaza2_precache" },
    { 611, "maps/so_survival_mp_hardhat_precache" },
// ...
    { 621, "character/mp_character_opforce_snow_shotgun" },
    { 622, "character/mp_character_opforce_snow_smg" },
    { 623, "character/mp_character_opforce_snow_sniper" },
    { 624, "character/mp_character_opforce_urban_assault" },
    { 625, "character/mp_character_opforce_urban_lmg" },
    { 626, "character/mp_character_opforce_urban_shotgun" },
    { 627, "character/mp_character_opforce_urban_smg" },
    { 628, "character/mp_character_opforce_urban_sniper" },
    { 629, "character/mp_character_opforce_woods_assault" },
// ...
    { 651, "character/mp_character_opforce_woods_lmg" },
    { 652, "character/mp_character_opforce_woods_shotgun" },
    { 653, "character/mp_character_opforce_woods_smg" },
    { 654, "character/mp_character_opforce_woods_sniper" },
    { 655, "xmodelalias/alias_africa_militia_heads_mp" },
    { 656, "character/mp_character_africa_militia_assault_a" },
//  {
    { 660, "character/mp_character_africa_militia_lmg_b" },
//  {
    { 662, "character/mp_character_africa_militia_shotgun_b" },
//  {
    { 664, "character/mp_character_africa_militia_smg_b" },
//  {
    { 666, "character/mp_character_africa_militia_sniper" },
    { 667, "xmodelalias/alias_henchmen_heads_mp" },
    { 668, "character/mp_character_opforce_hench_assault_a" },
    { 669, "character/mp_character_opforce_hench_assault_b" },
    { 670, "character/mp_character_opforce_hench_assault_c" },
    { 671, "character/mp_character_opforce_hench_assault_d" },
    { 672, "character/mp_character_opforce_hench_lmg_a" },
    { 673, "character/mp_character_opforce_hench_lmg_b" },
    { 674, "character/mp_character_opforce_hench_shgn_a" },
    { 675, "character/mp_character_opforce_hench_shgn_b" },
// ...
    { 690, "character/mp_character_pmc_africa_smg_a" },
// ...
    { 709, "character/mp_character_opforce_hench_smg_a" },
    { 710, "character/mp_character_opforce_hench_smg_b" },
    { 711, "character/mp_character_opforce_hench_sniper" },
    { 712, "maps/mp/gametypes/_hostmigration" },
    { 713, "mptype/mptype_ally_ghillie_desert" },
    { 714, "mptype/mptype_opforce_ghillie_desert" },
    { 715, "mptype/mptype_ally_ghillie_arctic" },
    { 716, "mptype/mptype_opforce_ghillie_arctic" },
    { 717, "mptype/mptype_ally_ghillie_urban" },
    { 718, "mptype/mptype_opforce_ghillie_urban" },
    { 719, "mptype/mptype_ally_ghillie_forest" },
    { 720, "mptype/mptype_opforce_ghillie_forest" },
    { 721, "mptype/mptype_opforce_ghillie_militia" },
    { 722, "mptype/mptype_delta_multicam_assault" },
    { 723, "mptype/mptype_delta_multicam_lmg" },
    { 724, "mptype/mptype_delta_multicam_smg" },
    { 725, "mptype/mptype_delta_multicam_shotgun" },
    { 726, "mptype/mptype_delta_multicam_riot" },
    { 727, "mptype/mptype_ally_juggernaut" },
    { 728, "mptype/mptype_sas_urban_assault" },
    { 729, "mptype/mptype_sas_urban_lmg" },
    { 730, "mptype/mptype_sas_urban_shotgun" },
    { 731, "mptype/mptype_sas_urban_smg" },
    { 732, "mptype/mptype_sas_urban_sniper" },
    { 733, "mptype/mptype_gign_paris_assault" },
    { 734, "mptype/mptype_gign_paris_lmg" },
    { 735, "mptype/mptype_gign_paris_shotgun" },
    { 736, "mptype/mptype_gign_paris_smg" },
    { 737, "maps/so_survival_mp_alpha_precache" },
    { 738, "maps/so_survival_mp_carbon_precache" },
    { 739, "maps/so_survival_mp_village_precache" },
    { 740, "maps/so_survival_mp_seatown_precache" },
    { 741, "maps/so_survival_mp_lambeth_precache" },
    { 742, "maps/so_survival_mp_bootleg_precache" },
    { 743, "maps/so_survival_mp_exchange_precache" },
    { 744, "maps/so_survival_mp_mogadishu_precache" },
// ...
    { 750, "maps/so_survival_mp_underground_precache" },
    { 751, "maps/so_survival_mp_interchange_precache" },
// ...
    { 767, "mptype/mptype_gign_paris_sniper" },
    { 768, "mptype/mptype_gign_paris_riot" },
    { 769, "mptype/mptype_pmc_africa_assault" },
    { 770, "mptype/mptype_pmc_africa_lmg" },
    { 771, "mptype/mptype_pmc_africa_smg" },
    { 772, "mptype/mptype_pmc_africa_shotgun" },
    { 773, "mptype/mptype_pmc_africa_sniper" },
    { 774, "mptype/mptype_pmc_africa_riot" },
    { 775, "mptype/mptype_opforce_air_assault" },
    { 776, "mptype/mptype_opforce_air_lmg" },
    { 777, "mptype/mptype_opforce_air_shotgun" },
    { 778, "mptype/mptype_opforce_air_smg" },
    { 779, "mptype/mptype_opforce_air_sniper" },
    { 780, "mptype/mptype_opforce_air_riot" },
    { 781, "mptype/mptype_opforce_juggernaut" },
    { 782, "mptype/mptype_opforce_snow_assault" },
    { 783, "mptype/mptype_opforce_snow_lmg" },
    { 784, "mptype/mptype_opforce_snow_shotgun" },
    { 785, "mptype/mptype_opforce_snow_smg" },
    { 786, "mptype/mptype_opforce_snow_sniper" },
    { 787, "mptype/mptype_opforce_snow_riot" },
    { 788, "mptype/mptype_opforce_urban_assault" },
    { 789, "mptype/mptype_opforce_urban_lmg" },
    { 790, "mptype/mptype_opforce_urban_shotgun" },
    { 791, "mptype/mptype_opforce_urban_smg" },
    { 792, "mptype/mptype_opforce_urban_sniper" },
    { 793, "mptype/mptype_opforce_urban_riot" },
    { 794, "mptype/mptype_opforce_woodland_assault" },
    { 795, "mptype/mptype_opforce_woodland_lmg" },
    { 796, "mptype/mptype_opforce_woodland_shotgun" },
// ...
    { 816, "maps/so_survival_mp_dome_precache" },
    { 817, "maps/so_survival_mp_radar_precache" },
    { 818, "maps/so_survival_mp_bravo_precache" },
    { 819, "mptype/mptype_opforce_woodland_smg" },
    { 820, "mptype/mptype_opforce_woodland_sniper" },
    { 821, "mptype/mptype_opforce_woodland_riot" },
    { 822, "mptype/mptype_opforce_africa_assault" },
    { 823, "mptype/mptype_opforce_africa_lmg" },
    { 824, "mptype/mptype_opforce_africa_shotgun" },
    { 825, "mptype/mptype_opforce_africa_smg" },
    { 826, "mptype/mptype_opforce_africa_sniper" },
    { 827, "mptype/mptype_opforce_africa_riot" },
    { 828, "mptype/mptype_opforce_henchmen_assault" },
    { 829, "mptype/mptype_delta_multicam_sniper" },
    { 830, "mptype/mptype_opforce_henchmen_lmg" },
    { 831, "mptype/mptype_opforce_henchmen_shotgun" },
    { 832, "mptype/mptype_opforce_henchmen_smg" },
    { 833, "mptype/mptype_opforce_henchmen_sniper" },
    { 834, "mptype/mptype_opforce_henchmen_riot" },
    { 835, "maps/mp/gametypes/_weapons" },
    { 836, "maps/mp/_entityheadicons" },
    { 837, "maps/mp/gametypes/_damagefeedback" },
    { 838, "maps/mp/_stinger" },
    { 839, "maps/mp/_flashgrenades" },
    { 840, "maps/mp/_empgrenade" },
    { 841, "maps/mp/gametypes/_class" },
    { 842, "maps/mp/_equipment" },
    { 843, "maps/mp/_javelin" },
    { 844, "maps/mp/gametypes/_shellshock" },
    { 845, "maps/mp/_matchdata" },
    { 846, "maps/mp/killstreaks/_perkstreaks" },
    { 847, "maps/mp/perks/_perkfunctions" },
    { 848, "maps/mp/gametypes/_scrambler" },
    { 849, "maps/mp/gametypes/_portable_radar" },
    { 850, "maps/mp/gametypes/_objpoints" },
    { 851, "maps/mp/gametypes/_hud_util" },
    { 852, "maps/mp/gametypes/_gameobjects" },
// ...
    { 873, "maps/mp/gametypes/_quickmessages" },
    { 874, "maps/mp/gametypes/_playerlogic" },
    { 875, "maps/mp/gametypes/_spectating" },
    { 876, "maps/mp/gametypes/_spawnlogic" },
    { 877, "maps/mp/_events" },
    { 878, "maps/mp/gametypes/_gamescore" },
    { 879, "maps/mp/gametypes/_menus" },
    { 880, "maps/mp/_minefields" },
    { 881, "maps/mp/_radiation" },
    { 882, "maps/mp/_shutter" },
    { 883, "maps/mp/_destructables" },
    { 884, "maps/mp/_audio" },
    { 885, "maps/mp/_art" },
    { 886, "maps/mp/_createfx" },
    { 887, "maps/mp/_global_fx" },
    { 888, "maps/mp/_animatedmodels" },
    { 889, "maps/mp/killstreaks/_helicopter" },
    { 890, "maps/mp/_skill" },
    { 891, "maps/mp/killstreaks/_remoteuav" },
    { 892, "maps/mp/gametypes/_battlechatter_mp" },
    { 893, "maps/mp/gametypes/_deathicons" },
    { 894, "maps/mp/gametypes/_killcam" },
    { 895, "maps/mp/perks/_perks" },
    { 896, "maps/mp/gametypes/_damage" },
    { 897, "maps/mp/_highlights" },
    { 898, "maps/mp/killstreaks/_escortairdrop" },
    { 899, "maps/mp/killstreaks/_juggernaut" },
    { 900, "maps/mp/killstreaks/_autosentry" },
    { 901, "maps/mp/killstreaks/_airdrop" },
    { 902, "maps/mp/gametypes/_hud" },
    { 903, "maps/mp/_load" },
    { 904, "maps/mp/gametypes/_serversettings" },
//  { 905, "905" },
    { 906, "maps/mp/mp_plaza2_precache" },
    { 907, "maps/mp/mp_plaza2_fx" },
    { 908, "maps/mp/mp_carbon_precache" },
    { 909, "maps/mp/mp_carbon_fx" },
//  { 910, "910" },
//  { 911, "911" },
    { 912, "maps/mp/mp_village_precache" },
    { 913, "maps/mp/mp_village_fx" },
    { 914, "maps/mp/mp_seatown_precache" },
    { 915, "maps/mp/mp_seatown_fx" },
    { 916, "maps/mp/mp_lambeth_precache" },
    { 917, "maps/mp/mp_lambeth_fx" },
    { 918, "maps/mp/mp_hardhat_precache" },
    { 919, "maps/mp/mp_hardhat_fx" },
    { 920, "maps/mp/mp_bootleg_precache" },
    { 921, "maps/mp/mp_bootleg_fx" },
    { 922, "maps/mp/mp_exchange_precache" },
    { 923, "maps/mp/mp_exchange_fx" },
    { 924, "maps/mp/mp_mogadishu_precache" },
    { 925, "maps/mp/mp_mogadishu_fx" },
    { 926, "maps/mp/mp_underground_precache" },
    { 927, "maps/mp/mp_underground_fx" },
    { 928, "maps/mp/mp_interchange_precache" },
    { 929, "maps/mp/mp_interchange_fx" },
//  { 930, "930" },
//  { 931, "931" },
//  { 932, "932" },
//  { 933, "933" },
    { 934, "maps/mp/gametypes/_healthoverlay" },
    { 935, "maps/mp/gametypes/_music_and_dialog" },
    { 936, "maps/mp/_awards" },
    { 937, "maps/mp/_areas" },
    { 938, "maps/mp/_defcon" },
    { 939, "maps/mp/_matchevents" },
    { 940, "maps/mp/gametypes/_friendicons" },
    { 941, "maps/mp/gametypes/_scoreboard" },
    { 942, "maps/mp/killstreaks/_harrier" },
    { 943, "maps/mp/gametypes/_callbacksetup" },
    { 944, "maps/mp/killstreaks/_airstrike" },
    { 945, "maps/mp/killstreaks/_emp" },
    { 946, "maps/mp/killstreaks/_uav" },
    { 947, "maps/mp/killstreaks/_ac130" },
    { 948, "maps/mp/killstreaks/_remotemissile" },
    { 949, "maps/mp/killstreaks/_helicopter_flock" },
// ...
    { 967, "maps/mp/mp_dome_precache" },
    { 968, "maps/mp/mp_dome_fx" },
//  { 969, "969" },
//  { 970, "970" },
    { 971, "maps/mp/mp_radar_precache" },
    { 972, "maps/mp/mp_radar_fx" },
    { 973, "maps/mp/mp_paris_precache" },
    { 974, "maps/mp/mp_paris_fx" },
    { 975, "maps/mp/mp_bravo_precache" },
    { 976, "maps/mp/mp_bravo_fx" },
    { 977, "maps/mp/mp_alpha_precache" },
    { 978, "maps/mp/mp_alpha_fx" },
    { 979, "maps/mp/killstreaks/_helicopter_guard" },
    { 980, "maps/mp/killstreaks/_nuke" },
    { 981, "maps/mp/killstreaks/_remotemortar" },
    { 982, "maps/mp/killstreaks/_deployablebox" },
    { 983, "maps/mp/killstreaks/_ims" },
    { 984, "maps/mp/killstreaks/_remoteturret" },
    { 985, "maps/mp/killstreaks/_remotetank" },
    { 986, "maps/mp/gametypes/_playercards" },
    { 987, "maps/mp/gametypes/_globallogic" },
// ...
    { 998, "maps/mp/gametypes/_teams" },
// ...
    { 1216, "vehicle_scripts/_pavelow_noai" },
    { 1217, "maps/mp/_fx" },
    { 1218, "maps/mp/_compass" },
    { 1219, "maps/mp/_menus" },
    { 1220, "maps/mp/_crib" },
    { 1221, "maps/mp/killstreaks/_autoshotgun" },
    { 1222, "maps/mp/killstreaks/_tank" },
    { 1223, "maps/mp/killstreaks/_mobilemortar" },
    { 1224, "maps/mp/killstreaks/_a10" },
    { 1225, "maps/mp/killstreaks/_teamammorefill" },
//  { 1226, "1226" }, // maps/mp/gametypes/_clientids maybe
    { 1227, "maps/mp/gametypes/_dev" },
//  { 1228, "1228" },// maps/mp/gametypes/_globalentities maybe
//  { 1229, "1229" },// maps/mp/gametypes/_hardpoints maybe
    { 1230, "maps/mp/killstreaks/_aamissile" },
    { 1231, "maps/mp/killstreaks/_aastrike" },
//    { 1232, "maps/mp/killstreaks/_ac130" }, // bad name!
// ...
    { 1343, "xmodelalias/alias_gign_heads" },
    { 1344, "codescripts/delete" },
    { 1345, "codescripts/struct" },
// ...
    { 1443, "common_scripts/_destructible_types_anim_airconditioner" },
    { 1444, "maps/animated_models/fence_tarp_107x56_med_01" },
    { 1445, "maps/animated_models/fence_tarp_130x56_med_01" },
    { 1446, "maps/animated_models/fence_tarp_134x56_med_01" },
    { 1447, "maps/animated_models/fence_tarp_134x76_med_01" },
    { 1448, "maps/animated_models/fence_tarp_134x76_med_02" },
    { 1449, "maps/animated_models/fence_tarp_140x56_med_01" },
    { 1450, "maps/animated_models/fence_tarp_151x56_med_01" },
    { 1451, "maps/animated_models/fence_tarp_167x56_med_01" },
    { 1452, "maps/animated_models/foliage_desertbrush_1_sway" },
    { 1453, "maps/animated_models/foliage_pacific_bushtree01" },
    { 1454, "maps/animated_models/foliage_pacific_flowers06_sway" },
    { 1455, "maps/animated_models/oil_pump_jack01" },
    { 1456, "maps/animated_models/oil_pump_jack02" },
    { 1457, "maps/animated_models/windmill_spin_med" },
    { 1458, "maps/animated_models/windsock_large_wind_medium" },
    { 1459, "maps/createart/mp_dome_art" },
    { 1460, "maps/mp/_explosive_barrels" },
    { 1461, "maps/createfx/mp_dome_fx" },
    { 1462, "maps/createart/mp_dome_fog" },
// ...
    { 1568, "vehicle_scripts/_mi17" },
//  { 1636, "" },
//  { 1637, "" },
//  { 1638, "" },
    { 1639, "vehicle_scripts/_gaz_dshk" },
    { 1640, "vehicle_scripts/_hind" },
//  { 1641, "" },
    { 1642, "maps/animated_models/fence_tarp_108x76_med_01" },
//  { 1643, "maps/animated_models/fence_tarp_132x82_med_01" },
    { 1644, "maps/_predator2" },
//  { 1645, "maps/_xm25" },
    { 1646, "vehicle_scripts/_blackhawk_minigun" },
//  { 1647, "vehicle_scripts/_stryker50cal" },
// ------- files chunck end ---------
//  { 1721,
// ...
    { 6358, "maps/createart/so_survival_mp_plaza2_art" },
    { 6359, "maps/createart/so_survival_mp_seatown_fog" },
    { 6360, "maps/createart/so_survival_mp_seatown_art" },
    { 6361, "maps/createart/so_survival_mp_underground_fog" },
    { 6362, "maps/createart/so_survival_mp_underground_art" },
    { 6363, "maps/createart/so_survival_mp_village_fog" },
    { 6364, "maps/createart/so_survival_mp_village_art" },
// ...
//  { 8217,
//  { 8218,
//  { 8219,
// ...
    { 8224, "maps/createart/so_survival_mp_hardhat_fog" },
    { 8225, "maps/createart/so_survival_mp_hardhat_art" },
    { 8226, "maps/createart/so_survival_mp_alpha_fog" },
    { 8227, "maps/createart/so_survival_mp_alpha_art" },
    { 8228, "maps/createart/so_survival_mp_bootleg_fog" },
    { 8229, "maps/createart/so_survival_mp_bootleg_art" },
    { 8230, "maps/createart/so_survival_mp_bravo_fog" },
    { 8231, "maps/createart/so_survival_mp_bravo_art" },
    { 8232, "maps/createart/so_survival_mp_carbon_fog" },
    { 8233, "maps/createart/so_survival_mp_carbon_art" },
    { 8234, "maps/createart/so_survival_mp_exchange_fog" },
    { 8235, "maps/createart/so_survival_mp_exchange_art" },
    { 8236, "maps/createart/so_survival_mp_interchange_fog" },
    { 8237, "maps/createart/so_survival_mp_interchange_art" },
    { 8238, "maps/createart/so_survival_mp_lambeth_fog" },
    { 8239, "maps/createart/so_survival_mp_lambeth_art" },
    { 8240, "maps/createart/so_survival_mp_mogadishu_fog" },
    { 8241, "maps/createart/so_survival_mp_mogadishu_art" },
    { 8242, "maps/createart/so_survival_mp_paris_fog" },
    { 8243, "maps/createart/so_survival_mp_paris_art" },
    { 8244, "maps/createart/so_survival_mp_plaza2_fog" },
// ...
    { 18378, "maps/createart/so_survival_mp_dome_fog" },
    { 18379, "maps/createart/so_survival_mp_dome_art" },
    { 18380, "character/character_so_juggernaut_lite" },
// ...
    { 18649, "common_scripts/_destructible_types_anim_generator" },
    { 18650, "common_scripts/_destructible_types_anim_lockers" },
    { 18651, "maps/animated_models/fence_tarp_124x52_a_med_01" },
    { 18652, "maps/animated_models/fence_tarp_124x52_b_med_01" },
    { 18653, "maps/animated_models/fence_tarp_126x76_a_med_01" },
    { 18654, "maps/animated_models/fence_tarp_126x76_med_01" },
    { 18655, "maps/animated_models/radar_spinning" },
    { 18656, "maps/createart/mp_radar_art" },
    { 18657, "maps/createfx/mp_radar_fx" },
    { 18658, "maps/createart/mp_radar_fog" },
    { 18659, "maps/createart/so_survival_mp_radar_fog" },
    { 18660, "maps/createart/so_survival_mp_radar_art" },
    { 18661, "common_scripts/_destructible_types_anim_chicken" },
    { 18662, "maps/animated_models/hanging_sheet_wind_medium" },
    { 18663, "maps/createart/mp_village_art" },
    { 18664, "maps/createfx/mp_village_fx" },
    { 18665, "maps/createart/mp_village_fog" },
    { 18666, "common_scripts/_destructible_types_anim_security_camera" },
    { 18667, "maps/createart/mp_underground_art" },
//  { 18668, "" },
//  { 18669, "" },
    { 18670, "maps/createfx/mp_underground_fx" },
    { 18671, "maps/createart/mp_underground_fog" },
    { 18672, "vehicle_scripts/_80s_hatch1" },
    { 18673, "common_scripts/_destructible_types_anim_me_fanceil1_spin" },
    { 18674, "maps/animated_models/hanging_apron_wind_medium" },
    { 18675, "maps/animated_models/hanging_longsleeve_wind_medium" },
    { 18676, "maps/animated_models/hanging_shortsleeve_wind_medium" },
    { 18677, "maps/animated_models/seatown_canopy_1section_01_sway" },
    { 18678, "maps/animated_models/seatown_canopy_1section_02_sway" },
    { 18679, "maps/animated_models/seatown_canopy_3section_01_sway" },
    { 18680, "maps/animated_models/seatown_canopy_stand_01_sway" },
    { 18681, "maps/animated_models/seatown_canopy_stand_02_sway" },
    { 18682, "maps/animated_models/seatown_lrg_wiregrp_sway" },
    { 18683, "maps/animated_models/seatown_mid01_wiregrp_sway" },
    { 18684, "maps/animated_models/seatown_wire_flags1_sway" },
    { 18685, "maps/animated_models/seatown_wire_flags2_sway" },
    { 18686, "maps/createart/mp_seatown_art" },
    { 18687, "maps/createfx/mp_seatown_fx" },
    { 18688, "maps/createart/mp_seatown_fog" },
    { 18689, "common_scripts/_destructible_types_anim_wallfan" },
    { 18690, "maps/animated_models/mi24p_hind_plaza_destroy_animated" },
    { 18691, "maps/createart/mp_plaza2_art" },
    { 18692, "maps/createfx/mp_plaza2_fx" },
    { 18693, "maps/createart/mp_plaza2_fog" },
    { 18694, "maps/createart/mp_paris_art" },
    { 18695, "maps/createfx/mp_paris_fx" },
    { 18696, "maps/createart/mp_paris_fog" },
    { 18697, "common_scripts/_destructible_types_anim_motorcycle_01" },
    { 18698, "maps/createart/mp_mogadishu_art" },
    { 18699, "maps/createfx/mp_mogadishu_fx" },
    { 18700, "maps/createart/mp_mogadishu_fog" },
    { 18701, "maps/animated_models/com_roofvent2" },
    { 18702, "maps/animated_models/foliage_tree_river_birch_lg_a" },
    { 18703, "maps/animated_models/foliage_tree_river_birch_xl_a" },
    { 18704, "maps/createart/mp_lambeth_art" },
    { 18705, "maps/createfx/mp_lambeth_fx" },
    { 18706, "maps/createart/mp_lambeth_fog" },
    { 18707, "maps/createart/mp_interchange_art" },
    { 18708, "maps/createfx/mp_interchange_fx" },
    { 18709, "maps/createart/mp_interchange_fog" },
    { 18710, "maps/animated_models/fence_tarp_128x84_med_01" },
    { 18711, "maps/animated_models/fence_tarp_192x50_med_01" },
    { 18712, "maps/animated_models/fence_tarp_192x84_a_med_01" },
    { 18713, "maps/animated_models/fence_tarp_196x146_med_01" },
    { 18714, "maps/animated_models/fence_tarp_196x36_med_01" },
    { 18715, "maps/animated_models/fence_tarp_196x56_med_01" },
    { 18716, "maps/animated_models/fence_tarp_208x42_med_01" },
    { 18717, "maps/animated_models/fence_tarp_352x88_med_01" },
    { 18718, "maps/animated_models/fence_tarp_draping_224x116_01" },
    { 18719, "maps/animated_models/fence_tarp_draping_98x94_med_01" },
    { 18720, "maps/animated_models/fence_tarp_draping_98x94_med_02" },
    { 18721, "maps/animated_models/fence_tarp_rooftop_set_01_med_01" },
    { 18722, "maps/animated_models/hanging_dead_paratrooper01_animated" },
    { 18723, "maps/animated_models/plastic_fence_232x88_med_01" },
    { 18724, "maps/animated_models/plastic_fence_234x88_med_01" },
    { 18725, "maps/animated_models/plastic_fence_256x48_med_01" },
    { 18726, "maps/animated_models/plastic_fence_264x40_med_01" },
    { 18727, "maps/animated_models/plastic_fence_300x88_med_01" },
    { 18728, "maps/animated_models/plastic_fence_400x88_med_01" },
    { 18729, "maps/animated_models/plastic_fence_528x88_med_01" },
    { 18730, "maps/createart/mp_hardhat_art" },
    { 18731, "maps/createfx/mp_hardhat_fx" },
    { 18732, "maps/createart/mp_hardhat_fog" },
    { 18733, "maps/createart/mp_exchange_art" },
    { 18734, "maps/createfx/mp_exchange_fx" },
    { 18735, "maps/createart/mp_exchange_fog" },
    { 18736, "common_scripts/_destructible_types_anim_light_fluo_single" },
    { 18737, "maps/animated_models/fence_tarp_110x64_med_01" },
    { 18738, "maps/animated_models/fence_tarp_124x64_med_01" },
    { 18739, "maps/animated_models/fence_tarp_128x64_med_01" },
    { 18740, "maps/animated_models/fence_tarp_130x76_med_01" },
    { 18741, "maps/animated_models/fence_tarp_130x82_a_med_01" },
    { 18742, "maps/animated_models/fence_tarp_130x82_med_01" },
    { 18743, "maps/animated_models/fence_tarp_132x62_med_01" },
    { 18744, "maps/animated_models/fence_tarp_132x82_a_med_01" },
    { 18745, "maps/animated_models/fence_tarp_140x68_med_01" },
    { 18746, "maps/animated_models/fence_tarp_160x82_med_01" },
    { 18747, "maps/animated_models/fence_tarp_162x64_med_01" },
    { 18748, "maps/animated_models/fence_tarp_40x58_med_01" },
    { 18749, "maps/animated_models/fence_tarp_68x58_med_01" },
    { 18750, "maps/animated_models/fence_tarp_70x82_med_01" },
    { 18751, "maps/animated_models/fence_tarp_80x84_med_01" },
    { 18752, "maps/animated_models/fence_tarp_90x64_med_01" },
    { 18753, "maps/animated_models/fence_tarp_94x64_med_01" },
    { 18754, "maps/createart/mp_carbon_art" },
//  { 18755, "" },
//  { 18756, "" },
//  { 18757, "" },
    { 18758, "maps/createfx/mp_carbon_fx" },
    { 18759, "maps/createart/mp_carbon_fog" },
    { 18760, "maps/animated_models/foliage_cod5_tree_jungle_03" },
    { 18761, "maps/animated_models/foliage_pacific_bushtree01" },
    { 18762, "maps/animated_models/foliage_pacific_palms08" },
    { 18763, "maps/createart/mp_bravo_art" },
    { 18764, "maps/createfx/mp_bravo_fx" },
    { 18765, "maps/createart/mp_bravo_fog" },
    { 18766, "common_scripts/_destructible_types_anim_light_fluo_on" },
    { 18767, "maps/createart/mp_bootleg_art" },
    { 18768, "maps/createfx/mp_bootleg_fx" },
    { 18769, "maps/createart/mp_bootleg_fog" },
    { 18770, "maps/animated_models/alpha_hanging_civs_animated" },
    { 18771, "maps/createart/mp_alpha_art" },
    { 18772, "maps/createfx/mp_alpha_fx" },
    { 18773, "maps/createart/mp_alpha_fog" },
// ...
    { 21505, "vehicle_scripts/_btr80" },
    { 21688, "vehicle_scripts/_uaz_castle" },
// ...
    { 33361, "maps/mp/mp_italy_precache" },
    { 33362, "maps/createart/mp_italy_art" },
    { 33363, "maps/mp/mp_italy_fx" },
    { 33364, "maps/createfx/mp_italy_fx" },
    { 33365, "maps/createart/mp_italy_fog" },
    { 33366, "maps/mp/mp_park_precache" },
    { 33367, "maps/createart/mp_park_art" },
    { 33368, "maps/mp/mp_park_fx" },
    { 33369, "maps/createfx/mp_park_fx" },
    { 33370, "maps/createart/mp_park_fog" },
    { 33371, "maps/mp/mp_overwatch_precache" },
    { 33372, "maps/createart/mp_overwatch_art" },
    { 33373, "maps/mp/mp_overwatch_fx" },
    { 33374, "maps/createfx/mp_overwatch_fx" },
    { 33375, "maps/createart/mp_overwatch_fog" },
    { 33376, "maps/mp/mp_morningwood_precache" },
    { 33377, "maps/createart/mp_morningwood_art" },
    { 33378, "maps/mp/mp_morningwood_fx" },
    { 33379, "maps/createfx/mp_morningwood_fx" },
    { 33380, "maps/createart/mp_morningwood_fog" },
    { 33381, "maps/createart/so_survival_mp_italy_fog" },
    { 33382, "maps/createart/so_survival_mp_italy_art" },
    { 33383, "maps/so_survival_mp_italy_precache" },
    { 33384, "maps/createart/so_survival_mp_park_fog" },
    { 33385, "maps/createart/so_survival_mp_park_art" },
    { 33386, "maps/so_survival_mp_park_precache" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 5534> token_list
{{
    { 1, "pl#" },
    { 17, "teamHasRemoteUAV" }, // was introduced in an IW patch, made up name
    { 18, "carriedRemoteUAV" }, // was introduced in an IW patch, made up name
    { 19, "imsBombSquadModel" }, // was introduced in an IW patch, made up name
    { 20, "zonesCycling" },
    { 54, "_unk_field_ID54" }, // was introduced in an IW patch, used in _nuke.gsc and _rank.gsc
    { 55, "isBuffEquippedOnWeapon" },
    { 56, "onDisconnect" },
    { 369, "gun_prevGuns" },
    { 370, "gun_curGun" },
    { 371, "getNextGun" },
    { 372, "addAttachments" },
    { 373, "gun_attachments" },
    { 416, "logKillsConfirmed" },
    { 417, "logKillsDenied" },
    { 418, "isPainted" },
    { 419, "lightWeightScalar" },
    { 420, "painted" },
    { 421, "hideCarryIconOnGameEnd" },
    { 422, "empTimeRemaining" },
    { 423, "juggRemoveRadarOnGameEnded" }, // was introduced in an IW patch, made up name, no way of knowing its official name
    { 424, "airstrikessfx" },
    { 425, "airstrikeexplosion" }, // an educated name guess based off of opaque strings
    { 426, "keepEMPTimeRemaining" },
    { 427, "nukeEmpTimeRemaining" },
    { 428, "keepNukeEMPTimeRemaining" },
    { 429, "handleToggleZoom" },
    { 430, "remote_mortar_toggleZoom" },
    { 431, "hintPickUp" },
    { 432, "pickup_message_deleted" },
    { 433, "getCustomClassLoc" },
    { 434, "isInUnlockTable" },
    { 435, "getChallengeFilter" },
    { 436, "getChallengeTable" },
    { 437, "getTierFromTable" },
    { 438, "getWeaponFromChallenge" },
    { 439, "getWeaponAttachmentFromChallenge" },
    { 440, "isKillstreakChallenge" },
    { 441, "getKillstreakFromChallenge" },
    { 442, "cac_getCustomClassLoc" },
    { 443, "reInitializeMatchRulesOnMigration" },
    { 444, "initializeMatchRules" },
    { 445, "matchRules_oneShotKill" },
    { 447, "matchRules_randomize" },
    { 448, "reInitializeScoreLimitOnMigration" },
    { 449, "gun_nextGuns" },
    { 464, "grnd_wasSpectator" },
    { 467, "grnd_previousCrateTypes" },
    { 520, "spawningAfterRemoteDeath" },
    { 521, "hideWorldIconOnGameEnd" },
    { 522, "fauxVehicleCount" },
    { 523, "incrementFauxVehicleCount" },
    { 524, "decrementFauxVehicleCount" },
    { 525, "crateTeamModelUpdater" },
    { 950, "getHighestProgressedPlayers" },
    { 951, "gun_progressDisplay" },
    { 1023, "delay_createfx_seconds" },
    { 1027, "lastVecToTarget" },
    { 1028, "markForDetete" },
// ...
    { 1648, "initFX" },
    { 1649, "func" },
    { 1650, "main" },
    { 1651, "CodeCallback_StartGameType" },
    { 1652, "CodeCallback_PlayerConnect" },
    { 1653, "CodeCallback_PlayerDisconnect" },
    { 1654, "CodeCallback_PlayerDamage" },
    { 1655, "CodeCallback_PlayerKilled" },
    { 1656, "CodeCallback_VehicleDamage" },
    { 1657, "CodeCallback_CodeEndGame" },
    { 1658, "CodeCallback_PlayerLastStand" },
    { 1659, "CodeCallback_PlayerMigrated" },
    { 1660, "CodeCallback_HostMigration" },
    { 1661, "InitStructs" },
    { 1662, "CreateStruct" },
    { 1664, "init" },
    { 1665, "precache" },
    { 1666, "spawner" },
    { 1667, "code_classname" }, // entity fields
    { 1668, "classname" },
    { 1669, "origin" },
    { 1670, "model" },
    { 1671, "spawnflags" },
    { 1672, "target" },
    { 1673, "targetname" },
    { 1674, "count" },
    { 1675, "health" },
    { 1676, "dmg" },
    { 1677, "angles" },
    { 1678, "birthtime" },
    { 1679, "script_linkname" },
    { 1680, "slidevelocity" },
    { 1681, "name" },           // client fields
    { 1682, "sessionteam" },
    { 1683, "sessionstate" },
    { 1684, "maxHealth" },
// ...
    { 1705, "bouncefraction" },
// ...
    { 1725, "score" },
    { 1726, "deaths" },
    { 1727, "statusicon" },
    { 1728, "headicon" },
    { 1729, "headiconteam" },
    { 1730, "canclimbladders" },
    { 1731, "kills" },
    { 1732, "assists" },
    { 1733, "hasradar" },
    { 1734, "isradarblocked" },
    { 1735, "radarstrength" },
    { 1736, "radarshowenemydirection" },
    { 1737, "radarmode" },
    { 1738, "forcespectatorclient" },
    { 1739, "killcamentity" },
    { 1740, "archivetime" },
    { 1741, "psoffsettime" },
    { 1742, "pers" },
    { 1743, "x" },
    { 1744, "y" },
    { 1745, "z" },
    { 1746, "fontScale" },
    { 1747, "font" },
    { 1748, "alignx" },
    { 1749, "aligny" },
    { 1750, "horzalign" },
    { 1751, "vertalign" },
    { 1752, "color" },
    { 1753, "alpha" },
    { 1754, "label" },
    { 1755, "sort" },
    { 1756, "foreground" },
    { 1757, "lowresbackground" },
    { 1758, "hidewhendead" },
    { 1759, "hidewheninmenu" },
    { 1760, "splatter" },
    { 1761, "glowcolor" },
    { 1762, "glowalpha" },
    { 1763, "archived" },
    { 1764, "hidein3rdperson" },
    { 1765, "hidewhenindemo" },
    { 1766, "veh_speed" },
    { 1767, "veh_pathspeed" },
    { 1768, "veh_transmission" },
    { 1769, "veh_pathdir" },
    { 1770, "veh_pathtype" },
    { 1771, "veh_topspeed" },
    { 1772, "veh_brake" },
    { 1773, "veh_throttle" },
    { 1774, "script_noteworthy" },
    { 1775, "speed" },
    { 1776, "lookahead" },
    { 1777, "ambienttrack" },
    { 1778, "ambienttrack_ac130" },
    { 1779, "message" },
    { 1780, "northyaw" },
    { 1781, "wait" },
    { 1782, "radius" },
    { 1783, "height" },
    { 1784, "accumulate" },
    { 1785, "threshold" },
    { 1786, "cursorhint" },
    { 1787, "hintstring" },
    { 1788, "script_delay" },
    { 1789, "script_visionset" },
    { 1790, "script_zone" },
    { 1791, "rightarc" },
    { 1792, "leftarc" },
    { 1793, "toparc" },
    { 1794, "bottomarc" },
    { 1795, "yawconvergencetime" },
    { 1796, "pitchconvergencetime" },
    { 1797, "suppressionTime" },
    { 1798, "maxrange" },
    { 1799, "aiSpread" },
    { 1800, "damage" },
    { 1801, "playerSpread" },
    { 1802, "convergencetime" },
    { 1803, "weaponinfo" },
    { 1804, "vehicletype" },
    { 1805, "type" },
    { 1806, "accuracy" },
    { 1807, "lookforward" },
    { 1808, "lookright" },
    { 1809, "lookup" },
    { 1810, "fovcosine" },
    { 1811, "fovcosinebusy" },
    { 1812, "upaimlimit" },
    { 1813, "downaimlimit" },
    { 1814, "rightaimlimit" },
    { 1815, "leftaimlimit" },
    { 1816, "maxsightdistsqrd" },
    { 1817, "sightlatency" },
    { 1818, "ignoreclosefoliage" },
    { 1819, "followmin" },
    { 1820, "followmax" },
    { 1821, "chainfallback" },
    { 1822, "chainnode" },
    { 1823, "interval" },
    { 1824, "teammovewaittime" },
    { 1825, "damagetaken" },
    { 1826, "damagedir" },
    { 1827, "damageyaw" },
    { 1828, "damagelocation" },
    { 1829, "damageweapon" },
    { 1830, "damagemod" },
    { 1831, "proneok" },
    { 1832, "walkdistfacingmotion" },
    { 1833, "killcamentitylookat" },
    { 1834, "walkdist" },
    { 1835, "desiredangle" },
    { 1836, "pacifist" },
    { 1837, "pacifistwait" },
    { 1838, "footstepdetectdist" },
    { 1839, "footstepdetectdistwalk" },
    { 1840, "footstepdetectdistsprint" },
    { 1841, "reactiontargetpos" },
    { 1842, "newenemyreactiondistsq" },
    { 1843, "ignoreexplosionevents" },
    { 1844, "ignoresuppression" },
    { 1845, "suppressionwait" },
    { 1846, "suppressionduration" },
    { 1847, "suppressionstarttime" },
    { 1848, "suppressionmeter" },
    { 1849, "weapon" },
    { 1850, "dontavoidplayer" },
    { 1851, "grenadeawareness" },
    { 1852, "grenade" },
    { 1853, "grenadeweapon" },
    { 1854, "grenadeammo" },
    { 1855, "favoriteenemy" },
    { 1856, "highlyawareradius" },
    { 1857, "minpaindamage" },
    { 1858, "allowpain" },
    { 1859, "allowdeath" },
    { 1860, "delayeddeath" },
    { 1861, "diequietly" },
    { 1862, "forceragdollimmediate" },
    { 1863, "providecoveringfire" },
    { 1864, "doingambush" },
    { 1865, "combatmode" },
    { 1866, "alertlevel" },
    { 1867, "alertlevelint" },
    { 1868, "useable" },
    { 1869, "ignoretriggers" },
    { 1870, "pushable" },
    { 1871, "script_pushable" },
    { 1872, "dropweapon" },
    { 1873, "drawoncompass" },
    { 1874, "groundtype" },
    { 1875, "anim_pose" },
    { 1876, "goalradius" },
    { 1877, "goalheight" },
    { 1878, "goalpos" },
    { 1879, "nodeoffsetpos" },
    { 1880, "ignoreforfixednodesafecheck" },
    { 1881, "fixednode" },
    { 1882, "fixednodesaferadius" },
    { 1883, "pathgoalpos" },
    { 1884, "pathrandompercent" },
    { 1885, "usechokepoints" },
    { 1886, "stopanimdistsq" },
    { 1887, "lastenemysightpos" },
    { 1888, "pathenemylookahead" },
    { 1889, "pathenemyfightdist" },
    { 1890, "meleeattackdist" },
    { 1891, "movemode" },
    { 1892, "usecombatscriptatcover" },
    { 1893, "safetochangescript" },
    { 1894, "keepclaimednode" },
    { 1895, "keepclaimednodeifvalid" },
    { 1896, "keepnodeduringscriptedanim" },
    { 1897, "dodangerreact" },
    { 1898, "dangerreactduration" },
    { 1899, "nododgemove" },
    { 1900, "leanamount" },
    { 1901, "turnrate" },
    { 1902, "badplaceawareness" },
    { 1903, "damageshield" },
    { 1904, "nogrenadereturnthrow" },
    { 1905, "noattackeraccuracymod" },
    { 1906, "frontshieldanglecos" },
    { 1907, "lookaheaddir" },
    { 1908, "lookaheaddist" },
    { 1909, "lookaheadhitsstairs" },
    { 1910, "velocity" },
    { 1911, "prevanimdelta" },
    { 1912, "exposedduration" },
    { 1913, "requestarrivalnotify" },
    { 1914, "scriptedarrivalent" },
    { 1915, "goingtoruntopos" },
    { 1916, "engagemindist" },
    { 1917, "engageminfalloffdist" },
    { 1918, "engagemaxdist" },
    { 1919, "engagemaxfalloffdist" },
    { 1920, "finalaccuracy" },
    { 1921, "facemotion" },
    { 1921, "gunblockedbywall" },
    { 1923, "relativedir" },
    { 1924, "lockorientation" },
    { 1925, "maxfaceenemydist" },
    { 1926, "stairsstate" },
    { 1927, "script" },
    { 1928, "prevscript" },
    { 1929, "threatbias" },
    { 1930, "syncedmeleetarget" },
    { 1931, "ignoreme" },
    { 1932, "ignoreall" },
    { 1933, "maxvisibledist" },
    { 1934, "surprisedbymedistsq" },
    { 1935, "ignorerandombulletdamage" },
    { 1936, "dodamagetoall" },
    { 1937, "turretinvulnerability" },
    { 1938, "onlygoodnearestnodes" },
    { 1938, "takedamage" },
    { 1940, "playername" },
    { 1941, "deathinvulnerabletime" },
    { 1942, "criticalbulletdamagedist" },
    { 1943, "attackercount" },
    { 1944, "damagemultiplier" },
    { 1945, "laststand" },
    { 1946, "motiontrackerenabled" },
    { 1947, "team" },
    { 1948, "threatbiasgroup" },
    { 1949, "node" },
    { 1950, "prevnode" },
    { 1951, "enemy" },
    { 1952, "lastattacker" },
    { 1953, "attackeraccuracy" },
    { 1954, "width" },
    { 1955, "enable" },
    { 1956, "freecamera" },
    { 1957, "fov" },
    { 1958, "dofnear" },
    { 1959, "doffar" },
    { 1960, "look" },
    { 1961, "up" },
    { 1962, "right" },
    { 1963, "thermalbodymaterial" },
    { 1964, "entity" },
    { 1965, "tag" },
    { 1966, "nearz" },
    { 1967, "blurradius" },
    { 1968, "lod" },
    { 1969, "clipdistance" },
    { 1970, "enableshadows" },
    { 1971, "visionsetnakedduration" },
    { 1972, "visionsetnaked" },
    { 1973, "visionsetnightduration" },
    { 1974, "visionsetnight" },
    { 1975, "visionsetmissilecamduration" },
    { 1976, "visionsetmissilecam" },
    { 1977, "visionsetthermalduration" },
    { 1978, "visionsetthermal" },
    { 1979, "visionsetpainduration" },
    { 1980, "visionsetpain" },
    { 1981, "activevisionset" },
    { 1982, "activevisionsetduration" },
    { 1983, "animscript" },
    { 1984, "anglelerprate" },
    { 1985, "activator" },
    { 1986, "gravity" },
    { 1987, "ambient" },
    { 1988, "diffusefraction" },
    { 1989, "sunlight" },
    { 1990, "suncolor" },
    { 1991, "sundirection" },
    { 1992, "reflection_clear_color" },
    { 1993, "minusedistsq" },
    { 1994, "owner" },
// symbols
    { 1995, "create_lock" },
    { 1996, "exploder_before_load" },
    { 1997, "exploderFunction" },
    { 1998, "exploder_after_load" },
    { 1999, "server_culled_sounds" },
    { 2000, "createFX_enabled" },
    { 2001, "createFXent" },
    { 2002, "set_forward_and_up_vectors" },
    { 2003, "v" },
    { 2004, "print_org" },
    { 2005, "OneShotfx" },
    { 2006, "exploderfx" },
    { 2007, "createExploder" },
    { 2008, "createFXexploders" },
    { 2009, "script_exploder" },
    { 2010, "script_fxid" },
    { 2011, "script_firefx" },
    { 2012, "script_firefxdelay" },
    { 2013, "script_firefxsound" },
    { 2014, "script_sound" },
    { 2015, "script_earthquake" },
    { 2016, "script_damage" },
    { 2017, "script_radius" },
    { 2018, "script_soundalias" },
    { 2019, "script_firefxtimeout" },
    { 2020, "script_repeat" },
    { 2021, "script_delay_min" },
    { 2022, "script_delay_max" },
    { 2023, "script_exploder_group" },
    { 2024, "targetPos" },
    { 2025, "_script_exploders" },
    { 2026, "createfx_showOrigin" },
    { 2027, "loopfx" },
    { 2028, "createLoopEffect" },
    { 2029, "create_looper" },
    { 2030, "_effect" },
    { 2031, "looper" },
    { 2032, "create_loopsound" },
    { 2033, "loop_fx_sound" },
    { 2034, "create_interval_sound" },
    { 2035, "loop_fx_sound_interval" },
    { 2036, "loopfxthread" },
    { 2037, "waitframe" },
    { 2038, "fxStart" },
    { 2039, "timeOut" },
    { 2040, "fxStop" },
    { 2041, "loopfxChangeID" },
    { 2042, "loopfxChangeOrg" },
    { 2043, "loopfxChangeDelay" },
    { 2044, "loopfxDeletion" },
    { 2045, "loopfxStop" },
    { 2046, "loopSound" },
    { 2047, "loopSoundthread" },
    { 2048, "gunfireloopfx" },
    { 2049, "gunfireloopfxthread" },
    { 2050, "gunfireloopfxVec" },
    { 2051, "gunfireloopfxVecthread" },
    { 2052, "fxfireloopmod" },
    { 2053, "setfireloopmod" },
    { 2054, "setup_fx" },
    { 2055, "script_fxcommand" },
    { 2056, "script_fxstart" },
    { 2057, "script_fxstop" },
    { 2058, "burnville_paratrooper_hack" },
    { 2059, "burnville_paratrooper_hack_loop" },
    { 2060, "create_triggerfx" },
    { 2061, "verify_effects_assignment" },
    { 2062, "_missing_FX" },
    { 2063, "verify_effects_assignment_print" },
    { 2064, "OneShotfxthread" },
    { 2065, "menu" },
    { 2066, "create_fx_menu" },
    { 2067, "setmenu" },
    { 2068, "button_is_clicked" },
    { 2069, "createLoopSound" },
    { 2070, "createNewExploder" },
    { 2071, "createIntervalSound" },
    { 2072, "last_displayed_ent" },
    { 2073, "clear_settable_fx" },
    { 2074, "clear_fx_hudElements" },
    { 2075, "_exit_menu" },
    { 2076, "clear_entity_selection" },
    { 2077, "update_selected_entities" },
    { 2078, "get_last_selected_entity" },
    { 2079, "selected_fx_ents" },
    { 2080, "menu_fx_creation" },
    { 2081, "func_get_level_fx" },
    { 2082, "effect_list_offset" },
    { 2083, "effect_list_offset_max" },
    { 2084, "createOneshotEffect" },
    { 2085, "finish_creating_entity" },
    { 2086, "post_entity_creation_function" },
    { 2087, "select_last_entity" },
    { 2088, "move_selection_to_cursor" },
    { 2089, "menu_init" },
    { 2090, "createFX_options" },
    { 2091, "mp_createfx" },
    { 2092, "createfxMasks" },
    { 2093, "get_last_selected_ent" },
    { 2094, "entities_are_selected" },
    { 2095, "menu_change_selected_fx" },
    { 2096, "prepare_option_for_change" },
    { 2097, "createfx_centerprint" },
    { 2098, "createfx_inputlocked" },
    { 2099, "createFxHudElements" },
    { 2100, "menu_fx_option_set" },
    { 2101, "apply_option_to_selected_fx" },
    { 2102, "set_option_index" },
    { 2103, "selected_fx_option_index" },
    { 2104, "get_selected_option" },
    { 2105, "mask" },
    { 2106, "addOption" },
    { 2107, "get_option" },
    { 2108, "display_fx_info" },
    { 2109, "set_fx_hudElement" },
    { 2110, "createFx_hudElements" },
    { 2111, "display_fx_add_options" },
    { 2112, "add_option_to_selected_entities" },
    { 2113, "menuNone" },
    { 2114, "draw_effects_list" },
    { 2115, "increment_list_offset" },
    { 2116, "createEffect" },
    { 2117, "drawn" },
    { 2118, "createFXbyFXID" },
    { 2119, "getLoopEffectDelayDefault" },
    { 2120, "getOneshotEffectDelayDefault" },
    { 2121, "getExploderDelayDefault" },
    { 2122, "getIntervalSoundDelayMinDefault" },
    { 2123, "getIntervalSoundDelayMaxDefault" },
    { 2124, "add_effect" },
    { 2125, "createExploderEx" },
    { 2126, "set_origin_and_angles" },
    { 2127, "createfx_common" },
    { 2128, "flag_init" },
    { 2129, "createFX" },
    { 2130, "createfx_loopcounter" },
    { 2131, "createFxLogic" },
    { 2132, "get_template_level" },
    { 2133, "func_position_player" },
    { 2134, "location" },
    { 2135, "clearTextMarker" },
    { 2136, "selectedMove_up" },
    { 2137, "selectedMove_forward" },
    { 2138, "selectedMove_right" },
    { 2139, "selectedRotate_pitch" },
    { 2140, "selectedRotate_roll" },
    { 2141, "selectedRotate_yaw" },
    { 2142, "selected_fx" },
    { 2143, "createfx_lockedList" },
    { 2144, "createfx_draw_enabled" },
    { 2145, "buttonIsHeld" },
    { 2146, "player" },
    { 2147, "fx_rotating" },
    { 2148, "createfx_selecting" },
    { 2149, "createfxCursor" },
    { 2150, "buttonClick" },
    { 2151, "button_is_kb" },
    { 2152, "fx_highLightedEnt" },
    { 2153, "func_process_fx_rotater" },
    { 2154, "func_position_player_get" },
    { 2155, "copy_angles_of_selected_ents" },
    { 2156, "reset_axis_of_selected_ents" },
    { 2157, "last_selected_entity_has_changed" },
    { 2158, "drop_selection_to_ground" },
    { 2159, "set_off_exploders" },
    { 2160, "exploder" },
    { 2161, "draw_distance" },
    { 2162, "createfx_autosave" },
    { 2163, "flag_waitopen" },
    { 2164, "rotate_over_time" },
    { 2165, "delete_pressed" },
    { 2166, "remove_selected_option" },
    { 2167, "remove_option" },
    { 2168, "delete_selection" },
    { 2169, "insert_effect" },
    { 2170, "show_help" },
    { 2171, "select_all_exploders_of_currently_selected" },
    { 2172, "copy_ents" },
    { 2173, "stored_ents" },
    { 2174, "textAlpha" },
    { 2175, "paste_ents" },
    { 2176, "add_and_select_entity" },
    { 2177, "get_center_of_array" },
    { 2178, "ent_draw_axis" },
    { 2179, "rotation_is_occuring" },
    { 2180, "print_fx_options" },
    { 2181, "createfxDefaults" },
    { 2182, "entity_highlight_disable" },
    { 2183, "entity_highlight_enable" },
    { 2184, "toggle_createfx_drawing" },
    { 2185, "manipulate_createfx_ents" },
    { 2186, "reset_fx_hud_colors" },
    { 2187, "button_is_held" },
    { 2188, "toggle_entity_selection" },
    { 2189, "select_entity" },
    { 2190, "ent_is_highlighted" },
    { 2191, "deselect_entity" },
    { 2192, "index_is_selected" },
    { 2193, "ent_is_selected" },
    { 2194, "draw_axis" },
    { 2195, "fxHudElements" },
    { 2196, "createfx_centerprint_thread" },
    { 2197, "buttonDown" },
    { 2198, "buttonPressed_internal" },
    { 2199, "get_selected_move_vector" },
    { 2200, "process_button_held_and_clicked" },
    { 2201, "locked" },
    { 2202, "kb_locked" },
    { 2203, "add_button" },
    { 2204, "add_kb_button" },
    { 2205, "set_anglemod_move_vector" },
    { 2206, "cfxprintlnStart" },
    { 2207, "fileprint_launcher_start_file" },
    { 2208, "cfxprintln" },
    { 2209, "fileprint_launcher" },
    { 2210, "cfxprintlnEnd" },
    { 2211, "fileprint_launcher_end_file" },
    { 2212, "func_updatefx" },
    { 2213, "hack_start" },
    { 2214, "get_player" },
    { 2215, "createfx_orgranize_array" },
    { 2216, "stop_fx_looper" },
    { 2217, "stop_loopsound" },
    { 2218, "_effect_keys" },
    { 2219, "alphabetize" },
    { 2220, "restart_fx_looper" },
    { 2221, "process_fx_rotater" },
    { 2222, "generate_fx_log" },
    { 2223, "scriptPrintln" },
    { 2224, "debugPrintln" },
    { 2225, "draw_debug_line" },
    { 2226, "waittillend" },
    { 2227, "noself_func" },
    { 2228, "self_func" },
    { 2229, "randomvector" },
    { 2230, "randomvectorrange" },
    { 2231, "angle_dif" },
    { 2232, "sign" },
    { 2233, "track" },
    { 2234, "current_target" },
    { 2235, "get_enemy_team" },
    { 2236, "clear_exception" },
    { 2237, "exception" },
    { 2238, "defaultException" },
    { 2239, "set_exception" },
    { 2240, "set_all_exceptions" },
    { 2241, "cointoss" },
    { 2242, "choose_from_weighted_array" },
    { 2243, "get_cumulative_weights" },
    { 2244, "waittill_string" },
    { 2245, "waittill_multiple" },
    { 2246, "threads" },
    { 2247, "waittill_multiple_ents" },
    { 2248, "waittill_any_return" },
    { 2249, "waittill_any_timeout" },
    { 2250, "_timeout" },
    { 2251, "waittill_any" },
    { 2252, "waittill_any_ents" },
    { 2253, "isFlashed" },
    { 2254, "flashEndTime" },
    { 2255, "flag_exist" },
    { 2256, "flag" },
    { 2257, "init_flags" },
    { 2258, "flags_lock" },
    { 2259, "generic_index" },
    { 2260, "sp_stat_tracking_func" },
    { 2261, "flag_struct" },
    { 2262, "trigger_flags" },
    { 2263, "empty_init_func" },
    { 2264, "issuffix" },
    { 2265, "flag_set" },
    { 2266, "assign_unique_id" },
    { 2267, "unique_id" },
    { 2268, "flag_wait" },
    { 2269, "flag_clear" },
    { 2270, "waittill_either" },
    { 2271, "array_thread" },
    { 2272, "array_call" },
    { 2273, "array_thread4" },
    { 2274, "array_thread5" },
    { 2275, "trigger_on" },
    { 2276, "trigger_on_proc" },
    { 2277, "realOrigin" },
    { 2278, "trigger_off" },
    { 2279, "trigger_off_proc" },
    { 2280, "set_trigger_flag_permissions" },
    { 2281, "update_trigger_based_on_flags" },
    { 2282, "script_flag_true" },
    { 2283, "script_flag_false" },
    { 2284, "trigger_func" },
    { 2285, "create_flags_and_return_tokens" },
    { 2286, "init_trigger_flags" },
    { 2287, "getstruct" },
    { 2288, "struct_class_names" },
    { 2289, "getstructarray" },
    { 2290, "struct_class_init" },
    { 2291, "struct" },
    { 2292, "fileprint_start" },
    { 2293, "fileprint_map_start" },
    { 2294, "fileprint_map_header" },
    { 2295, "fileprint_map_keypairprint" },
    { 2296, "fileprint_map_entity_start" },
    { 2297, "fileprint_map_entity_end" },
    { 2298, "fileprint_radiant_vec" },
    { 2299, "array_remove" },
    { 2300, "array_remove_array" },
    { 2301, "array_removeUndefined" },
    { 2302, "array_levelthread" },
    { 2303, "array_levelcall" },
    { 2304, "add_to_array" },
    { 2305, "flag_assert" },
    { 2306, "flag_wait_either" },
    { 2307, "flag_wait_either_return" },
    { 2308, "flag_wait_any" },
    { 2309, "flag_wait_any_return" },
    { 2310, "flag_wait_all" },
    { 2311, "flag_wait_or_timeout" },
    { 2312, "flag_waitopen_or_timeout" },
    { 2313, "wait_for_flag_or_time_elapses" },
    { 2314, "delayCall" },
    { 2315, "delayCall_proc" },
    { 2316, "noself_delayCall" },
    { 2317, "noself_delayCall_proc" },
    { 2318, "isSP" },
    { 2319, "isSP_TowerDefense" },
    { 2320, "string_starts_with" },
    { 2321, "plot_points" },
    { 2322, "draw_line_for_time" },
    { 2323, "array_combine" },
    { 2324, "flat_angle" },
    { 2325, "flat_origin" },
    { 2326, "draw_arrow_time" },
    { 2327, "get_linked_ents" },
    { 2328, "script_linkto" },
    { 2329, "get_linked_vehicle_nodes" },
    { 2330, "get_linked_ent" },
    { 2331, "get_linked_vehicle_node" },
    { 2332, "get_links" },
    { 2333, "run_thread_on_targetname" },
    { 2334, "getNodeArrayFunction" },
    { 2335, "run_thread_on_noteworthy" },
    { 2336, "draw_arrow" },
    { 2337, "getfx" },
    { 2338, "fxExists" },
    { 2339, "print_csv_asset" },
    { 2340, "csv_lines" },
    { 2341, "fileprint_csv_start" },
    { 2342, "_loadfx" },
    { 2343, "getLastWeapon" },
    { 2344, "saved_lastWeapon" },
    { 2345, "PlayerUnlimitedAmmoThread" },
    { 2346, "isUsabilityEnabled" },
    { 2347, "disabledUsability" },
    { 2348, "_disableUsability" },
    { 2349, "_enableUsability" },
    { 2350, "resetUsability" },
    { 2351, "_disableWeapon" },
    { 2352, "disabledWeapon" },
    { 2353, "_enableWeapon" },
    { 2354, "isWeaponEnabled" },
    { 2355, "_disableWeaponSwitch" },
    { 2356, "disabledWeaponSwitch" },
    { 2357, "_enableWeaponSwitch" },
    { 2358, "isWeaponSwitchEnabled" },
    { 2359, "_disableOffhandWeapons" },
    { 2360, "disabledOffhandWeapons" },
    { 2361, "_enableOffhandWeapons" },
    { 2362, "isOffhandWeaponEnabled" },
    { 2363, "random" },
    { 2364, "spawn_tag_origin" },
    { 2365, "waittill_notify_or_timeout" },
    { 2366, "fileprintlauncher_linecount" },
    { 2367, "launcher_write_clipboard" },
    { 2368, "isDestructible" },
    { 2369, "destructible_type" },
    { 2370, "pauseEffect" },
    { 2371, "activate_individual_exploder" },
    { 2372, "brush_delete" },
    { 2373, "connectPathsFunction" },
    { 2374, "exploded" },
    { 2375, "brush_throw" },
    { 2376, "get_target_ent" },
    { 2377, "getNodeFunction" },
    { 2378, "brush_show" },
    { 2379, "script_modelname" },
    { 2380, "brush_shown" },
    { 2381, "disconnect_paths" },
    { 2382, "disconnectPathsFunction" },
    { 2383, "exploder_earthquake" },
    { 2384, "do_earthquake" },
    { 2385, "earthquake" },
    { 2386, "exploder_rumble" },
    { 2387, "exploder_delay" },
    { 2388, "exploder_damage" },
    { 2389, "effect_loopsound" },
    { 2390, "loopsound_ent" },
    { 2391, "play_loopsound_in_space" },
    { 2392, "sound_effect" },
    { 2393, "effect_soundalias" },
    { 2394, "play_sound_in_space" },
    { 2395, "cannon_effect" },
    { 2396, "exploder_playSound" },
    { 2397, "fire_effect" },
    { 2398, "first_frame" },
    { 2399, "loop_sound_delete" },
    { 2400, "activate_exploder" },
    { 2401, "is_later_in_alphabet" },
    { 2402, "alphabet_compare" },
    { 2403, "play_loop_sound_on_entity" },
    { 2404, "stop_loop_sound_on_entity" },
    { 2405, "delete_on_death" },
    { 2406, "error" },
    { 2407, "create_dvar" },
    { 2408, "void" },
    { 2409, "tag_project" },
    { 2410, "ter_op" },
    { 2411, "lock" },
    { 2412, "max_count" },
    { 2413, "is_locked" },
    { 2414, "unlock_wait" },
    { 2415, "unlock" },
    { 2416, "unlock_thread" },
    { 2417, "template_script" },
    { 2418, "setParent" },
    { 2419, "parent" },
    { 2420, "point" },
    { 2421, "yOffset" },
    { 2422, "xOffset" },
    { 2423, "relativePoint" },
    { 2424, "getParent" },
    { 2425, "addChild" },
    { 2426, "children" },
    { 2427, "index" },
    { 2428, "removeChild" },
    { 2429, "setPoint" },
    { 2430, "uiParent" },
    { 2431, "elemType" },
    { 2432, "setPointBar" },
    { 2433, "bar" },
    { 2434, "padding" },
    { 2435, "frac" },
    { 2436, "updateBar" },
    { 2437, "shader" },
    { 2438, "hidebar" },
    { 2439, "orig_alpha" },
    { 2440, "createFontString" },
    { 2441, "fontHeight" },
    { 2442, "createServerClientFontString" },
    { 2443, "createClientTimer" },
    { 2444, "createServerFontString" },
    { 2445, "createServerTimer" },
    { 2446, "createIcon" },
    { 2447, "createClientIcon" },
    { 2448, "createIcon_Hudelem" },
    { 2449, "createBar" },
    { 2450, "flashFrac" },
    { 2451, "createClientProgressBar" },
    { 2452, "createClientBar" },
    { 2453, "setFlashFrac" },
    { 2454, "fade_over_time" },
    { 2455, "flashThread" },
    { 2456, "destroyElem" },
    { 2457, "setIconShader" },
    { 2458, "setWidth" },
    { 2459, "setHeight" },
    { 2460, "setSize" },
    { 2461, "updateChildren" },
    { 2462, "stance_carry_icon_enamble" },
    { 2463, "stance_carry" },
    { 2464, "console" },
// ...
    { 2556, "delayThread" },
    { 2562, "gameskill" },
    { 2574, "player_damage" },

    { 2577, "players" },
// ...
    { 2609, "stats" },
    { 2623, "isPrimaryWeapon" },
    { 2664, "forward" },
// ...
    { 2700, "attacker" },
    { 2703, "updateOrigin" },
    { 2711, "A" },
    { 2738, "voice" },
    { 2755, "getRank" },
    { 2764, "priority" },
// ...
    { 2810, "sunradiosity" },
    { 2811, "skycolor" },
    { 2812, "skylight" },
    { 2813, "_color" },
    { 2814, "ltOrigin" },
    { 2815, "gndlt" },
    { 2816, "sound_csv_include" },
    { 2817, "csv_include" },
    { 2818, "precache_script" },
//  { 2819, "" },
    { 2820, "maxbounces" },
    { 2821, "radiosityscale" },
//  { 2822, "" },
    { 2823, "def" },
    { 2824, "exponent" },
    { 2825, "fov_inner" },
    { 2826, "fov_outer" },
    { 2827, "__smorigin" },
    { 2828, "__smangles" },
    { 2829, "__smname" },
    { 2830, "__smid" },
//  { 2831, "" },
//  { 2832, "" },
//  { 2833, "" },
//  { 2834, "" },
//  { 2835, "" },
    { 2836, "contrastGain" }, // MAYBE
// ...
    { 2944, "isTeamSpeaking" }, // animscripts/battlechater_ai

    { 2961, "string" },
// ...
    { 3075, "getVectorRightAngle" },
    { 3105, "within_fov" },

    { 3226, "primaryWeapon" },
// ...
    { 3297, "isSniper" },
// ...
    { 3360, "lastCarExplosionTime" },
    { 3361, "lastCarExplosionRange" },
    { 3362, "lastCarExplosionDamageLocation" },
    { 3363, "lastCarExplosionLocation" },

    { 3445, "playDeathSound" },
    { 3460, "makeType" },
    { 3461, "getInfoIndex" },
    { 3463, "destructible_create" },
    { 3464, "destructible_state" },
    { 3465, "destructible_anim" },
    { 3466, "destructible_fx" },
    { 3467, "destructible_explode" },
    { 3469, "prop_ac_prs_enm_con_digger_a" },
    { 3470, "prop_ac_prs_enm_con_dump_truck_a" },
    { 3471, "prop_ac_prs_enm_fuel_tank_a" },
    { 3472, "prop_ac_prs_enm_hanger_a" },
    { 3473, "prop_ac_prs_enm_maz_a" },
    { 3474, "prop_ac_prs_enm_mi26_halo_a" },
    { 3475, "prop_ac_prs_enm_mstas_a" },
    { 3476, "prop_ac_prs_enm_radar_maz_a" },
    { 3477, "prop_ac_prs_enm_s300v_a" },
    { 3484, "prop_ac_prs_enm_truck_a" },
    { 3491, "prop_ac_prs_enm_mobile_crane_a" },
    { 3492, "prop_ac_prs_enm_landing_craft_a" },
    { 3493, "prop_ac_prs_enm_speed_boat_a" },
    { 3494, "prop_ac_prs_prp_satellite_dish_a_dish" },
// ...
    { 3509, "prop_ac_prs_enm_missile_boat_a" },
    { 3510, "toy_glass" },
    { 3511, "destructible_splash_damage_scaler" },
    { 3512, "destructible_sound" },
    { 3513, "destructible_part" },
    { 3514, "toy_dt_mirror" },
    { 3515, "toy_icbm_consolemonitor" },
    { 3516, "toy_tubetv_" },
    { 3517, "toy_tvs_flatscreen" },
    { 3518, "toy_tvs_flatscreen_sturdy" },
    { 3519, "toy_transformer_ratnest01" },
    { 3520, "destructible_loopfx" },
    { 3521, "destructible_loopsound" },
    { 3522, "destructible_healthdrain" },
    { 3523, "toy_transformer_small01" },
    { 3524, "toy_generator" },
    { 3525, "toy_generator_on" },
    { 3526, "toy_oxygen_tank" },
    { 3527, "toy_electricbox2" },
    { 3528, "toy_electricbox4" },
    { 3529, "toy_airconditioner" },
    { 3530, "toy_ceiling_fan" },
    { 3531, "toy_wall_fan" },
    { 3532, "toy_propane_tank02" },
    { 3533, "destructible_physics" },
    { 3534, "toy_propane_tank02_small" },
    { 3535, "toy_copier" },
    { 3536, "toy_firehydrant" },
    { 3537, "toy_parkingmeter" },
    { 3538, "damage_not" },
    { 3539, "toy_mailbox" },
    { 3540, "toy_mailbox2" },
    { 3541, "toy_newspaper_stand_red" },
    { 3542, "toy_newspaper_stand_blue" },
    { 3543, "toy_filecabinet" },
    { 3544, "toy_trashbin_01" },
    { 3545, "toy_trashbin_02" },
    { 3546, "toy_trashbag1" },
    { 3547, "toy_recyclebin_01" },
    { 3548, "toy_trashcan_metal_closed" },
    { 3549, "toy_water_collector" },
    { 3550, "toy_foliage_tree_oak_1" },
    { 3551, "toy_paris_tree_plane_large" },
    { 3552, "toy_usa_gas_station_trash_bin_01" },
    { 3553, "toy_usa_gas_station_trash_bin_02" },
    { 3554, "toy_light_ceiling_round" },
    { 3555, "destructible_lights_out" },
    { 3556, "toy_light_ceiling_fluorescent" },
    { 3557, "toy_light_ceiling_fluorescent_spotlight" },
    { 3558, "destructible_spotlight" },
    { 3559, "toy_light_ceiling_fluorescent_single" },
    { 3560, "toy_light_ceiling_fluorescent_single_spotlight" },
    { 3561, "toy_bookstore_bookstand4_books" },
    { 3562, "toy_locker_double" },
    { 3563, "toy_dubai_fish_sculpture" },
    { 3564, "toy_intro_concrete_chipaway" },
    { 3565, "toy_chicken" },
    { 3566, "toy_hide_with_fx" },
    { 3567, "vehicle_ac130_80s_sedan1" },
    { 3568, "vehicle_bus_destructible" },
    { 3569, "vehicle_80s_sedan1" },
    { 3570, "vehicle_80s_hatch1" },
    { 3571, "vehicle_80s_hatch2" },
    { 3572, "vehicle_80s_wagon1" },
    { 3573, "vehicle_civ_car_a" },
    { 3574, "vehicle_small_hatch" },
    { 3575, "vehicle_london_cab_black" },
    { 3576, "vehicle_pickup" },
    { 3577, "vehicle_hummer" },
    { 3578, "vehicle_gaz" },
    { 3579, "vehicle_gaz_harbor" },
    { 3580, "vehicle_bm21" },
    { 3581, "vehicle_moving_truck" },
    { 3582, "vehicle_subway_cart" },
    { 3583, "create_vehicle_subway_cart_window_single" },
    { 3584, "vehicle_subway_cart_windows" },
    { 3585, "vehicle_subway_cart_windows_small" },
    { 3586, "vehicle_luxurysedan" },
    { 3587, "destructible_car_alarm" },
    { 3588, "vehicle_mig29_landed" },
    { 3589, "vehicle_mack_truck_short" },
    { 3590, "vehicle_semi_truck" },
    { 3591, "vehicle_motorcycle" },
    { 3592, "vehicle_scooter" },
    { 3593, "vehicle_subcompact" },
    { 3594, "vehicle_van" },
    { 3595, "vehicle_uaz_van" },
    { 3596, "vehicle_van_iw5" },
    { 3597, "vehicle_delivery_theme_park_truck_destructible" },
    { 3598, "vehicle_suburban" },
    { 3599, "vehicle_snowmobile" },
    { 3600, "destructible_gaspump" },
    { 3601, "destructible_electrical_transformer_large" },
    { 3602, "get_precached_anim" },
    { 3603, "_destructible_preanims" },
    { 3604, "get_precached_animtree" },
    { 3605, "_destructible_preanimtree" },
    { 3606, "vehicle_coupe" },
    { 3607, "vehicle_mini" },
    { 3608, "vehicle_uk_truck" },
    { 3609, "vehicle_uk_police_estate" },
    { 3610, "vehicle_uaz_winter" },
    { 3611, "vehicle_uaz_fabric" },
    { 3612, "vehicle_uaz_hardtop" },
    { 3613, "vehicle_jeep" },
    { 3614, "vehicle_jeep_dusty" },
    { 3615, "vehicle_uaz_open" },
    { 3616, "vehicle_india_compact_destructible" },
    { 3617, "vehicle_india_rickshaw" },
    { 3618, "vehicle_tuk_tuk" },
    { 3619, "vehicle_india_suv" },
    { 3620, "vehicle_policecar" },
    { 3621, "vehicle_policecar_russia" },
    { 3622, "vehicle_taxi" },
    { 3623, "random_dynamic_attachment" },
    { 3624, "vehicle_taxi_dubai" },
    { 3625, "toy_security_camera" },
    { 3626, "toy_building_collapse_paris_ac130" },
    { 3627, "toy_poison_gas_attack" },
    { 3628, "toy_arcade_machine" },
    { 3629, "toy_pinball_machine" },
    { 3630, "toy_fortune_machine" },
    { 3631, "toy_trashcan_clown" },
    { 3632, "toy_afrShanty1" },
    { 3633, "vehicle_slava_ny_harbor_zonea" },
    { 3634, "rooftop_skylight_destructible" },
    { 3635, "satellite_dish_big_destructible" },
    { 3636, "dest_onestate" },
    { 3637, "dest_pb_planter" },
    { 3638, "berlin_hotel_lights_ceiling1" },
    { 3639, "rus_vx_gas_canister" },
    { 3640, "destructibleSpawnedEntsLimit" },
    { 3641, "destructibleSpawnedEnts" },
    { 3642, "currentCarAlarms" },
    { 3643, "commonStartTime" },
    { 3644, "fast_destructible_explode" },
    { 3645, "warn_about_old_destructible" },
    { 3646, "find_destructibles" },
    { 3647, "setup_destructibles" },
    { 3648, "modeldummyon" },
    { 3649, "destructibleInfo" },
    { 3650, "parts" },
    { 3651, "destructible_parts" },
    { 3652, "modeldummy" },
    { 3653, "add_key_to_destructible" },
    { 3654, "add_keypairs_to_destructible" },
    { 3655, "add_array_to_destructible" },
    { 3656, "destructible_info" },
    { 3657, "precache_destructibles" },
    { 3658, "add_destructible_fx" },
    { 3659, "canDamageDestructible" },
    { 3660, "destructibles" },
    { 3661, "destructible_think" },
    { 3662, "damageOwner" },
    { 3663, "gunner" },
    { 3664, "is_shotgun_damage" },
    { 3665, "enable_ai_shotgun_destructible_damage" },
    { 3666, "getPartAndStateIndex" },
    { 3667, "destructible_update_part" },
    { 3668, "non_player_damage" },
    { 3669, "waiting_for_queue" },
    { 3670, "exploding" },
    { 3671, "loopingSoundStopNotifies" },
    { 3672, "damage_type" },
    { 3673, "destructible_splash_rotatation" },
    { 3674, "destructible_splash_damage" },
    { 3675, "getAllActiveParts" },
    { 3676, "setDistanceOnParts" },
    { 3677, "getLowestPartDistance" },
    { 3678, "isValidSoundCause" },
    { 3679, "isAttackerValid" },
    { 3680, "forceExploding" },
    { 3681, "dontAllowExplode" },
    { 3682, "damageIsFromPlayer" },
    { 3683, "isAIfunc" },
    { 3684, "isValidDamageCause" },
    { 3685, "godmode" },
    { 3686, "script_bulletshield" },
    { 3687, "getDamageType" },
    { 3688, "damage_mirror" },
    { 3689, "add_damage_owner_recorder" },
    { 3690, "car_damage_owner_recorder" },
    { 3691, "loopfx_onTag" },
    { 3692, "health_drain" },
    { 3693, "destructible_badplace_radius_multiplier" },
    { 3694, "destructible_health_drain_amount_multiplier" },
    { 3695, "healthDrain" },
    { 3696, "disable_destructible_bad_places" },
    { 3697, "disableBadPlace" },
    { 3698, "badplace_cylinder_func" },
    { 3699, "badplace_remove" },
    { 3700, "badplace_delete_func" },
    { 3701, "physics_launch" },
    { 3702, "physics_object_remove" },
    { 3703, "explode" },
    { 3704, "destructible_explosion_radius_multiplier" },
    { 3705, "destructible_protection_func" },
    { 3706, "cleanupVars" },
    { 3707, "animsapplied" },
    { 3708, "caralarm" },
    { 3709, "destructible_cleans_up_more" },
    { 3710, "script_noflip" },
    { 3711, "car_alarm_org" },
    { 3712, "set_disable_friendlyfire_value_delayed" },
    { 3713, "friendlyFireDisabledForDestructible" },
    { 3714, "connectTraverses" },
    { 3715, "disconnectTraverses" },
    { 3716, "get_traverse_disconnect_brush" },
    { 3717, "script_destruct_collision" },
    { 3718, "hideapart" },
    { 3719, "showapart" },
    { 3720, "disable_explosion" },
    { 3721, "force_explosion" },
    { 3722, "get_dummy" },
    { 3723, "play_loop_sound_on_destructible" },
    { 3724, "force_stop_sound" },
    { 3725, "notifyDamageAfterFrame" },
    { 3726, "play_sound" },
    { 3727, "toString" },
    { 3728, "do_car_alarm" },
    { 3729, "lastCarAlarmTime" },
    { 3730, "car_alarm_timeout" },
    { 3731, "should_do_car_alarm" },
    { 3732, "do_random_dynamic_attachment" },
    { 3733, "get_closest_with_targetname" },
    { 3734, "player_touching_post_clip" },
    { 3735, "get_player_touching" },
    { 3736, "is_so" },
    { 3737, "destructible_handles_collision_brushes" },
    { 3738, "collision_brush_pre_explosion" },
    { 3739, "collision_brush_post_explosion" },
    { 3740, "func_destructible_crush_player" },
    { 3741, "debug_player_in_post_clip" },
    { 3742, "destructible_get_my_breakable_light" },
    { 3743, "breakable_light" },
    { 3744, "break_nearest_light" },
    { 3745, "debug_radiusdamage_circle" },
    { 3746, "debug_circle_drawlines" },
    { 3747, "debug_line" },
    { 3748, "spotlight_tag_origin_cleanup" },
    { 3749, "spotlight_fizzles_out" },
    { 3750, "destructible_spotlight_think" },
    { 3751, "is_valid_damagetype" },
    { 3752, "destructible_sound_think" },
    { 3753, "destructible_fx_think" },
    { 3754, "destructible_animation_think" },
    { 3755, "no_destructible_animation" },
    { 3756, "clear_anims" },
    { 3757, "init_destroyed_count" },
    { 3758, "destroyedCount" },
    { 3759, "destroyedCountTimeout" },
    { 3760, "init_destroyed_count" },
    { 3761, "add_to_destroyed_count" },
    { 3762, "get_destroyed_count" },
    { 3763, "get_max_destroyed_count" },
    { 3764, "init_destructible_frame_queue" },
    { 3765, "destructibleFrameQueue" },
    { 3766, "add_destructible_to_frame_queue" },
    { 3767, "entNum" },
    { 3768, "destructible" },
    { 3769, "totalDamage" },
    { 3770, "nearDistance" },
    { 3771, "fxCost" },
    { 3772, "handle_destructible_frame_queue" },
    { 3773, "sort_destructible_frame_queue" },
    { 3774, "get_better_destructible" },
    { 3775, "get_part_FX_cost_for_action_state" },
// ...
    { 3791, "hatModel" },

    { 3844, "vehicle" },

    { 3974, "rockets" },

    { 4095, "secondaryWeapon" },
    { 4116, "startPos" },
    { 4145, "winner" },

    { 4538, "scr_sound" },
    { 4570, "play_sound_on_entity" },
    { 4595, "play_sound_on_tag" },
    { 4630, "script_bombmode_original" },

    { 4778, "challenge_targetVal" },
    { 4779, "challenge_rewardVal" },
    { 4780, "getChallengeStatus" },
    { 4781, "challengeData" },
    { 4782, "ch_getProgress" },
    { 4783, "ch_getState" },
    { 4784, "ch_setProgress" },
    { 4785, "ch_setState" },
    { 4786, "ch_getTarget" },
    { 4787, "buildChallengeTableInfo" },
    { 4788, "challengeInfo" },
    { 4790, "challengeSplashNotify" },
    { 4791, "optionalNumber" },
    { 4792, "sound" },
    { 4793, "slot" },
    { 4794, "actionNotify" },
    { 4795, "updateChallenges" },
    { 4797, "giveRankXpAfterWait" },
    { 4799, "processChallenge" },
    { 4800, "initNotifyMessage" },
    { 4802, "notifyTitle" },
    { 4803, "notifyText" },
    { 4804, "notifyText2" },
    { 4805, "notifyIcon" },
    { 4808, "doingSplash" },
    { 4809, "splashQueue" },
    { 4810, "maxRank" },
    { 4811, "rankTable" },
    { 4823, "scoreInfo" },
    { 4824, "xpScale" },
    { 4839, "fontPulseInit" },
    { 4840, "baseFontScale" },
    { 4841, "maxFontScale" },
    { 4842, "inFrames" },
    { 4843, "outFrames" },
    { 4844, "fontPulse" },
    { 4845, "updateRank" },
    { 4847, "updateRankAnnounceHUD" },
    { 4848, "titleText" },
    { 4849, "iconName" },
    { 4850, "duration" },
    { 4851, "textLabel" },
    { 4854, "notifyMessage" },
    { 4855, "stringToFloat" },
    { 4856, "actionNotifyMessage" },
    { 4857, "removeTypeFromQueue" },
    { 4858, "showNotifyMessage" },
    { 4859, "titleLabel" },
    { 4860, "titleIsString" },
    { 4861, "text2Label" },
    { 4862, "resetOnCancel" },
    { 4863, "waitRequireVisibility" },
    { 4864, "canReadText" },
    { 4865, "isFlashbanged" },
    { 4866, "dispatchNotify" },
    { 4867, "registerScoreInfo" },
    { 4868, "getScoreInfoValue" },
    { 4869, "getRankInfoMinXP" },
    { 4870, "getRankInfoXPAmt" },
    { 4871, "getRankInfoMaxXp" },
    { 4872, "getRankInfoFull" },
    { 4873, "getRankInfoIcon" },
    { 4874, "getRankForXp" },
    { 4875, "getRankXP" },

    { 5170, "intensity" },
    { 5176, "start" },
    { 5177, "end" },
    { 5179, "time" },
    { 5182, "fx" },
    { 5192, "delayThread_proc" },
    { 5197, "currentNode" },
    { 5243, "prev" },
    { 5329, "mode" },
    { 5614, "offset" },
    { 5711, "_unk_field_ID5711" }, // was introduced in an IW patch, used in _destructible.gsc
    { 5797, "_unk_field_ID5797" }, // was introduced in an IW patch, used in _destructible.gsc
    { 5801, "inUse" },
    { 5972, "pos" },
    { 5980, "delay" },
// ...
    { 6070, "points" },
    { 6296, "icon" },
    { 6415, "ps3" },
    { 6695, "hud_damagefeedback" },
    { 6702, "updateDamageFeedback" },
    { 6797, "hidden" },
    { 6858, "artStartVisionFileExport" },
    { 6859, "artEndVisionFileExport" },
    { 6860, "artStartFogFileExport" },
    { 6861, "artEndFogFileExport" },
    { 6862, "artCommonfxprintln" },
    { 6863, "setfogsliders" },
    { 6864, "translateFogSlidersToScript" },
    { 6865, "fogexphalfplane" },
    { 6866, "fognearplane" },
    { 6867, "fogcolor" },
    { 6868, "fogmaxopacity" },
    { 6869, "sunFogEnabled" },
    { 6870, "sunFogColor" },
    { 6871, "sunFogDir" },
    { 6872, "sunFogBeginFadeAngle" },
    { 6873, "sunFogEndFadeAngle" },
    { 6874, "sunFogScale" },
    { 6875, "limit" },
    { 6876, "updateFogFromScript" },
    { 6877, "artfxprintlnFog" },
    { 6886, "tweakart" },
    { 6890, "startdist" },
    { 6891, "halfwaydist" },
    { 6892, "red" },
    { 6893, "green" },
    { 6894, "blue" },
    { 6895, "maxopacity" },
    { 6923, "fovslidercheck" },
    { 6930, "create_vision_set_fog" },
    { 6931, "transitionTime" },
    { 6932, "dumpsettings" },
    { 6943, "scale" },
    { 6973, "vision_set_fog_changes" },
// ...
    { 7122, "timeLimitOverride" },
    { 7156, "func_loopfxthread" },
    { 7157, "func_oneshotfxthread" },
    { 7158, "func_create_loopsound" },
    { 7159, "global_FX" },
    { 7162, "global_FX_create" },
    { 7163, "watchGrenadeUsage" },
    { 7164, "c4explodethisframe" },
    { 7165, "c4array" },
    { 7166, "throwingGrenade" },
    { 7177, "beginGrenadeTracking" },
    { 7178, "grenade_earthQuake" },
    { 7180, "beginC4Tracking" },
    { 7181, "watchC4" },
    { 7182, "c4Death" },
    { 7183, "array_remove_nokeys" },
    { 7184, "watchClaymores" },
    { 7187, "claymoreDetonation" },
    { 7190, "deleteOnDeath" },
    { 7191, "watchC4Detonation" },
    { 7192, "watchC4AltDetonation" },
    { 7193, "waitAndDetonate" },
    { 7194, "c4Damage" },
    { 7195, "resetC4ExplodeThisFrame" },
    { 7196, "saydamaged" },
    { 7200, "getDamageableEnts" },
    { 7201, "isPlayer" },
    { 7202, "isADestructable" },
    { 7203, "damageCenter" },
    { 7204, "weaponDamageTracePassed" },
    { 7205, "damageEnt" },
    { 7206, "damageOrigin" },
    { 7207, "callbackPlayerDamage" },
    { 7208, "debugline" },
    { 7209, "onWeaponDamage" },
    { 7210, "watchC4AltDetonate" },
    { 7352, "elevators" },
    { 7353, "elevator_callbutton_link_v" },
    { 7354, "elevator_callbutton_link_h" },
    { 7355, "elevator_update_global_dvars" },
    { 7356, "elevator_accel" },
    { 7357, "elevator_decel" },
    { 7358, "elevator_music" },
    { 7359, "elevator_speed" },
    { 7360, "elevator_innerdoorspeed" },
    { 7361, "elevator_outterdoorspeed" },
    { 7362, "elevator_return" },
    { 7363, "elevator_waittime" },
    { 7364, "elevator_aggressive_call" },
    { 7365, "elevator_debug" },
    { 7366, "elevator_motion_detection" },
    { 7367, "elevator_think" },
    { 7368, "elevator_call" },
    { 7369, "elevator_callbuttons" },
    { 7370, "floor_override" },
    { 7371, "overrider" },
    { 7372, "elevator_fsm" },
    { 7373, "eState" },
    { 7374, "moveto_floor" },
    { 7375, "motion_trigger" },
    { 7376, "elevator_interrupted" },
    { 7377, "monitor_callbutton" },
    { 7378, "e" },
    { 7379, "call_elevator" },
    { 7380, "get_floor" },
    { 7381, "elevator_interrupt" },
    { 7382, "elevator_floor_update" },
    { 7383, "elevator_sound_think" },
    { 7384, "listen_for" },
    { 7385, "position_elevators" },
    { 7386, "elevator_move" },
    { 7387, "close_inner_doors" },
    { 7388, "open_inner_doors" },
    { 7389, "close_outer_doors" },
    { 7390, "open_outer_doors" },
    { 7391, "build_elevators" },
    { 7392, "build_call_buttons" },
    { 7393, "setup_hints" },
    { 7394, "make_discrete_trigger" },
    { 7395, "enabled" },
    { 7396, "discrete_waittill" },
    { 7397, "discrete_enable_triggerwaittill" },
    { 7398, "disable_trigger" },
    { 7399, "disable_trigger_helper" },
    { 7400, "get_outer_doorset" },
    { 7401, "get_outer_doorsets" },
    { 7402, "get_outer_closedpos" },
    { 7403, "get_outer_leftdoor" },
    { 7404, "get_outer_rightdoor" },
    { 7405, "get_outer_leftdoor_openedpos" },
    { 7406, "get_outer_rightdoor_openedpos" },
    { 7407, "get_housing_children" },
    { 7408, "get_housing_mainframe" },
    { 7409, "get_housing_models" },
    { 7410, "get_housing_primarylight" },
    { 7411, "get_housing_musak_model" },
    { 7412, "get_housing_door_trigger" },
    { 7413, "get_housing_inside_trigger" },
    { 7414, "get_housing_closedpos" },
    { 7415, "get_housing_leftdoor" },
    { 7416, "get_housing_rightdoor" },
    { 7417, "get_housing_leftdoor_opened_pos" },
    { 7418, "get_housing_rightdoor_opened_pos" },
    { 7419, "get_curFloor" },
    { 7420, "get_initFloor" },
    { 7421, "waittill_finish_moving" },
    { 7422, "isInbound" },
    { 7423, "isInBoundingSpere" },
    { 7424, "waittill_or_timeout" },
    { 7425, "elevator_get_dvar_int" },
    { 7426, "elevator_get_dvar" },
    { 7427, "_pipe_fx_time" },
    { 7428, "_pipes" },
    { 7429, "num_pipe_fx" },
    { 7430, "pipesetup" },
    { 7431, "pipe_fx_array" },
    { 7432, "B" },
    { 7433, "pipe_wait_loop" },
    { 7434, "pipe_logic" },
    { 7435, "_pipe_methods" },
    { 7436, "pipefx" },
    { 7437, "fx_time" },
    { 7438, "_sound" },
    { 7439, "pipe_damage" },
    { 7440, "_dmg" },
    { 7441, "allow_pipe_damage" },
    { 7442, "pipesDamage" },
    { 7443, "methodsInit" },
    { 7444, "pipe_calc_ballistic" },
    { 7445, "pipe_calc_splash" },
    { 7446, "pipe_calc_nofx" },
    { 7447, "precacheFX" },
    { 7448, "onPlayerConnect" },
    { 7449, "player_init" },
    { 7450, "touchTriggers" },
    { 7451, "ai_init" },
    { 7452, "civilian_jet_flyby" },
    { 7453, "jet_init" },
    { 7454, "jet_parts" },
    { 7455, "jet_flyto" },
    { 7456, "engine_fxs" },
    { 7457, "flash_fxs" },
    { 7458, "jet_engine_fx" },
    { 7459, "jet_flash_fx_red" },
    { 7460, "jet_flash_fx_green" },
    { 7461, "jet_flash_fx_blink" },
    { 7462, "civilianJetFlyBy" },
    { 7463, "old_origin" },
    { 7464, "jet_fly_vec" },
    { 7465, "jet_flight_time" },
    { 7466, "jet_reset" },
    { 7467, "jet_timer" },
    { 7468, "civilianJetFlyBy_timer" },
    { 7469, "airstrikeInProgress" },
    { 7470, "ac130player" },
    { 7471, "chopper" },
    { 7472, "remoteMissileInProgress" },
    { 7473, "getTimeInterval" },
    { 7474, "getWatchedDvar" },
    { 7475, "gameType" },
    { 7476, "overrideWatchDvars" },
    { 7477, "watchDvars" },
    { 7478, "value" },
    { 7479, "jet_flyby" },
    { 7480, "mapCenter" },
    { 7481, "jet_planeSound" },
    { 7482, "playsound_float" },
    { 7483, "playsound_loop_on_ent" },
    { 7484, "targetIsInFront" },
    { 7485, "targetIsClose" },
    { 7486, "vending_machine" },
    { 7487, "vm_normal" },
    { 7488, "vm_launch_from" },
    { 7489, "vm_launch_to" },
    { 7490, "vm_fx_loc" },
    { 7491, "vm_normal_model" },
    { 7492, "vm_damaged_model" },
    { 7493, "vm_soda_model" },
    { 7494, "vm_soda_start_pos" },
    { 7495, "vm_soda_start_angle" },
    { 7496, "vm_soda_stop_pos" },
    { 7497, "vm_soda_stop_angle" },
    { 7498, "soda_array" },
    { 7499, "soda_count" },
    { 7500, "soda_slot" },
    { 7501, "hp" },
    { 7502, "vending_machine_damage_monitor" },
    { 7503, "spawn_soda" },
    { 7504, "soda_can_drop" },
    { 7505, "soda_can_eject" },
    { 7506, "ejected" },
    { 7507, "freefall" },
    { 7508, "metal_detector" },
    { 7509, "alarm_interval" },
    { 7510, "alarm_playing" },
    { 7511, "alarm_annoyance" },
    { 7512, "tolerance" },
    { 7513, "playsound_and_light" },
    { 7514, "annoyance_tracker" },
    { 7515, "waittill_any_or_timeout" },
    { 7516, "metal_detector_weapons" },
    { 7517, "waittill_weapon_placed" },
    { 7518, "weapon_notify_loop" },
    { 7519, "isInBound_single" },
    { 7520, "metal_detector_dmg_monitor" },
    { 7521, "metal_detector_touch_monitor" },
    { 7522, "alarm_validate_damage" },
    { 7523, "creaky_board" },
    { 7524, "do_creak" },
    { 7525, "motion_light" },
    { 7526, "moveTracker" },
    { 7527, "lightsOn" },
    { 7528, "lightRigs" },
    { 7529, "touchList" },
    { 7530, "distMoved" },
    { 7531, "motion_light_timeout" },
    { 7532, "outdoor_motion_dlight" },
    { 7533, "outdoor_motion_light" },
    { 7534, "lightEnt" },
    { 7535, "outdoor_motion_dlight_timeout" },
    { 7536, "dog_bark" },
    { 7537, "trigger_door" },
    { 7538, "doorEnt" },
    { 7539, "doorAngle" },
    { 7540, "baseYaw" },
    { 7541, "doorOpen" },
    { 7542, "doorClose" },
    { 7543, "getDoorSide" },
    { 7544, "use_toggle" },
    { 7545, "bird_startle" },
    { 7546, "photo_copier_init" },
    { 7547, "copier" },
    { 7548, "copy_bar" },
    { 7549, "start_pos" },
    { 7550, "light" },
    { 7551, "end_pos" },
    { 7552, "get_photo_copier" },
    { 7553, "waittill_copier_copies" },
    { 7554, "photo_copier" },
    { 7555, "photo_copier_no_light" },
    { 7556, "reset_copier" },
    { 7557, "photo_copier_copy_bar_goes" },
    { 7558, "photo_copier_light_on" },
    { 7559, "photo_copier_stop" },
    { 7560, "photo_light_flicker" },
    { 7561, "fan_blade_rotate" },
    { 7562, "triggerTouchThink" },
    { 7563, "finished_spawning" },
    { 7564, "playerTouchTriggerThink" },
    { 7565, "guid" },
    { 7566, "moveTrackers" },
    { 7567, "gameEnded" },
    { 7568, "movementTracker" },
    { 7569, "DisablemovementTracker" },
    { 7570, "anythingTouchingTrigger" },
    { 7571, "playerTouchingTrigger" },

    { 7621, "inc" },
    { 7622, "startYaw" },
    { 7623, "windController" },
    { 7629, "shutterWanderLeft" },
    { 7630, "shutterWanderRight" },
    { 7632, "wireWander" },
    { 7644, "breakables_fx" },
    { 7647, "barrelExpSound" },
    { 7649, "barrelHealth" },
    { 7650, "precachemodeltype" },
    { 7651, "barrelExplodingThisFrame" },
    { 7652, "breakables_clip" },
    { 7656, "_breakable_utility_modelarray" },
    { 7657, "_breakable_utility_modelindex" },
    { 7658, "_breakable_utility_maxnum" },
    { 7674, "oil_spill_think" },
    { 7675, "barrel" },
    { 7676, "oilspill" },
    { 7677, "extra" },
    { 7678, "oil_spill_burn_after" },
    { 7679, "oil_spill_burn" },
    { 7680, "oil_spill_burn_section" },
    { 7681, "explodable_barrel_think" },
    { 7682, "explodable_barrel_burn" },
    { 7683, "explodable_barrel_explode" },
    { 7684, "remove" },
    { 7697, "flags" },

    { 7706, "breakable_clip" },
    { 7712, "modelscale" },
    { 7714, "getClosestEnt" },
    { 7724, "item" },
    { 7743, "anim_prop_models" },
    { 7748, "weight" },
    { 7752, "animateModel" },
    { 7755, "script_print_fx" },
    { 7756, "script_playfx" },
    { 7757, "script_playfxontag" },
    { 7758, "GrenadeExplosionfx" },
    { 7759, "soundfx" },
    { 7760, "soundfxDelete" },
    { 7763, "blendDelete" },
    { 7765, "setModelFromArray" },
    { 7766, "precacheModelArray" },
    { 7767, "attachHead" },
    { 7768, "character_head_index" },
    { 7769, "script_char_index" },
    { 7770, "headModel" },
    { 7771, "attachHat" },
    { 7772, "character_hat_index" },
    { 7773, "new" },
    { 7774, "anim_gunHand" },
    { 7775, "PutGunInHand" },
    { 7776, "save" },
    { 7777, "anim_gunInHand" },
    { 7778, "load" },
    { 7779, "get_random_character" },
    { 7780, "script_char_group" },
    { 7781, "character_index_cache" },
    { 7782, "get_least_used_index" },
    { 7783, "initialize_character_group" },
    { 7784, "get_random_weapon" },
    { 7785, "setupMiniMap" },
    { 7787, "_loadStarted" },
    { 7791, "mapSize" },
    { 7809, "createClientFontString_func" },
    { 7810, "HUDsetPoint_func" },
    { 7813, "laserOn_func" },
    { 7814, "laserOff_func" },
    { 7829, "playerHealthRegen" },
    { 7864, "script_prefab_exploder" },
    { 7876, "script_accel" },
    { 7889, "exploder_load" },
    { 7890, "script_chance" },
    { 7900, "script_disconnectpaths" },
    { 7902, "setupExploders" },
    { 7907, "script_ender" },
// ...
    { 8106, "watchWeaponChange" },
    { 8160, "friendlyfire" },
    { 8366, "stance" },
    { 8377, "monitorFlash" },
    { 8440, "onDeath" },
    { 8677, "oldRadius" },
    { 8745, "debugprint" },

    { 8809, "export" },
    { 8813, "deathtime" },
    { 8908, "is_in_array" },
    { 8951, "playerHealth_RegularRegenDelay" },
    { 8954, "healthOverlayCutoff" },
    { 9005, "text" },
    { 9469, "mgTurret" },
    { 9482, "script_team" },
    { 9808, "draw_line" },
    { 9844, "array_remove_index" },
    { 9870, "flashRumbleLoop" },
// ...
    { 10039, "xenon" },
    { 10226, "endOnDeath" },
    { 10234, "dirtEffect" },
    { 10302, "entityHeadIconOffset" },
    { 10304, "entityHeadIcon" },
    { 10338, "script_targetoffset_z" },
    { 10346, "currentWeapon" },
    { 10351, "cobra_missile_models" },
    { 10352, "fire_missile" },
    { 10359, "cosine" },
    { 10372, "attractor" },
    { 10381, "turretType" },
    { 10389, "turrets" },
    { 10396, "script_airspeed" },
    { 10437, "heli_damage_monitor" },
    { 10635, "_unk_field_ID10635" }, // was introduced in an IW patch, used in _destructible.gsc
    { 10637, "_unk_field_ID10637" }, // was introduced in an IW patch, used in _destructible.gsc
    { 10732, "notifyString" },
    { 11038, "init_animatedmodels" },
    { 11039, "animation" },
    { 11082, "strip_suffix" },
    { 11083, "updateBarScale" },
    { 11084, "rateOfChange" },
    { 11085, "lastUpdateTime" },
    { 11086, "createTimer" },
    { 11087, "baseWidth" },
    { 11088, "baseHeight" },
    { 11089, "createServerIcon" },
    { 11090, "createServerBar" },
    { 11091, "getCurrentFraction" },
    { 11092, "createPrimaryProgressBar" },
    { 11093, "primaryProgressBarHeight" },
    { 11094, "primaryProgressBarWidth" },
    { 11095, "primaryProgressBarY" },
    { 11096, "primaryProgressBarX" },
    { 11097, "createPrimaryProgressBarText" },
    { 11098, "primaryProgressBarFontSize" },
    { 11099, "primaryProgressBarTextY" },
    { 11100, "primaryProgressBarTextX" },
    { 11101, "createTeamProgressBar" },
    { 11102, "teamProgressBarHeight" },
    { 11103, "teamProgressBarWidth" },
    { 11104, "teamProgressBarY" },
    { 11105, "createTeamProgressBarText" },
    { 11106, "teamProgressBarFontSize" },
    { 11107, "teamProgressBarTextY" },
    { 11108, "hideElem" },
    { 11109, "showElem" },
    { 11110, "getIconShader" },
    { 11111, "setIconSize" },
    { 11112, "transitionReset" },
    { 11113, "transitionZoomIn" },
    { 11114, "transitionPulseFXIn" },
    { 11115, "transitionSlideIn" },
    { 11116, "transitionSlideOut" },
    { 11117, "transitionZoomOut" },
    { 11118, "transitionFadeIn" },
    { 11119, "maxAlpha" },
    { 11120, "transitionFadeOut" },
    { 11121, "getWeeklyRef" },
    { 11122, "getDailyRef" },
    { 11123, "hud" },
    { 11124, "splitscreen" },
    { 11125, "lowerTextYAlign" },
    { 11126, "lowerTextY" },
    { 11127, "lowerTextFontSize" },
    { 11128, "getHighestScoringPlayer" },
    { 11129, "placement" },
    { 11130, "getLosingPlayers" },
    { 11131, "givePlayerScore" },
    { 11132, "rankingEnabled" },
    { 11133, "hardcoreMode" },
    { 11134, "xpPointsPopup" },
    { 11135, "statAdd" },
    { 11136, "statSetChild" },
    { 11137, "teamBased" },
    { 11138, "checkPlayerScoreLimitSoon" },
    { 11139, "checkScoreLimit" },
    { 11140, "onPlayerScore" },
    { 11141, "objectivePointsMod" },
    { 11142, "_getPlayerScore" },
    { 11143, "_setPlayerScore" },
    { 11144, "giveTeamScoreForObjective" },
    { 11145, "otherTeam" },
    { 11146, "wasWinning" },
    { 11147, "lastStatusTime" },
    { 11148, "getScoreLimit" },
    { 11149, "leaderDialog" },
    { 11150, "getWinningTeam" },
    { 11151, "_setTeamScore" },
    { 11152, "overtimeScoreWinOverride" },
    { 11153, "onScoreLimit" },
    { 11154, "checkTeamScoreLimitSoon" },
    { 11155, "updateTeamScore" },
    { 11156, "isRoundBased" },
    { 11157, "isObjectiveBased" },
    { 11158, "_getTeamScore" },
    { 11159, "sendUpdatedTeamScores" },
    { 11160, "WaitTillSlowProcessAllowed" },
    { 11161, "sendUpdatedDMScores" },
    { 11162, "updatedDMScores" },
    { 11163, "removeDisconnectedPlayerFromPlacement" },
    { 11164, "updatePlacement" },
    { 11165, "connectedPostGame" },
    { 11166, "getBetterPlayer" },
    { 11167, "updateTeamPlacement" },
    { 11168, "initialDMScoreUpdate" },
    { 11169, "processAssist" },
    { 11170, "onXPEvent" },
    { 11171, "incPersStat" },
    { 11172, "getPersStat" },
    { 11173, "incPlayerStat" },
    { 11174, "giveAdrenaline" },
    { 11175, "playerAssist" },
    { 11176, "processShieldAssist" },
    { 11177, "splashNotifyDelayed" },
    { 11178, "getTweakableDVarValue" },
    { 11179, "rules" },
    { 11180, "dVar" },
    { 11181, "gameTweaks" },
    { 11182, "teamTweaks" },
    { 11183, "playerTweaks" },
    { 11184, "classTweaks" },
    { 11185, "weaponTweaks" },
    { 11186, "hardpointTweaks" },
    { 11187, "hudTweaks" },
    { 11188, "getTweakableDVar" },
    { 11189, "getTweakableValue" },
    { 11190, "getTweakableLastValue" },
    { 11191, "lastValue" },
    { 11192, "setTweakableValue" },
    { 11193, "setTweakableLastValue" },
    { 11194, "registerTweakable" },
    { 11195, "clientTweakables" },
    { 11196, "tweakablesInitialized" },
    { 11197, "minefields" },
    { 11198, "minefield_think" },
    { 11199, "minefield_kill" },
    { 11200, "minefield" },
    { 11201, "radiation" },
    { 11202, "numAreas" },
    { 11203, "playerEnterArea" },
    { 11204, "playerLeaveArea" },
    { 11205, "poison" },
    { 11206, "radiationOverlay" },
    { 11207, "soundWatcher" },
    { 11208, "radiationEffect" },
    { 11209, "radiationSound" },
    { 11210, "blackout" },
    { 11211, "doRadiationDamage" },
    { 11212, "fadeinBlackOut" },
    { 11213, "fadeoutBlackOut" },
    { 11214, "destructable_think" },
    { 11215, "script_accumulate" },
    { 11216, "script_threshold" },
    { 11217, "script_destructable_area" },
    { 11218, "destructable_destruct" },
    { 11219, "blockArea" },
    { 11220, "blockEntsInArea" },
    { 11221, "blockedoff" },
    { 11222, "unblockArea" },
    { 11223, "unblockEntsInArea" },
    { 11224, "Callback_HostMigration" },
    { 11225, "hostMigrationReturnedPlayerCount" },
    { 11226, "hostMigrationTimer" },
    { 11227, "UpdateTimerPausedness" },
    { 11228, "updateGameEvents" },
    { 11229, "hostMigrationWait" },
    { 11230, "inGracePeriod" },
    { 11231, "matchStartTimer" },
    { 11232, "hostMigrationWaitForPlayers" },
    { 11233, "hostMigrationTimerThink_Internal" },
    { 11234, "hostMigrationControlsFrozen" },
    { 11235, "isReallyAlive" },
    { 11236, "freezeControlsWrapper" },
    { 11237, "hostMigrationTimerThink" },
    { 11238, "waitTillHostMigrationDone" },
    { 11239, "waitTillHostMigrationDone" },
    { 11240, "waitLongDurationWithHostMigrationPause" },
    { 11241, "waitLongDurationWithGameEndTimeUpdate" },
    { 11242, "teamBalance" },
    { 11243, "maxClients" },
    { 11244, "freeplayers" },
    { 11245, "initScoreBoard" },
    { 11246, "onFreePlayerConnect" },
    { 11247, "onJoinedTeam" },
    { 11248, "onJoinedSpectators" },
    { 11249, "trackPlayedTime" },
    { 11250, "timePlayed" },
    { 11251, "gameFlagWait" },
    { 11252, "updatePlayerTimes" },
    { 11253, "rankedmatch" },
    { 11254, "updatePlayedTime" },
    { 11255, "statAddBuffered" },
    { 11256, "statAddChildBuffered" },
    { 11257, "bufferedChildStatsMax" },
    { 11258, "statAddChildBufferedWithMax" },
    { 11259, "bufferedStatsMax" },
    { 11260, "statAddBufferedWithMax" },
    { 11261, "updateTeamTime" },
    { 11262, "updateTeamBalanceDvar" },
    { 11263, "updateTeamBalance" },
    { 11264, "teamLimit" },
    { 11265, "getTeamBalance" },
    { 11266, "balanceTeams" },
    { 11267, "dont_auto_balance" },
    { 11268, "axis" },
    { 11269, "allies" },
    { 11270, "setGhillieModels" },
    { 11271, "environment" },
    { 11272, "setTeamModels" },
    { 11273, "setPlayerModels" },
    { 11274, "playerModelForWeapon" },
    { 11275, "isJuggernaut" },
    { 11276, "CountPlayers" },
    { 11277, "trackFreePlayedTime" },
    { 11278, "updateFreePlayerTimes" },
    { 11279, "updateFreePlayedTime" },
    { 11280, "getJoinTeamPermissions" },
    { 11281, "onPlayerSpawned" },
    { 11282, "getTeamName" },
    { 11283, "getTeamShortName" },
    { 11284, "getTeamForfeitedString" },
    { 11285, "getTeamEliminatedString" },
    { 11286, "getTeamIcon" },
    { 11287, "getTeamHudIcon" },
    { 11288, "getTeamHeadIcon" },
    { 11289, "getTeamVoicePrefix" },
    { 11290, "getTeamSpawnMusic" },
    { 11291, "getTeamWinMusic" },
    { 11292, "getTeamFlagModel" },
    { 11293, "getTeamFlagCarryModel" },
    { 11294, "getTeamFlagIcon" },
    { 11295, "getTeamFlagFX" },
    { 11296, "getTeamColor" },
    { 11297, "getTeamCrateModel" },
    { 11298, "getTeamDeployModel" },
    { 11299, "initedEntityHeadIcons" },
    { 11300, "setHeadIcon" },
    { 11301, "entityHeadIcons" },
    { 11302, "getPlayerForGuid" },
    { 11303, "destroyOnOwnerDisconnect" },
    { 11304, "destroyIconsOnDeath" },
    { 11305, "keepPositioned" },
    { 11306, "setTeamHeadIcon" },
    { 11307, "entityHeadIconTeam" },
    { 11308, "setPlayerHeadIcon" },
    { 11309, "keepIconPositioned" },
    { 11310, "destroyHeadIconsOnDeath" },
    { 11311, "updateHeadIconOrigin" },
    { 11312, "watchTrophyUsage" },
    { 11313, "trophyArray" },
    { 11314, "maxPerPlayerExplosives" },
    { 11315, "createBombSquadModel" },
    { 11316, "weaponName" },
    { 11317, "trophyRemainingAmmo" },
    { 11318, "ammo" },
    { 11319, "trigger" },
    { 11320, "c4EMPKillstreakWait" },
    { 11321, "trophyUseListener" },
    { 11322, "setSelfUsable" },
    { 11323, "notUsableForJoiningPlayers" },
    { 11324, "givePerk" },
    { 11325, "trophyPlayerSpawnWaiter" },
    { 11326, "trophyDisconnectWaiter" },
    { 11327, "trophyActive" },
    { 11328, "grenades" },
    { 11329, "missiles" },
    { 11330, "disabled" },
    { 11331, "combineArrays" },
    { 11332, "sentry_fire" },
    { 11333, "projectileExplode" },
    { 11334, "empGrenadeExplode" },
    { 11335, "mine_explode" },
    { 11336, "trophyDamage" },
    { 11337, "friendlyFireCheck" },
    { 11338, "iDFLAGS_PENETRATION" },
    { 11339, "wasDamagedFromBulletPenetration" },
    { 11340, "wasDamaged" },
    { 11341, "trophyBreak" },
    { 11342, "startMonitoringFlash" },
    { 11343, "stopMonitoringFlash" },
    { 11344, "usingRemote" },
    { 11345, "stunScaler" },
    { 11346, "applyFlash" },
    { 11347, "flashDuration" },
    { 11348, "flashRumbleDuration" },
    { 11349, "monitorEMPGrenade" },
    { 11350, "empEndTime" },
    { 11351, "_hasPerk" },
    { 11352, "applyEMP" },
    { 11353, "empDuration" },
    { 11354, "empGrenaded" },
    { 11355, "empGrenadeDeathWaiter" },
    { 11356, "checkToTurnOffEmp" },
    { 11357, "teamEMPed" },
    { 11358, "EMPPlayer" },
    { 11359, "empRumbleLoop" },
    { 11360, "isEMPGrenaded" },
    { 11361, "InitStingerUsage" },
    { 11362, "stingerStage" },
    { 11363, "stingerTarget" },
    { 11364, "stingerLockStartTime" },
    { 11365, "stingerLostSightlineTime" },
    { 11366, "stingerTargets" },
    { 11367, "ResetStingerLocking" },
    { 11368, "stingerUseEntered" },
    { 11369, "ResetStingerLockingOnDeath" },
    { 11370, "StillValidStingerLock" },
    { 11371, "ac130" },
    { 11372, "planemodel" },
    { 11373, "LoopStingerLockingFeedback" },
    { 11374, "LoopStingerLockedFeedback" },
    { 11375, "LockSightTest" },
    { 11376, "StingerDebugDraw" },
    { 11377, "SoftSightTest" },
    { 11378, "GetTargetList" },
    { 11379, "harriers" },
    { 11380, "uavmodels" },
    { 11381, "littleBirds" },
    { 11382, "ugvs" },
    { 11383, "StingerUsageLoop" },
    { 11384, "InitJavelinUsage" },
    { 11385, "javelinStage" },
    { 11386, "javelinPoints" },
    { 11387, "javelinNormals" },
    { 11388, "javelinLockMisses" },
    { 11389, "javelinTargetPoint" },
    { 11390, "javelinTargetNormal" },
    { 11391, "javelinLockStartTime" },
    { 11392, "ResetJavelinLocking" },
    { 11393, "javelinUseEntered" },
    { 11394, "currentlyLocking" },
    { 11395, "currentlyLocked" },
    { 11396, "javelinTarget" },
    { 11397, "EyeTraceForward" },
    { 11398, "LockMissesReset" },
    { 11399, "LockMissesIncr" },
    { 11400, "LockMissesPassedThreshold" },
    { 11401, "TargetPointTooClose" },
    { 11402, "LoopLocalSeekSound" },
    { 11403, "TopAttackPasses" },
    { 11404, "JavelinUsageLoop" },
    { 11405, "isEMPed" },
    { 11406, "javelinLostSightlineTime" },
    { 11407, "DebugSightLine" },
    { 11408, "VehicleLockSightTest" },
    { 11409, "javelinLockVehicle" },
    { 11410, "StillValidJavelinLock" },
    { 11411, "shellshockOnDamage" },
    { 11412, "shellShockReduction" },
    { 11413, "isUsingRemote" },
    { 11414, "bloodEffect" },
    { 11415, "c4_earthQuake" },
    { 11416, "barrel_earthQuake" },
    { 11417, "artillery_earthQuake" },
    { 11418, "attachmentGroup" },
    { 11419, "getAttachmentList" },
    { 11420, "scavenger_altmode" },
    { 11421, "scavenger_secondary" },
    { 11422, "getIntProperty" },
    { 11423, "riotShieldXPBullets" },
    { 11424, "weaponList" },
    { 11425, "claymoreDetectionDot" },
    { 11426, "claymoreDetectionMinDist" },
    { 11427, "claymoreDetectionGracePeriod" },
    { 11428, "claymoreDetonateRadius" },
    { 11429, "mineDetectionGracePeriod" },
    { 11430, "mineDetectionRadius" },
    { 11431, "mineDetectionHeight" },
    { 11432, "mineDamageRadius" },
    { 11433, "mineDamageMin" },
    { 11434, "mineDamageMax" },
    { 11435, "mineDamageHalfHeight" },
    { 11436, "mineSelfDestructTime" },
    { 11437, "mine_launch" },
    { 11438, "mine_spin" },
    { 11439, "mine_beacon" },
    { 11440, "delayMineTime" },
    { 11441, "stingerFXid" },
    { 11442, "primary_weapon_array" },
    { 11443, "side_arm_array" },
    { 11444, "grenade_array" },
    { 11445, "missile_array" },
    { 11446, "inventory_array" },
    { 11447, "mines" },
    { 11448, "dumpIt" },
    { 11449, "bombSquadWaiter" },
    { 11450, "bombSquadVisibilityUpdater" },
    { 11451, "hits" },
    { 11452, "hasDoneCombat" },
    { 11453, "currentWeaponAtSpawn" },
    { 11454, "concussionEndTime" },
    { 11455, "trackingWeaponName" },
    { 11456, "trackingWeaponShots" },
    { 11457, "trackingWeaponKills" },
    { 11458, "trackingWeaponHits" },
    { 11459, "trackingWeaponHeadShots" },
    { 11460, "trackingWeaponDeaths" },
    { 11461, "trackRiotShield" },
    { 11462, "lastHitTime" },
    { 11463, "droppedDeathWeapon" },
    { 11464, "tookWeaponFrom" },
    { 11465, "sniperDustWatcher" },
    { 11466, "getWeaponClass" },
    { 11467, "WatchStingerUsage" },
    { 11468, "WatchJavelinUsage" },
    { 11469, "lastDroppableWeapon" },
    { 11470, "hitsThisMag" },
    { 11471, "isCACPrimaryWeapon" },
    { 11472, "bothBarrels" },
    { 11473, "isKillstreakWeapon" },
    { 11474, "changingWeapon" },
    { 11475, "class_num" },
    { 11476, "loadoutPrimaryBuff" },
    { 11477, "cac_getWeapon" },
    { 11478, "_unsetPerk" },
    { 11479, "loadoutSecondaryBuff" },
    { 11480, "watchStartWeaponChange" },
    { 11481, "watchWeaponReload" },
    { 11482, "watchRangerUsage" },
    { 11483, "isHackWeapon" },
    { 11484, "mayDropWeapon" },
    { 11485, "dropWeaponForDeath" },
    { 11486, "blockWeaponDrops" },
    { 11487, "ownersattacker" },
    { 11488, "detachIfAttached" },
    { 11489, "deletePickupAfterAWhile" },
    { 11490, "getItemWeaponName" },
    { 11491, "watchPickup" },
    { 11492, "itemRemoveAmmoFromAltModes" },
    { 11493, "handleScavengerBagPickup" },
    { 11494, "dropScavengerForDeath" },
    { 11495, "getWeaponBasedGrenadeCount" },
    { 11496, "getWeaponBasedSmokeGrenadeCount" },
    { 11497, "getFragGrenadeCount" },
    { 11498, "getSmokeGrenadeCount" },
    { 11499, "setWeaponStat" },
    { 11500, "watchWeaponUsage" },
    { 11501, "statGetBuffered" },
    { 11502, "statSetBuffered" },
    { 11503, "lastStandParams" },
    { 11504, "lastStandStartTime" },
    { 11505, "updateMagShots" },
    { 11506, "checkHitsThisMag" },
    { 11507, "genericChallenge" },
    { 11508, "checkHit" },
    { 11509, "attackerCanDamageItem" },
    { 11510, "gotPullbackNotify" },
    { 11511, "claymorearray" },
    { 11512, "bouncingbettyArray" },
    { 11513, "isCooked" },
    { 11514, "originalOwner" },
    { 11515, "watchSmokeExplode" },
    { 11516, "inPlayerSmokeScreen" },
    { 11517, "waitSmokeTime" },
    { 11518, "AddMissileToSightTraces" },
    { 11519, "missilesForSightTraces" },
    { 11520, "watchMissileUsage" },
    { 11521, "setAltSceneObj" },
    { 11522, "watchSentryUsage" },
    { 11523, "empExplodeWaiter" },
    { 11524, "watchForThrowbacks" },
    { 11525, "threwBack" },
    { 11526, "activated" },
    { 11527, "c4EMPDamage" },
    { 11528, "setClaymoreTeamHeadIcon" },
    { 11529, "equipmentWatchUse" },
    { 11530, "shouldAffectClaymore" },
    { 11531, "c4Activate" },
    { 11532, "deleteC4AndClaymoresOnDisconnect" },
    { 11533, "wasChained" },
    { 11534, "waitTillEnabled" },
    { 11535, "c4DetectionTrigger" },
    { 11536, "detectId" },
    { 11537, "bombSquadIcon" },
    { 11538, "claymoreDetectionTrigger" },
    { 11539, "detectIconWaiter" },
    { 11540, "detectExplosives" },
    { 11541, "bombSquadIds" },
    { 11542, "setupBombSquad" },
    { 11543, "bombSquadIcons" },
    { 11544, "showHeadIcon" },
    { 11545, "get_damageable_player_pos" },
    { 11546, "get_damageable_player" },
    { 11547, "get_damageable_grenade_pos" },
    { 11548, "get_damageable_grenade" },
    { 11549, "get_damageable_sentry" },
    { 11550, "get_damageable_mine" },
    { 11551, "getEMPDamageEnts" },
    { 11552, "debugcircle" },
    { 11553, "isAltModeWeapon" },
    { 11554, "isInventoryWeapon" },
    { 11555, "isRiotShield" },
    { 11556, "isOffhandWeapon" },
    { 11557, "isSideArm" },
    { 11558, "isGrenade" },
    { 11559, "updateSavedLastWeapon" },
    { 11560, "updateWeaponRank" },
    { 11561, "getWeaponRank" },
    { 11562, "isDeathStreakWeapon" },
    { 11563, "clearEMPOnDeath" },
    { 11564, "updateMoveSpeedScale" },
    { 11565, "getBaseWeaponName" },
    { 11566, "weaponSpeed" },
    { 11567, "moveSpeedScaler" },
    { 11568, "stanceRecoilAdjuster" },
    { 11569, "setRecoilScale" },
    { 11570, "buildWeaponData" },
    { 11571, "baseName" },
    { 11572, "assetName" },
    { 11573, "variants" },
    { 11574, "monitorSemtex" },
    { 11575, "isStuck" },
    { 11576, "stuckEnemyEntity" },
    { 11577, "playerCardSplashNotify" },
    { 11578, "splashNotify" },
    { 11579, "turret_monitorUse" },
    { 11580, "turret_playerThread" },
    { 11581, "spawnMine" },
    { 11582, "killCamOffset" },
    { 11583, "killCamEnt" },
    { 11584, "equipmentMines" },
    { 11585, "killstreakMines" },
    { 11586, "mineDamageMonitor" },
    { 11587, "mineProximityTrigger" },
    { 11588, "mineDeleteTrigger" },
    { 11589, "mineSelfDestruct" },
    { 11590, "mineBounce" },
    { 11591, "mineExplode" },
    { 11592, "playSpinnerFX" },
    { 11593, "mineDamageDebug" },
    { 11594, "mineDamageHeightPassed" },
    { 11595, "getStanceCenter" },
    { 11596, "watchMineUsage" },
    { 11598, "mineThrown" },
    { 11599, "mineBeacon" },
    { 11600, "mineBeaconTeamUpdater" },
    { 11601, "MaxLives" },
    { 11602, "MaxNameLength" },
    { 11603, "MaxEvents" },
    { 11604, "MaxKillstreaks" },
    { 11605, "MaxLogClients" },
    { 11606, "MaxNumChallengesPerPlayer" },
    { 11607, "MaxNumAwardsPerPlayer" },
    { 11608, "getMatchDateTime" },
    { 11609, "logKillstreakEvent" },
    { 11610, "clientid" },
    { 11611, "logGameEvent" },
    { 11612, "logKillEvent" },
    { 11613, "logMultiKill" },
    { 11614, "logPlayerLife" },
    { 11615, "spawnPos" },
    { 11616, "wasTI" },
    { 11617, "spawnTime" },
    { 11618, "logPlayerXP" },
    { 11619, "logLoadout" },
    { 11620, "curClass" },
    { 11621, "getClassIndex" },
    { 11622, "cac_getWeaponAttachment" },
    { 11623, "cac_getWeaponAttachmentTwo" },
    { 11624, "cac_getOffhand" },
    { 11625, "cac_getPerk" },
    { 11626, "cac_getDeathstreak" },
    { 11627, "cac_getWeaponBuff" },
    { 11628, "cac_getKillstreak" },
    { 11629, "classTableName" },
    { 11630, "table_getWeapon" },
    { 11631, "table_getWeaponAttachment" },
    { 11632, "table_getOffhand" },
    { 11633, "table_getEquipment" },
    { 11634, "table_getPerk" },
    { 11635, "table_getDeathstreak" },
    { 11636, "table_getWeaponBuff" },
    { 11637, "table_getKillstreak" },
    { 11638, "validateAttachment" },
    { 11639, "logPlayerDeath" },
    { 11640, "isAttachment" },
    { 11641, "logPlayerData" },
    { 11642, "endOfGameSummaryLogger" },
    { 11643, "weaponsUsed" },
    { 11644, "weaponXpEarned" },
    { 11645, "challengesCompleted" },
    { 11646, "doubleBubbleSort" },
    { 11647, "gameEndListener" },
    { 11648, "getNextLifeId" },
    { 11649, "canLogClient" },
    { 11650, "canLogEvent" },
    { 11651, "canLogKillstreak" },
    { 11652, "canLogLife" },
    { 11653, "logWeaponStat" },
    { 11654, "logAttachmentStat" },
    { 11655, "buildBaseWeaponList" },
    { 11656, "logChallenge" },
    { 11657, "logAward" },
    { 11658, "killstreakFuncs" },
    { 11659, "tryUseAllPerks" },
    { 11660, "tryUseBlindEye" },
    { 11661, "tryUsePaint" },
    { 11662, "tryUseAssists" },
    { 11663, "tryUseSteadyAim" },
    { 11664, "tryUseStalker" },
    { 11665, "tryUseExtremeConditioning" },
    { 11666, "tryUseSleightOfHand" },
    { 11667, "tryUseScavenger" },
    { 11668, "tryUseHardline" },
    { 11669, "setStreakCountToNext" },
    { 11670, "tryUseColdBlooded" },
    { 11671, "tryUseQuickdraw" },
    { 11672, "tryUseBlastshield" },
    { 11673, "tryUseSitRep" },
    { 11674, "tryUseIronLungs" },
    { 11675, "tryUseAssassin" },
    { 11676, "tryUseDeadSilence" },
    { 11677, "doPerkFunctions" },
    { 11678, "watchDeath" },
    { 11679, "_unsetExtraPerks" },
    { 11680, "checkForPerkUpgrade" },
    { 11681, "getPerkUpgrade" },
    { 11682, "isPerkStreakOn" },
    { 11683, "streakName" },
    { 11684, "available" },
    { 11685, "setScrambler" },
    { 11686, "_giveWeapon" },
    { 11687, "unsetScrambler" },
    { 11688, "deleteScrambler" },
    { 11689, "inPlayerScrambler" },
    { 11690, "deployedScrambler" },
    { 11691, "scramblers" },
    { 11692, "cleanArray" },
    { 11693, "monitorScramblerUse" },
    { 11694, "scramblerSetup" },
    { 11695, "scramblerWatchOwner" },
    { 11696, "scramblerBeepSounds" },
    { 11697, "scramblerDamageListener" },
    { 11698, "deathEffect" },
    { 11699, "scramblerUseListener" },
    { 11700, "scramblerProximityTracker" },
    { 11701, "scramProxyActive" },
    { 11702, "scramProxyPerk" },
    { 11703, "setPortableRadar" },
    { 11704, "unsetPortableRadar" },
    { 11705, "deletePortableRadar" },
    { 11706, "inPlayerPortableRadar" },
    { 11707, "deployedPortableRadar" },
    { 11708, "monitorPortableRadarUse" },
    { 11709, "portableRadarSetup" },
    { 11710, "portableRadarWatchOwner" },
    { 11711, "portableRadarBeepSounds" },
    { 11712, "portableRadarDamageListener" },
    { 11713, "portableRadarUseListener" },
    { 11714, "portableRadarProximityTracker" },
    { 11715, "perkFuncs" },
    { 11716, "precacheModel" },
    { 11717, "spawnGlow" },
    { 11718, "spawnFire" },
    { 11719, "scriptPerks" },
    { 11720, "perkSetFuncs" },
    { 11721, "perkUnsetFuncs" },
    { 11722, "fauxPerks" },
    { 11723, "setBlastShield" },
    { 11724, "unsetBlastShield" },
    { 11725, "setSiege" },
    { 11726, "unsetSiege" },
    { 11727, "setFreefall" },
    { 11728, "unsetFreefall" },
    { 11729, "setLocalJammer" },
    { 11730, "unsetLocalJammer" },
    { 11731, "setThermal" },
    { 11732, "unsetThermal" },
    { 11733, "setBlackBox" },
    { 11734, "unsetBlackBox" },
    { 11735, "setLightWeight" },
    { 11736, "unsetLightWeight" },
    { 11737, "setSteelNerves" },
    { 11738, "unsetSteelNerves" },
    { 11739, "setDelayMine" },
    { 11740, "unsetDelayMine" },
    { 11741, "setChallenger" },
    { 11742, "unsetChallenger" },
    { 11743, "setSaboteur" },
    { 11744, "unsetSaboteur" },
    { 11745, "setEndGame" },
    { 11746, "unsetEndGame" },
    { 11747, "setRearView" },
    { 11748, "unsetRearView" },
    { 11749, "setAC130" },
    { 11750, "unsetAC130" },
    { 11751, "setSentryMinigun" },
    { 11752, "unsetSentryMinigun" },
    { 11753, "setPredatorMissile" },
    { 11754, "unsetPredatorMissile" },
    { 11755, "setTank" },
    { 11756, "unsetTank" },
    { 11757, "setPrecision_airstrike" },
    { 11758, "unsetPrecision_airstrike" },
    { 11759, "setHelicopterMinigun" },
    { 11760, "unsetHelicopterMinigun" },
    { 11761, "setOneManArmy" },
    { 11762, "unsetOneManArmy" },
    { 11763, "setLittlebirdSupport" },
    { 11764, "unsetLittlebirdSupport" },
    { 11765, "setTacticalInsertion" },
    { 11766, "unsetTacticalInsertion" },
    { 11767, "setSteadyAimPro" },
    { 11768, "unsetSteadyAimPro" },
    { 11769, "setStunResistance" },
    { 11770, "unsetStunResistance" },
    { 11771, "setMarksman" },
    { 11772, "unsetMarksman" },
    { 11773, "setDoubleLoad" },
    { 11774, "unsetDoubleLoad" },
    { 11775, "setSharpFocus" },
    { 11776, "unsetSharpFocus" },
    { 11777, "setHardShell" },
    { 11778, "unsetHardShell" },
    { 11779, "setRegenSpeed" },
    { 11780, "unsetRegenSpeed" },
    { 11781, "setAutoSpot" },
    { 11782, "unsetAutoSpot" },
    { 11783, "setEmpImmune" },
    { 11784, "unsetEmpImmune" },
    { 11785, "setOverkillPro" },
    { 11786, "unsetOverkillPro" },
    { 11787, "setCombatHigh" },
    { 11788, "unsetCombatHigh" },
    { 11789, "setLightArmor" },
    { 11790, "unsetLightArmor" },
    { 11791, "setRevenge" },
    { 11792, "unsetRevenge" },
    { 11793, "setC4Death" },
    { 11794, "unsetC4Death" },
    { 11795, "setFinalStand" },
    { 11796, "unsetFinalStand" },
    { 11797, "setJuiced" },
    { 11798, "unsetJuiced" },
    { 11799, "setCarePackage" },
    { 11800, "unsetCarePackage" },
    { 11801, "setStoppingPower" },
    { 11802, "unsetStoppingPower" },
    { 11803, "setUAV" },
    { 11804, "unsetUAV" },
    { 11805, "precacheShaders" },
    { 11806, "validatePerk" },
    { 11807, "perks" },
    { 11808, "omaClassChanged" },
    { 11809, "playerProximityTracker" },
    { 11810, "proximityActive" },
    { 11811, "drawLine" },
    { 11812, "cac_modified_damage" },
    { 11813, "xpScaler" },
    { 11814, "isBulletDamage" },
    { 11816, "setPainted" },
    { 11817, "bulletDamageMod" },
    { 11818, "armorVestMod" },
    { 11819, "explosiveDamageMod" },
    { 11820, "blastShieldMod" },
    { 11821, "dangerCloseMod" },
    { 11822, "hasLightArmor" },
    { 11823, "damageBlockedTotal" },
    { 11824, "initPerkDvars" },
    { 11825, "armorVestDefMod" },
    { 11826, "armorPiercingMod" },
    { 11827, "riotShieldMod" },
    { 11828, "cac_selector" },
    { 11829, "specialty" },
    { 11830, "gambitUseTracker" },
    { 11831, "giveBlindEyeAfterSpawn" },
    { 11832, "avoidKillstreakOnSpawnTimer" },
    { 11833, "objPointNames" },
    { 11834, "objPoints" },
    { 11835, "objPointSize" },
    { 11836, "objpoint_alpha_default" },
    { 11837, "objPointScale" },
    { 11838, "createTeamObjpoint" },
    { 11839, "isFlashing" },
    { 11840, "isShown" },
    { 11841, "baseAlpha" },
    { 11842, "deleteObjPoint" },
    { 11843, "setOriginByName" },
    { 11844, "getObjPointByName" },
    { 11845, "getObjPointByIndex" },
    { 11846, "startFlashing" },
    { 11847, "stopFlashing" },
    { 11848, "script_gameobjectname" },
    { 11849, "numGametypeReservedObjectives" },
    { 11850, "objIDStart" },
    { 11851, "carryObject" },
    { 11852, "claimTrigger" },
    { 11853, "canPickupObject" },
    { 11854, "killedInUse" },
    { 11855, "onPlayerDisconnect" },
    { 11856, "createCarryObject" },
    { 11857, "curOrigin" },
    { 11858, "ownerTeam" },
    { 11859, "triggerType" },
    { 11860, "baseOrigin" },
    { 11861, "useWeapon" },
    { 11862, "offset3d" },
    { 11863, "baseAngles" },
    { 11864, "visuals" },
    { 11865, "compassIcons" },
    { 11866, "objIDAllies" },
    { 11867, "objIDAxis" },
    { 11868, "objIDPingFriendly" },
    { 11869, "objIDPingEnemy" },
    { 11870, "carrier" },
    { 11871, "isResetting" },
    { 11872, "interactTeam" },
    { 11873, "allowWeapons" },
    { 11874, "worldIcons" },
    { 11875, "carrierVisible" },
    { 11876, "visibleTeam" },
    { 11877, "carryIcon" },
    { 11878, "onDrop" },
    { 11879, "onPickup" },
    { 11880, "onReset" },
    { 11881, "curProgress" },
    { 11882, "useTime" },
    { 11883, "useRate" },
    { 11884, "teamUseTimes" },
    { 11885, "teamUseTexts" },
    { 11886, "numTouching" },
    { 11887, "claimTeam" },
    { 11888, "claimPlayer" },
    { 11889, "lastClaimTeam" },
    { 11890, "lastClaimTime" },
    { 11891, "carryObjectUseThink" },
    { 11892, "carryObjectProxThink" },
    { 11893, "carryObjectProxThinkInstant" },
    { 11894, "carryObjectProxThinkDelayed" },
    { 11895, "onEndUse" },
    { 11896, "onUseUpdate" },
    { 11897, "pickupObjectDelay" },
    { 11898, "setPickedUp" },
    { 11899, "onPickupFailed" },
    { 11900, "updateCarryObjectOrigin" },
    { 11901, "wait_endon" },
    { 11902, "giveObject" },
    { 11903, "returnHome" },
    { 11904, "isHome" },
    { 11905, "setPosition" },
    { 11906, "onPlayerLastStand" },
    { 11907, "setDropped" },
    { 11908, "body" },
    { 11909, "safeOrigin" },
    { 11910, "setCarrier" },
    { 11911, "clearCarrier" },
    { 11912, "pickupTimeout" },
    { 11913, "autoResetTime" },
    { 11914, "takeObject" },
    { 11915, "trackCarrier" },
    { 11916, "manualDropThink" },
    { 11917, "deleteUseObject" },
    { 11918, "_objective_delete" },
    { 11919, "createUseObject" },
    { 11920, "keyObject" },
    { 11921, "onUse" },
    { 11922, "onCantUse" },
    { 11923, "useText" },
    { 11924, "setKeyObject" },
    { 11925, "useObjectUseThink" },
    { 11926, "onBeginUse" },
    { 11927, "cantUseHintThink" },
    { 11928, "getEarliestClaimPlayer" },
    { 11929, "startTime" },
    { 11930, "useObjectProxThink" },
    { 11931, "proxTriggerThink" },
    { 11932, "carryFlag" },
    { 11933, "proxTriggerLOS" },
    { 11934, "requiresLOS" },
    { 11935, "setClaimTeam" },
    { 11936, "getClaimTeam" },
    { 11937, "noUseBar" },
    { 11938, "updateProxBar" },
    { 11939, "proxBar" },
    { 11940, "proxBarText" },
    { 11941, "lastUseRate" },
    { 11942, "lastHostMigrationState" },
    { 11943, "updateUseRate" },
    { 11944, "objectiveScaler" },
    { 11945, "isArena" },
    { 11946, "attachUseModel" },
    { 11947, "attachedUseModel" },
    { 11948, "useHoldThink" },
    { 11949, "lastNonUseWeapon" },
    { 11950, "detachUseModels" },
    { 11951, "takeUseWeapon" },
    { 11952, "useHoldThinkLoop" },
    { 11953, "personalUseBar" },
    { 11954, "updateTrigger" },
    { 11955, "updateWorldIcons" },
    { 11956, "updateWorldIcon" },
    { 11957, "updateTimer" },
    { 11958, "updateCompassIcons" },
    { 11959, "updateCompassIcon" },
    { 11960, "shouldPingObject" },
    { 11961, "getUpdateTeams" },
    { 11962, "shouldShowCompassDueToRadar" },
    { 11963, "updateVisibilityAccordingToRadar" },
    { 11964, "setOwnerTeam" },
    { 11965, "getOwnerTeam" },
    { 11966, "setUseTime" },
    { 11967, "setUseText" },
    { 11968, "setTeamUseTime" },
    { 11969, "setTeamUseText" },
    { 11970, "setUseHintText" },
    { 11971, "allowCarry" },
    { 11972, "allowUse" },
    { 11973, "setVisibleTeam" },
    { 11974, "setModelVisibility" },
    { 11975, "_suicide" },
    { 11976, "makeSolid" },
    { 11977, "setCarrierVisible" },
    { 11978, "setCanUse" },
    { 11979, "useTeam" },
    { 11980, "set2DIcon" },
    { 11981, "set3DIcon" },
    { 11982, "set3DUseIcon" },
    { 11983, "worldUseIcons" },
    { 11984, "setCarryIcon" },
    { 11985, "disableObject" },
    { 11986, "enableObject" },
    { 11987, "getRelativeTeam" },
    { 11988, "isFriendlyTeam" },
    { 11989, "canInteractWith" },
    { 11990, "isTeam" },
    { 11991, "isRelativeTeam" },
    { 11992, "getEnemyTeam" },
    { 11993, "getNextObjID" },
    { 11994, "reclaimedReservedObjectives" },
    { 11995, "getLabel" },
    { 11996, "script_label" },
    { 11997, "autoSpotDeathWatcher" },
    { 11998, "autoSpotAdsWatcher" },
    { 11999, "recoilScale" },
    { 12000, "blastshieldUseTracker" },
    { 12001, "perkUseDeathTracker" },
    { 12002, "_usePerkEnabled" },
    { 12003, "endGame" },
    { 12004, "attackerTable" },
    { 12005, "endGameTimer" },
    { 12006, "endGameIcon" },
    { 12007, "revertVisionSet" },
    { 12008, "nukeDetonated" },
    { 12009, "nukeVisionSet" },
    { 12010, "endGameDeath" },
    { 12011, "trackSiegeEnable" },
    { 12012, "trackSiegeDissable" },
    { 12013, "stanceStateListener" },
    { 12014, "jumpStateListener" },
    { 12015, "killStreakScaler" },
    { 12016, "setBackShield" },
    { 12017, "unsetBackShield" },
    { 12018, "killstreakThink" },
    { 12019, "giveKillstreak" },
    { 12020, "killstreakSplashNotify" },
    { 12021, "oneManArmyWeaponChangeTracker" },
    { 12022, "isOneManArmyMenu" },
    { 12023, "selectOneManArmyClass" },
    { 12024, "closeOMAMenuOnDeath" },
    { 12025, "giveOneManArmyClass" },
    { 12026, "giveLoadout" },
    { 12027, "omaUseBar" },
    { 12028, "clearPreviousTISpawnpoint" },
    { 12029, "setSpawnpoint" },
    { 12030, "updateTISpawnPosition" },
    { 12031, "TISpawnPosition" },
    { 12032, "isValidTISpawnPosition" },
    { 12033, "monitorTIUse" },
    { 12034, "touchingBadTrigger" },
    { 12035, "enemyTrigger" },
    { 12036, "playerSpawnPos" },
    { 12037, "GlowStickSetupAndWaitForDeath" },
    { 12038, "GlowStickTeamUpdater" },
    { 12039, "GlowStickDamageListener" },
    { 12040, "leaderDialogOnPlayer" },
    { 12041, "GlowStickUseListener" },
    { 12042, "updateEnemyUse" },
    { 12043, "deleteTI" },
    { 12044, "dummyGlowStickDelete" },
    { 12045, "GlowStickEnemyUseListener" },
    { 12046, "makeEnemyUsable" },
    { 12047, "getOtherTeam" },
    { 12048, "watchPaintedDeath" },
    { 12049, "unsetPainted" },
    { 12050, "watchStoppingPowerKill" },
    { 12052, "juicedTimer" },
    { 12053, "juicedIcon" },
    { 12054, "unsetJuicedOnRide" },
    { 12055, "unsetJuicedOnDeath" },
    { 12056, "combatHighOverlay" },
    { 12057, "combatHighTimer" },
    { 12058, "combatHighIcon" },
    { 12059, "unsetCombatHighOnDeath" },
    { 12060, "unsetCombatHighOnRide" },
    { 12061, "giveLightArmor" },
    { 12062, "previousMaxHealth" },
    { 12063, "removeLightArmorOnDeath" },
    { 12064, "removeLightArmor" },
    { 12065, "lastKilledBy" },
    { 12066, "showTo" },
    { 12067, "constantSize" },
    { 12068, "pinToScreenEdge" },
    { 12069, "fadeOutPinnedIcon" },
    { 12070, "is3D" },
    { 12071, "revengeParams" },
    { 12072, "watchRevengeDeath" },
    { 12073, "watchRevengeKill" },
    { 12074, "watchRevengeDisconnected" },
    { 12075, "watchStopRevenge" },
    { 12076, "watchRevengeVictimDisconnected" },
    { 12077, "objIdFriendly" },
    { 12078, "spectateOverride" },
    { 12079, "onSpectatingClient" },
    { 12080, "updateSpectateSettings" },
    { 12081, "setSpectatePermissions" },
    { 12082, "gameEndTime" },
    { 12083, "allowFreeSpectate" },
    { 12084, "allowEnemySpectate" },
    { 12085, "registerAdrenalineInfo" },
    { 12086, "numKills" },
    { 12087, "killedPlayers" },
    { 12088, "killedPlayersCurrent" },
    { 12089, "killedBy" },
    { 12090, "greatestUniquePlayerKills" },
    { 12091, "recentKillCount" },
    { 12092, "lastKillTime" },
    { 12093, "damagedPlayers" },
    { 12094, "damagedPlayer" },
    { 12095, "killedPlayer" },
    { 12096, "lastKilledPlayer" },
    { 12097, "modifiers" },
    { 12098, "attackers" },
    { 12099, "attackerData" },
    { 12100, "firstTimeDamaged" },
    { 12101, "xpEventPopup" },
    { 12102, "assistedSuicide" },
    { 12103, "attackerPosition" },
    { 12104, "killstreaks" },
    { 12105, "setPlayerStat" },
    { 12106, "isLongShot" },
    { 12107, "checkMatchDataKills" },
    { 12108, "iDFlags" },
    { 12109, "inFinalStand" },
    { 12110, "loadoutPrimary" },
    { 12111, "loadoutSecondary" },
    { 12112, "matchMakingGame" },
    { 12113, "setPlayerStatIfLower" },
    { 12114, "setPlayerStatIfGreater" },
    { 12115, "checkMatchDataWeaponKills" },
    { 12116, "checkMatchDataEquipmentKills" },
    { 12117, "camperCheck" },
    { 12118, "lastKillWasCamping" },
    { 12119, "lastKillLocation" },
    { 12120, "lastCampKillTime" },
    { 12121, "consolation" },
    { 12122, "proximityAssist" },
    { 12123, "giveRankXP" },
    { 12124, "proximityKill" },
    { 12125, "longshot" },
    { 12126, "execution" },
    { 12127, "headShot" },
    { 12128, "avengedPlayer" },
    { 12129, "defendedPlayer" },
    { 12130, "postDeathKill" },
    { 12131, "backStab" },
    { 12132, "revenge" },
    { 12133, "multiKill" },
    { 12134, "teamPlayerCardSplash" },
    { 12135, "firstBlood" },
    { 12136, "winningShot" },
    { 12137, "buzzKill" },
    { 12138, "comeBack" },
    { 12139, "disconnected" },
    { 12140, "monitorHealed" },
    { 12141, "updateRecentKills" },
    { 12142, "monitorCrateJacking" },
    { 12143, "monitorObjectives" },
    { 12144, "quickcommands" },
    { 12145, "spamdelay" },
    { 12146, "quickstatements" },
    { 12147, "quickresponses" },
    { 12148, "doQuickMessage" },
    { 12149, "QuickMessageToAll" },
    { 12150, "saveHeadIcon" },
    { 12151, "oldheadicon" },
    { 12152, "oldheadiconteam" },
    { 12153, "restoreHeadIcon" },
    { 12154, "isOptionsMenu" },
    { 12155, "onMenuResponse" },
    { 12156, "forceEnd" },
    { 12157, "autoassign" },
    { 12158, "spectator" },
    { 12159, "selectedClass" },
    { 12160, "class" },
    { 12161, "getTeamAssignment" },
    { 12162, "menuAutoAssign" },
    { 12163, "closeMenus" },
    { 12164, "switching_teams" },
    { 12165, "joining_team" },
    { 12166, "leaving_team" },
    { 12167, "beginClassChoice" },
    { 12168, "allowClassChoice" },
    { 12169, "predictAboutToSpawnPlayerOverTime" },
    { 12170, "bypassClassChoice" },
    { 12171, "beginTeamChoice" },
    { 12172, "showMainMenuForTeam" },
    { 12173, "menuAllies" },
    { 12174, "hasSpawned" },
    { 12175, "menuAxis" },
    { 12176, "menuSpectator" },
    { 12177, "spawnSpectator" },
    { 12178, "menuClass" },
    { 12179, "getClassChoice" },
    { 12180, "getWeaponChoice" },
    { 12181, "lastClass" },
    { 12182, "setClass" },
    { 12183, "tag_stowed_back" },
    { 12184, "tag_stowed_hip" },
    { 12185, "isInKillcam" },
    { 12186, "spawnClient" },
    { 12187, "addToTeam" },
    { 12188, "removeFromTeamCount" },
    { 12189, "allowTeamChoice" },
    { 12190, "addToTeamCount" },
    { 12191, "updateObjectiveText" },
    { 12192, "updateMainMenu" },
    { 12193, "TimeUntilWaveSpawn" },
    { 12194, "lastWave" },
    { 12195, "waveDelay" },
    { 12196, "respawnTimerStartTime" },
    { 12197, "waveSpawnIndex" },
    { 12198, "TeamKillDelay" },
    { 12199, "maxAllowedTeamKills" },
    { 12200, "TimeUntilSpawn" },
    { 12201, "onRespawnDelay" },
    { 12202, "tiSpawnDelay" },
    { 12203, "maySpawn" },
    { 12204, "getGametypeNumLives" },
    { 12205, "disableSpawning" },
    { 12206, "gameHasStarted" },
    { 12207, "setLowerMessage" },
    { 12208, "isLastRound" },
    { 12209, "waitingToSpawn" },
    { 12210, "waitAndSpawnClient" },
    { 12211, "clearLowerMessage" },
    { 12212, "wavePlayerSpawnIndex" },
    { 12213, "waitForTimeOrNotify" },
    { 12214, "wantSafeSpawn" },
    { 12215, "waitRespawnButton" },
    { 12216, "removeSpawnMessageShortly" },
    { 12217, "lastStandRespawnPlayer" },
    { 12218, "dieHardMode" },
    { 12219, "revived" },
    { 12220, "standardmaxHealth" },
    { 12221, "freezePlayerForRoundEnd" },
    { 12222, "getDeathSpawnPoint" },
    { 12223, "showSpawnNotifies" },
    { 12224, "defconSplashNotify" },
    { 12225, "isRested" },
    { 12226, "predictedSpawnPoint" },
    { 12227, "predictedSpawnPointTime" },
    { 12228, "predictAboutToSpawnPlayer" },
    { 12229, "getSpawnpoint_Random" },
    { 12230, "getSpawnPoint" },
    { 12231, "checkPredictedSpawnpointCorrectness" },
    { 12232, "percentage" },
    { 12233, "printPredictedSpawnpointCorrectness" },
    { 12234, "getSpawnOrigin" },
    { 12235, "alternates" },
    { 12236, "tiValidationCheck" },
    { 12237, "spawnPlayer" },
    { 12238, "fauxDead" },
    { 12239, "killsThisLife" },
    { 12240, "ClearKillcamState" },
    { 12241, "cancelKillcam" },
    { 12242, "friendlydamage" },
    { 12243, "afk" },
    { 12244, "inC4Death" },
    { 12245, "inLastStand" },
    { 12246, "clampedHealth" },
    { 12247, "shieldDamage" },
    { 12248, "shieldBulletHits" },
    { 12249, "recentShieldXP" },
    { 12250, "wasAliveAtMatchStart" },
    { 12251, "getTimeLimit" },
    { 12252, "getTimePassed" },
    { 12253, "finalizeSpawnpointChoice" },
    { 12254, "lastspawntime" },
    { 12255, "onSpawnPlayer" },
    { 12256, "playerSpawned" },
    { 12257, "setThirdPersonDOF" },
    { 12258, "gameFlag" },
    { 12259, "getObjectiveHintText" },
    { 12260, "oldNotifyMessage" },
    { 12261, "hidePerksAfterTime" },
    { 12262, "hidePerksOnDeath" },
    { 12263, "hidePerksOnKill" },
    { 12264, "respawn_asSpectator" },
    { 12265, "in_spawnSpectator" },
    { 12266, "getPlayerFromClientNum" },
    { 12267, "onSpawnSpectator" },
    { 12268, "spawnIntermission" },
    { 12269, "clearLowerMessages" },
    { 12270, "postGamePromotion" },
    { 12271, "postGameNotifies" },
    { 12272, "spawnEndOfGame" },
    { 12273, "setSpawnVariables" },
    { 12274, "notifyConnecting" },
    { 12275, "Callback_PlayerDisconnect" },
    { 12276, "connected" },
    { 12277, "removePlayerOnDisconnect" },
    { 12278, "initClientDvars" },
    { 12279, "hitlocInited" },
    { 12280, "getLowestAvailableClientId" },
    { 12281, "Callback_PlayerConnect" },
    { 12282, "usingOnlineDataOffline" },
    { 12283, "resetAdrenaline" },
    { 12284, "leaderDialogQueue" },
    { 12285, "leaderDialogActive" },
    { 12286, "leaderDialogGroups" },
    { 12287, "leaderDialogGroup" },
    { 12288, "statGet" },
    { 12289, "kill_streak" },
    { 12290, "lastGrenadeSuicideTime" },
    { 12291, "teamkillsThisRound" },
    { 12292, "saved_actionSlotData" },
    { 12293, "updateLossStats" },
    { 12294, "OnPlayerConnectAudioInit" },
    { 12295, "isValidClass" },
    { 12296, "Callback_PlayerMigrated" },
    { 12297, "AddLevelsToExperience" },
    { 12298, "GetRestXPCap" },
    { 12299, "setRestXPGoal" },
    { 12300, "forceSpawn" },
    { 12301, "kickIfDontSpawn" },
    { 12302, "kickWait" },
    { 12303, "updateSessionState" },
    { 12304, "initPlayerStats" },
    { 12305, "initBufferedStats" },
    { 12306, "initPersStat" },
    { 12307, "suicides" },
    { 12308, "headshots" },
    { 12309, "captures" },
    { 12310, "returns" },
    { 12311, "defends" },
    { 12312, "plants" },
    { 12313, "defuses" },
    { 12314, "destructions" },
    { 12315, "confirmed" },
    { 12316, "denied" },
    { 12317, "statSetChildBuffered" },
    { 12318, "teamCount" },
    { 12319, "addToAliveCount" },
    { 12320, "aliveCount" },
    { 12321, "maxPlayerCount" },
    { 12322, "removeFromAliveCount" },
    { 12323, "addToLivesCount" },
    { 12324, "livesCount" },
    { 12325, "removeFromLivesCount" },
    { 12326, "removeAllFromLivesCount" },
    { 12327, "callbackStartGameType" },
    { 12328, "callbackPlayerConnect" },
    { 12329, "callbackPlayerDisconnect" },
    { 12330, "callbackPlayerKilled" },
    { 12331, "callbackCodeEndGame" },
    { 12332, "callbackPlayerLastStand" },
    { 12333, "callbackPlayerMigrated" },
    { 12334, "ufo_mode" },
    { 12335, "initGameFlags" },
    { 12336, "initLevelFlags" },
    { 12337, "requiredMapAspectRatio" },
    { 12338, "leaderDialogOnPlayer_func" },
    { 12339, "init_audio" },
    { 12340, "setMapCenterForReflections" },
    { 12341, "lanterns" },
    { 12342, "hurtPlayersThink" },
    { 12343, "setupDestructibleKillCamEnts" },
    { 12344, "deleteDestructibleKillCamEnt" },
    { 12345, "sendMatchData" },
    { 12346, "allowvote" },
    { 12347, "updateServerSettings" },
    { 12348, "constrainGameType" },
    { 12349, "script_gametype_dm" },
    { 12350, "script_gametype_tdm" },
    { 12351, "script_gametype_ctf" },
    { 12352, "script_gametype_hq" },
    { 12353, "script_gametype_sd" },
    { 12354, "script_gametype_koth" },
    { 12355, "killcam" },
    { 12356, "showingFinalKillcam" },
    { 12357, "killcamlength" },
    { 12358, "kc_timer" },
    { 12359, "kc_skiptext" },
    { 12360, "kc_othertext" },
    { 12361, "kc_icon" },
    { 12362, "doFinalKillCamFX" },
    { 12363, "doingFinalKillcamFx" },
    { 12364, "calculateKillCamTime" },
    { 12365, "waittillKillcamOver" },
    { 12366, "setKillCamEntity" },
    { 12367, "waitSkipKillcamButton" },
    { 12368, "waitKCCopyCatButton" },
    { 12369, "waitDeathCopyCatButton" },
    { 12370, "waitCopyCatButton" },
    { 12371, "waitSkipKillcamSafeSpawnButton" },
    { 12372, "endKillcamIfNothingToShow" },
    { 12373, "spawnedKillcamCleanup" },
    { 12374, "endedKillcamCleanup" },
    { 12375, "killcamCleanup" },
    { 12376, "cancelKillCamOnUse" },
    { 12377, "cancelKillCamUseButton" },
    { 12378, "cancelKillCamSafeSpawnButton" },
    { 12379, "cancelKillCamCallback" },
    { 12380, "cancelKillCamSafeSpawnCallback" },
    { 12381, "cancelKillCamOnUse_specificButton" },
    { 12382, "initKCElements" },
    { 12383, "selfDeathIcons" },
    { 12384, "updateDeathIconsEnabled" },
    { 12385, "addDeathIcon" },
    { 12386, "lastDeathIcon" },
    { 12387, "destroySlowly" },
    { 12388, "connectTime" },
    { 12389, "targets" },
    { 12390, "onPlayerDisconnect" },
    { 12391, "onWeaponFired" },
    { 12392, "processKill" },
    { 12393, "heli_types" },
    { 12394, "heli_start_nodes" },
    { 12395, "heli_loop_nodes" },
    { 12396, "heli_leave_nodes" },
    { 12397, "heli_crash_nodes" },
    { 12398, "heli_missile_rof" },
    { 12399, "heli_maxhealth" },
    { 12400, "heli_debug" },
    { 12401, "heli_targeting_delay" },
    { 12402, "heli_turretReloadTime" },
    { 12403, "heli_turretClipSize" },
    { 12404, "heli_visual_range" },
    { 12405, "heli_target_spawnprotection" },
    { 12406, "heli_target_recognition" },
    { 12407, "heli_missile_friendlycare" },
    { 12408, "heli_missile_target_cone" },
    { 12409, "heli_armor_bulletdamage" },
    { 12410, "heli_attract_strength " },
    { 12411, "heli_attract_range" },
    { 12412, "heli_angle_offset" },
    { 12413, "heli_forced_wait" },
    { 12414, "chopper_fx" },
    { 12415, "fx_heli_dust" },
    { 12416, "fx_heli_water" },
    { 12417, "heliDialog" },
    { 12418, "lastHeliDialogTime" },
    { 12419, "queueCreate" },
    { 12420, "makeHeliType" },
    { 12421, "lightFxFunc" },
    { 12422, "addAirExplosion" },
    { 12423, "pavelowLightFx" },
    { 12424, "defaultLightFX" },
    { 12425, "useHelicopter" },
    { 12426, "useHelicopterBlackbox" },
    { 12427, "useHelicopterFlares" },
    { 12428, "useHelicopterMinigun" },
    { 12429, "useHelicopterMK19" },
    { 12430, "tryUseHelicopter" },
    { 12431, "isAirDenied" },
    { 12432, "updateKillstreaks" },
    { 12433, "lifeId" },
    { 12434, "heliType" },
    { 12435, "queueAdd" },
    { 12438, "setUsingRemote" },
    { 12436, "currentActiveVehicleCount" },
    { 12437, "maxVehiclesAllowed" },
    { 12439, "initRideKillstreak" },
    { 12440, "clearUsingRemote" },
    { 12441, "deleteOnEntNotify" },
    { 12442, "startHelicopter" },
    { 12443, "precacheHelicopter" },
    { 12444, "heli_sound" },
    { 12445, "spawn_helicopter" },
    { 12446, "heli_type" },
    { 12447, "zOffset" },
    { 12448, "heliRide" },
    { 12449, "heliRideLifeId" },
    { 12450, "thermalVision" },
    { 12451, "enhanced_vision" },
    { 12452, "thermal_vision" },
    { 12453, "weaponLockThink" },
    { 12454, "heliTargetOrigin" },
    { 12455, "remoteHeliLOS" },
    { 12456, "endRide" },
    { 12457, "getKillstreakWeapon" },
    { 12458, "endRideOnHelicopterDone" },
    { 12459, "getPosNearEnemies" },
    { 12460, "remoteHeliDist" },
    { 12461, "updateAreaNodes" },
    { 12462, "validPlayers" },
    { 12463, "nodeScore" },
    { 12464, "heli_think" },
    { 12465, "targeting_delay" },
    { 12466, "primaryTarget" },
    { 12467, "secondaryTarget" },
    { 12468, "currentstate" },
    { 12469, "makeGunShip" },
    { 12470, "mgTurretLeft" },
    { 12471, "mgTurretRight" },
    { 12472, "deleteTurretsWhenDone" },
    { 12473, "sentry_attackTargets" },
    { 12474, "sentry_burstFireStart" },
    { 12475, "sentry_burstFireStop" },
    { 12476, "heli_existance" },
    { 12477, "queueRemoveFirst" },
    { 12478, "usedKillstreak" },
    { 12479, "heli_targeting" },
    { 12480, "canTarget_turret" },
    { 12481, "getBestPrimaryTarget" },
    { 12482, "threatlevel" },
    { 12483, "update_player_threat" },
    { 12484, "antithreat" },
    { 12485, "heli_reset" },
    { 12486, "addRecentDamage" },
    { 12487, "recentDamageAmount" },
    { 12488, "largeProjectileDamage" },
    { 12489, "vehicleKilled" },
    { 12490, "heli_health" },
    { 12491, "laststate" },
    { 12492, "heli_crash" },
    { 12493, "heli_secondary_explosions" },
    { 12494, "heli_spin" },
    { 12495, "spinSoundShortly" },
    { 12496, "heli_explode" },
    { 12497, "check_owner" },
    { 12498, "heli_leave_on_disconnect" },
    { 12499, "heli_leave_on_changeTeams" },
    { 12500, "heli_leave_on_spawned" },
    { 12501, "heli_leave_on_gameended" },
    { 12502, "heli_leave_on_timeout" },
    { 12503, "attack_targets" },
    { 12504, "attack_secondary" },
    { 12505, "missileTarget" },
    { 12506, "missile_target_sight_check" },
    { 12507, "missile_support" },
    { 12508, "attack_primary" },
    { 12509, "waitOnTargetOrDeath" },
    { 12510, "fireMissile" },
    { 12511, "getOriginOffsets" },
    { 12512, "travelToNode" },
    { 12513, "heli_fly_simple_path" },
    { 12514, "heli_fly_loop_path" },
    { 12515, "desired_speed" },
    { 12516, "desired_accel" },
    { 12517, "heli_loop_speed_control" },
    { 12518, "heli_is_threatened" },
    { 12519, "heli_fly_well" },
    { 12520, "get_best_area_attack_node" },
    { 12521, "heli_leave" },
    { 12522, "pathGoal" },
    { 12523, "wait_and_delete" },
    { 12524, "debug_print3d" },
    { 12525, "debug_print3d_simple" },
    { 12526, "draw_text" },
    { 12527, "addToHeliList" },
    { 12528, "helis" },
    { 12529, "removeFromHeliList" },
    { 12530, "addToLittleBirdList" },
    { 12531, "removeFromLittleBirdListOnDeath" },
    { 12532, "exceededMaxLittlebirds" },
    { 12533, "playFlareFx" },
    { 12534, "deployFlares" },
    { 12535, "heli_flares_monitor" },
    { 12536, "numFlares" },
    { 12537, "handleIncomingStinger" },
    { 12538, "watchStingerProximity" },
    { 12539, "handleIncomingSAM" },
    { 12540, "watchSAMProximity" },
    { 12541, "deleteAfterTime" },
    { 12542, "pavelowMadeSelectionVO" },
    { 12543, "RemoteUAV_fx" },
    { 12544, "RemoteUAV_dialog" },
    { 12545, "RemoteUAV_lastDialogTime" },
    { 12546, "RemoteUAV_noDeployZones" },
    { 12547, "remote_uav" },
    { 12548, "useRemoteUAV" },
    { 12549, "tryUseRemoteUAV" },
    { 12550, "nukeIncoming" },
    { 12551, "isCarrying" },
    { 12552, "giveCarryRemoteUAV" },
    { 12553, "createCarryRemoteUAV" },
    { 12554, "sentryType" },
    { 12555, "canBePlaced" },
    { 12556, "inHeliProximity" },
    { 12557, "rangeTrigger" },
    { 12558, "maxHeight" },
    { 12559, "maxDistance" },
    { 12560, "setCarryingRemoteUAV" },
    { 12561, "carriedBy" },
    { 12562, "carryRemoteUAV_setCarried" },
    { 12563, "isInRemoteNoDeploy" },
    { 12564, "updateCarryRemoteUAVPlacement" },
    { 12565, "carryRemoteUAV_handleExistence" },
    { 12566, "removeRemoteWeapon" },
    { 12567, "startRemoteUAV" },
    { 12568, "remoteUAV_clearRideIntro" },
    { 12569, "lockPlayerForRemoteUAVLaunch" },
    { 12570, "clearPlayerLockFromRemoteUAVLaunch" },
    { 12571, "createRemoteUAV" },
    { 12572, "destroyed" },
    { 12573, "specialDamageCallback" },
    { 12574, "scrambler" },
    { 12575, "smoking" },
    { 12576, "markedPlayers" },
    { 12577, "hasIncoming" },
    { 12578, "incomingMissiles" },
    { 12579, "remoteUAV_Ride" },
    { 12580, "playerLinked" },
    { 12581, "restoreAngles" },
    { 12582, "juggernautOverlay" },
    { 12583, "remote_uav_rideLifeId" },
    { 12584, "remoteUAV" },
    { 12585, "remoteUAV_delayLaunchDialog" },
    { 12586, "remoteUAV_endride" },
    { 12587, "remoteUAV_freezeBuffer" },
    { 12588, "remoteUAV_playerExit" },
    { 12589, "remoteUAV_Track" },
    { 12590, "lastTrackingDialogTime" },
    { 12591, "lockedTarget" },
    { 12592, "trace" },
    { 12593, "remoteUAV_trackEntities" },
    { 12594, "uavType" },
    { 12595, "UAVRemoteMarkedBy" },
    { 12596, "isLeaving" },
    { 12597, "remoteUAV_canTargetUAV" },
    { 12598, "remoteUAV_Fire" },
    { 12599, "remoteUAV_Rumble" },
    { 12600, "remoteUAV_markPlayer" },
    { 12601, "birth_time" },
    { 12602, "remoteUAVMarkedObjID01" },
    { 12603, "remoteUAVMarkedObjID02" },
    { 12604, "remoteUAVMarkedObjID03" },
    { 12605, "remoteUAV_processTaggedAssist" },
    { 12606, "remoteUAV_unmarkRemovedPlayer" },
    { 12607, "remoteUAV_clearMarkedForOwner" },
    { 12608, "remoteUAV_operationRumble" },
    { 12609, "remoteUAV_watch_distance" },
    { 12610, "centerRef" },
    { 12611, "rangeCountdownActive" },
    { 12612, "heliInProximity" },
    { 12613, "remoteUAV_in_range" },
    { 12614, "remoteUAV_staticFade" },
    { 12615, "remoteUAV_rangeCountdown" },
    { 12616, "remoteUAV_explode_on_disconnect" },
    { 12617, "remoteUAV_explode_on_changeTeams" },
    { 12618, "remoteUAV_clear_marked_on_gameEnded" },
    { 12619, "remoteUAV_leave_on_timeout" },
    { 12620, "remoteUAV_leave" },
    { 12621, "remoteUAV_explode_on_death" },
    { 12622, "remoteUAV_cleanup" },
    { 12623, "remoteUAV_light_fx" },
    { 12624, "remoteUAV_handleIncomingStinger" },
    { 12625, "remoteUAV_handleIncomingSAM" },
    { 12626, "remoteUAV_clearIncomingWarning" },
    { 12627, "missile_isIncoming" },
    { 12628, "remoteUAV_watchHeliProximity" },
    { 12629, "remoteUAV_handleDamage" },
    { 12630, "Callback_VehicleDamage" },
    { 12631, "speakers" },
    { 12632, "bcSounds" },
    { 12633, "grenadeProximityTracking" },
    { 12634, "suppressingFireTracking" },
    { 12635, "suppressWaiter" },
    { 12636, "claymoreTracking" },
    { 12637, "reloadTracking" },
    { 12638, "grenadeTracking" },
    { 12639, "sayLocalSoundDelayed" },
    { 12640, "sayLocalSound" },
    { 12641, "doSound" },
    { 12642, "timeHack" },
    { 12643, "isSpeakerInRange" },
    { 12644, "addSpeaker" },
    { 12645, "removeSpeaker" },
    { 12646, "isSwitchingTeams" },
    { 12647, "isTeamSwitchBalanced" },
    { 12648, "isFriendlyFire" },
    { 12649, "killedSelf" },
    { 12650, "isHeadShot" },
    { 12651, "isEnvironmentWeapon" },
    { 12652, "handleTeamChangeDeath" },
    { 12653, "handleWorldDeath" },
    { 12654, "onNormalDeath" },
    { 12655, "handleSuicideDeath" },
    { 12656, "handleFriendlyFireDeath" },
    { 12657, "gracePeriod" },
    { 12658, "handleNormalDeath" },
    { 12659, "updatePersRatio" },
    { 12660, "cloneLoadout" },
    { 12661, "streakType" },
    { 12662, "killShouldAddToKillstreak" },
    { 12663, "setPersStat" },
    { 12664, "statGetChild" },
    { 12665, "statSet" },
    { 12666, "lastAttackedShieldPlayer" },
    { 12667, "lastAttackedShieldTime" },
    { 12668, "isPlayerWeapon" },
    { 12669, "Callback_PlayerKilled" },
    { 12670, "QueueShieldForRemoval" },
    { 12671, "shieldTrashArray" },
    { 12672, "LaunchShield" },
    { 12673, "hasRiotShield" },
    { 12674, "hasRiotShieldEquipped" },
    { 12675, "PlayerKilled_internal" },
    { 12676, "useLastStandParams" },
    { 12677, "eInflictor" },
    { 12678, "iDamage" },
    { 12679, "sMeansOfDeath" },
    { 12680, "sWeapon" },
    { 12681, "sPrimaryWeapon" },
    { 12682, "vDir" },
    { 12683, "sHitLoc" },
    { 12684, "lasttimedamaged" },
    { 12685, "nuked" },
    { 12686, "playerKilled" },
    { 12687, "previousPrimary" },
    { 12688, "trackLeaderBoardDeathStats" },
    { 12689, "trackAttackerLeaderBoardDeathStats" },
    { 12690, "lastDeathPos" },
    { 12691, "sameShotDamage" },
    { 12692, "streakTypeResetsOnDeath" },
    { 12693, "onPlayerKilled" },
    { 12694, "timeUntilRoundEnd" },
    { 12695, "checkForceBleedout" },
    { 12696, "checkKillSteal" },
    { 12697, "attackerEnt" },
    { 12698, "initFinalKillCam" },
    { 12699, "finalKillCam_delay" },
    { 12700, "finalKillCam_victim" },
    { 12701, "finalKillCam_attacker" },
    { 12702, "finalKillCam_attackerNum" },
    { 12703, "finalKillCam_killCamEntityIndex" },
    { 12704, "finalKillCam_killCamEntityStartTime" },
    { 12705, "finalKillCam_sWeapon" },
    { 12706, "finalKillCam_deathTimeOffset" },
    { 12707, "finalKillCam_psOffsetTime" },
    { 12708, "finalKillCam_timeRecorded" },
    { 12709, "finalKillCam_timeGameEnded" },
    { 12710, "finalKillCam_winner" },
    { 12711, "recordFinalKillCam" },
    { 12712, "getSecondsPassed" },
    { 12713, "eraseFinalKillCam" },
    { 12714, "doFinalKillcam" },
    { 12715, "finalKill" },
    { 12716, "anyPlayersInKillcam" },
    { 12717, "resetPlayerVariables" },
    { 12718, "getKillcamEntity" },
    { 12719, "samTurret" },
    { 12720, "imsKillCamEnt" },
    { 12721, "attackerInRemoteKillstreak" },
    { 12722, "remote_mortar" },
    { 12723, "using_remote_turret" },
    { 12724, "using_remote_tank" },
    { 12725, "HitlocDebug" },
    { 12726, "damageInfo" },
    { 12727, "hitloc" },
    { 12728, "bp" },
    { 12729, "jugg" },
    { 12730, "colorIndex" },
    { 12731, "damageInfoColorIndex" },
    { 12732, "damageInfoVictim" },
    { 12733, "giveRecentShieldXP" },
    { 12734, "Callback_PlayerDamage_internal" },
    { 12735, "iDFLAGS_STUN" },
    { 12736, "iDFLAGS_SHIELD_EXPLOSIVE_IMPACT" },
    { 12737, "iDFLAGS_SHIELD_EXPLOSIVE_IMPACT_HUGE" },
    { 12738, "iDFLAGS_SHIELD_EXPLOSIVE_SPLASH" },
    { 12739, "modifyPlayerDamage" },
    { 12740, "iDFlagsTime" },
    { 12741, "canDoCombat" },
    { 12742, "iDFLAGS_NO_KNOCKBACK" },
    { 12743, "killstreakSpawnShield" },
    { 12744, "iDFLAGS_NO_PROTECTION" },
    { 12745, "lastspawnpoint" },
    { 12746, "explosiveInfo" },
    { 12747, "setInflictorStat" },
    { 12748, "lastDamageWasFromEnemy" },
    { 12749, "lastLegitimateAttacker" },
    { 12750, "wasCooked" },
    { 12751, "playerDamaged" },
    { 12752, "useStartSpawn" },
    { 12753, "shouldWeaponFeedback" },
    { 12754, "checkVictimStutter" },
    { 12755, "findIsFacing" },
    { 12756, "stutterStep" },
    { 12757, "inStutter" },
    { 12758, "addAttacker" },
    { 12759, "isPrimary" },
    { 12760, "vPoint" },
    { 12761, "resetAttackerList" },
    { 12762, "Callback_PlayerDamage" },
    { 12763, "finishPlayerDamageWrapper" },
    { 12764, "Callback_PlayerLastStand" },
    { 12765, "dieAfterTime" },
    { 12766, "detonateOnUse" },
    { 12767, "detonateOnDeath" },
    { 12768, "c4DeathDetonate" },
    { 12769, "c4DeathEffect" },
    { 12770, "enableLastStandWeapons" },
    { 12771, "lastStandTimer" },
    { 12772, "maxHealthOverlay" },
    { 12773, "lastStandBleedOut" },
    { 12774, "beingRevived" },
    { 12775, "lastStandAllowSuicide" },
    { 12776, "lastStandKeepOverlay" },
    { 12777, "lastStandWaittillDeath" },
    { 12778, "mayDoLastStand" },
    { 12779, "ensureLastStandParamsValidity" },
    { 12780, "getHitLocHeight" },
    { 12781, "delayStartRagdoll" },
    { 12782, "noRagdollEnts" },
    { 12783, "getMostKilledBy" },
    { 12784, "getMostKilled" },
    { 12785, "damageShellshockAndRumble" },
    { 12786, "reviveSetup" },
    { 12787, "deleteOnReviveOrDeathOrDisconnect" },
    { 12788, "updateUsableByTeam" },
    { 12789, "trackTeamChanges" },
    { 12790, "trackLastStandChanges" },
    { 12791, "reviveTriggerThink" },
    { 12792, "Callback_KillingBlow" },
    { 12793, "combatHigh" },
    { 12794, "emitFallDamage" },
    { 12795, "isFlankKill" },
    { 12796, "_obituary" },
    { 12797, "logPrintPlayerDeath" },
    { 12798, "destroyOnReviveEntDeath" },
    { 12799, "gamemodeModifyPlayerDamage" },
    { 12800, "matchRules_damageMultiplier" },
    { 12801, "matchRules_vampirism" },
    { 12802, "healthRegenDisabled" },
    { 12803, "healthRegenLevel" },
    { 12804, "regenSpeed" },
    { 12805, "atBrinkOfDeath" },
    { 12806, "breathingManager" },
    { 12807, "breathingStopTime" },
    { 12808, "healthRegeneration" },
    { 12809, "healthRegenerated" },
    { 12810, "wait_for_not_using_remote" },
    { 12811, "playerPainBreathingSound" },
    { 12812, "playedStartingMusic" },
    { 12813, "onLastAlive" },
    { 12814, "onRoundSwitch" },
    { 12815, "onGameEnded" },
    { 12816, "playSoundOnPlayers" },
    { 12817, "roundWinnerDialog" },
    { 12818, "roundEndDelay" },
    { 12819, "gameWinnerDialog" },
    { 12820, "postRoundTime" },
    { 12821, "musicController" },
    { 12822, "leaderDialogOnPlayers" },
    { 12823, "suspenseMusic" },
    { 12824, "finalKillcamMusic" },
    { 12825, "giveHighlight" },
    { 12826, "clientMatchDataId" },
    { 12827, "awards" },
    { 12828, "defaultvalue" },
    { 12829, "initPlayerStat" },
    { 12830, "prevPos" },
    { 12831, "previousDeaths" },
    { 12832, "altitudePolls" },
    { 12833, "totalAltitudeSum" },
    { 12834, "usedWeapons" },
    { 12835, "initAwards" },
    { 12836, "initGametypeAwards" },
    { 12837, "initBaseAward" },
    { 12838, "winners" },
    { 12839, "exclusive" },
    { 12840, "initAwardProcess" },
    { 12841, "process" },
    { 12842, "var1" },
    { 12843, "var2" },
    { 12844, "initStat" },
    { 12845, "initStatAward" },
    { 12846, "initDerivedAward" },
    { 12847, "initAwardFlag" },
    { 12848, "initMultiAward" },
    { 12849, "award1_ref" },
    { 12850, "award2_ref" },
    { 12851, "initThresholdAward" },
    { 12852, "setMatchRecordIfGreater" },
    { 12853, "getPlayerStat" },
    { 12854, "getPlayerStatTime" },
    { 12855, "setMatchRecordIfLower" },
    { 12856, "getDecodedRatio" },
    { 12857, "setPersonalBestIfGreater" },
    { 12858, "setPersonalBestIfLower" },
    { 12859, "incPlayerRecord" },
    { 12860, "addAwardWinner" },
    { 12861, "getAwardWinners" },
    { 12862, "clearAwardWinners" },
    { 12863, "setAwardRecord" },
    { 12864, "getAwardRecord" },
    { 12865, "getAwardRecordTime" },
    { 12866, "assignAwards" },
    { 12867, "playerForClientId" },
    { 12868, "assignAward" },
    { 12869, "getAwardType" },
    { 12870, "isMultiAward" },
    { 12871, "isStatAward" },
    { 12872, "isThresholdAward" },
    { 12873, "isAwardFlag" },
    { 12874, "isAwardExclusive" },
    { 12875, "hasDisplayValue" },
    { 12876, "giveAward" },
    { 12877, "getFormattedValue" },
    { 12878, "limitDecimalPlaces" },
    { 12879, "highestWins" },
    { 12880, "lowestWins" },
    { 12881, "lowestWithHalfPlayedTime" },
    { 12882, "statValueChanged" },
    { 12883, "isAtleast" },
    { 12884, "isAtMost" },
    { 12885, "isAtMostWithHalfPlayedTime" },
    { 12886, "setRatio" },
    { 12887, "getKillstreakAwardRef" },
    { 12888, "monitorReloads" },
    { 12889, "monitorShotsFired" },
    { 12890, "monitorSwaps" },
    { 12891, "monitorMovementDistance" },
    { 12892, "monitorPositionCamping" },
    { 12893, "lastCampChecked" },
    { 12894, "positionArray" },
    { 12895, "encodeRatio" },
    { 12896, "getRatioHiVal" },
    { 12897, "getRatioLoVal" },
    { 12898, "monitorEnemyDistance" },
    { 12900, "monitorExplosionsSurvived" },
    { 12901, "monitorShieldBlocks" },
    { 12902, "monitorFlashHits" },
    { 12903, "monitorStunHits" },
    { 12904, "monitorStanceTime" },
    { 12905, "softLandingTriggers" },
    { 12906, "script_type" },
    { 12907, "softLanding" },
    { 12908, "playerEnterSoftLanding" },
    { 12909, "playerLeaveSoftLanding" },
    { 12910, "softLandingWaiter" },
    { 12911, "defconMode" },
    { 12912, "defconStreakAdd" },
    { 12913, "defconPointMod" },
    { 12914, "defconKillstreakWait" },
    { 12915, "defconKillstreakThread" },
    { 12916, "updateDefcon" },
    { 12917, "juggSettings" },
    { 12918, "splashUsedName" },
    { 12920, "overlay" },
    { 12921, "giveJuggernaut" },
    { 12922, "isJuggernautRecon" },
    { 12923, "personalRadar" },
    { 12924, "matchRules_showJuggRadarIcon" },
    { 12925, "objIdEnemy" },
    { 12926, "clearKillstreaks" },
    { 12927, "juggernautSounds" },
    { 12928, "radarMover" },
    { 12929, "juggRemover" },
    { 12930, "isJuggernautDef" },
    { 12931, "isJuggernautGL" },
    { 12932, "juggRemoveOnGameEnded" },
    { 12933, "sentrySettings" },
    { 12934, "burstMin" },
    { 12935, "burstMax" },
    { 12936, "pauseMin" },
    { 12937, "pauseMax" },
    { 12938, "sentryModeOn" },
    { 12939, "sentryModeOff" },
    { 12940, "spinupTime" },
    { 12941, "overheatTime" },
    { 12942, "cooldownTime" },
    { 12943, "fxTime" },
    { 12944, "modelBase" },
    { 12945, "modelPlacement" },
    { 12946, "modelPlacementFailed" },
    { 12947, "modelDestroyed" },
    { 12948, "teamSplash" },
    { 12949, "shouldSplash" },
    { 12950, "voDestroyed" },
    { 12951, "ownerHintString" },
    { 12952, "tryUseAutoSentry" },
    { 12953, "tryUseSAM" },
    { 12954, "giveSentry" },
    { 12955, "validateUseStreak" },
    { 12956, "last_sentry" },
    { 12957, "setCarryingSentry" },
    { 12958, "removeWeapons" },
    { 12959, "restoreWeapon" },
    { 12960, "removePerks" },
    { 12961, "restorePerk" },
    { 12962, "restoreWeapons" },
    { 12963, "restorePerks" },
    { 12964, "waitRestorePerks" },
    { 12965, "createSentryForPlayer" },
    { 12966, "sentry_initSentry" },
    { 12967, "laser_on" },
    { 12968, "momentum" },
    { 12969, "heatLevel" },
    { 12970, "overheated" },
    { 12971, "cooldownWaitTime" },
    { 12972, "sentry_handleDamage" },
    { 12973, "sentry_watchDisabled" },
    { 12974, "sentry_handleDeath" },
    { 12975, "ownerTrigger" },
    { 12976, "forceDisable" },
    { 12977, "inUseBy" },
    { 12978, "turret_overheat_bar" },
    { 12979, "sentry_handleUse" },
    { 12980, "turret_handlePickup" },
    { 12981, "turret_handleUse" },
    { 12982, "sentry_handleOwnerDisconnect" },
    { 12983, "sentry_setOwner" },
    { 12984, "sentry_setPlaced" },
    { 12985, "sentry_setCancelled" },
    { 12986, "sentry_setCarried" },
    { 12987, "updateSentryPlacement" },
    { 12988, "sentry_onCarrierDeath" },
    { 12989, "sentry_onCarrierDisconnect" },
    { 12990, "sentry_onCarrierChangedTeam" },
    { 12991, "sentry_onGameEnded" },
    { 12992, "sentry_setActive" },
    { 12993, "sentry_setInactive" },
    { 12994, "sentry_makeSolid" },
    { 12995, "sentry_makeNotSolid" },
    { 12996, "isFriendlyToSentry" },
    { 12997, "addToTurretList" },
    { 12998, "removeFromTurretList" },
    { 12999, "sentry_timeOut" },
    { 13000, "sentry_targetLockSound" },
    { 13001, "sentry_spinUp" },
    { 13002, "sentry_spinDown" },
    { 13003, "turret_shotMonitor" },
    { 13004, "sentry_heatMonitor" },
    { 13005, "turret_heatMonitor" },
    { 13006, "turret_coolMonitor" },
    { 13007, "PlayHeatFX" },
    { 13008, "playSmokeFX" },
    { 13009, "sentry_beepSounds" },
    { 13010, "sam_attackTargets" },
    { 13011, "samTargetEnt" },
    { 13012, "samMissileGroups" },
    { 13013, "sam_acquireTarget" },
    { 13014, "ac130InUse" },
    { 13015, "sam_fireOnTarget" },
    { 13016, "samMissileGroup" },
    { 13017, "sam_watchLineOfSight" },
    { 13018, "sam_watchLaser" },
    { 13019, "sam_watchCrashing" },
    { 13020, "sam_watchLeaving" },
    { 13021, "airDropCrates" },
    { 13022, "oldAirDropCrates" },
    { 13023, "airDropCrateCollision" },
    { 13024, "numDropCrates" },
    { 13025, "crateTypes" },
    { 13026, "crateMaxVal" },
    { 13027, "lowSpawn" },
    { 13028, "addCrateType" },
    { 13029, "crateFuncs" },
    { 13030, "getRandomCrateType" },
    { 13031, "getCrateTypeForDropType" },
    { 13032, "tryUseAssaultAirdrop" },
    { 13033, "tryUseSupportAirdrop" },
    { 13034, "tryUseAirdropPredatorMissile" },
    { 13035, "tryUseAirdropSentryMinigun" },
    { 13036, "tryUseJuggernautAirdrop" },
    { 13037, "tryUseJuggernautGLAirdrop" },
    { 13038, "tryUseJuggernautReconAirdrop" },
    { 13039, "tryUseJuggernautDefAirdrop" },
    { 13040, "tryUseTrophyAirdrop" },
    { 13041, "tryUseMegaAirdrop" },
    { 13042, "tryUseAirdropTrap" },
    { 13043, "tryUseAirdropRemoteTank" },
    { 13044, "tryUseAmmo" },
    { 13045, "tryUseExplosiveAmmo" },
    { 13046, "tryUseLightArmor" },
    { 13047, "tryUseAirdrop" },
    { 13048, "watchDisconnect" },
    { 13049, "beginAirdropViaMarker" },
    { 13050, "threwAirDropMarker" },
    { 13051, "watchAirDropWeaponChange" },
    { 13052, "isChangingWeapon" },
    { 13053, "watchAirDropMarkerUsage" },
    { 13054, "watchAirDropMarker" },
    { 13055, "beginAirDropMarkerTracking" },
    { 13056, "airDropMarkerActivate" },
    { 13057, "finishSupportEscortUsage" },
    { 13058, "initAirDropCrate" },
    { 13059, "collision" },
    { 13060, "deleteOnOwnerDeath" },
    { 13061, "crateModelTeamUpdater" },
    { 13062, "crateModelPlayerUpdater" },
    { 13063, "crateUseTeamUpdater" },
    { 13064, "crateUseJuggernautUpdater" },
    { 13065, "crateType" },
    { 13066, "crateUsePostJuggernautUpdater" },
    { 13067, "createAirDropCrate" },
    { 13068, "dropType" },
    { 13069, "friendlyModel" },
    { 13070, "enemyModel" },
    { 13071, "dropCrateExistence" },
    { 13072, "trap_createBombSquadModel" },
    { 13073, "crateSetupForUse" },
    { 13074, "setUsableByTeam" },
    { 13075, "dropTheCrate" },
    { 13076, "waitForDropCrateMsg" },
    { 13077, "physicsWaiter" },
    { 13078, "dropTimeOut" },
    { 13079, "getPathStart" },
    { 13080, "getPathEnd" },
    { 13081, "getFlyHeightOffset" },
    { 13082, "airstrikeHeightScale" },
    { 13083, "doFlyBy" },
    { 13084, "leaving" },
    { 13085, "doMegaFlyBy" },
    { 13086, "doC130FlyBy" },
    { 13087, "doMegaC130FlyBy" },
    { 13088, "dropNuke" },
    { 13089, "stopLoopAfter" },
    { 13090, "playloopOnEnt" },
    { 13091, "c130Setup" },
    { 13092, "c130" },
    { 13093, "heliSetup" },
    { 13094, "isAirdrop" },
    { 13095, "watchTimeOut" },
    { 13096, "heli_existence" },
    { 13097, "heli_handleDamage" },
    { 13098, "alreadyDead" },
    { 13099, "heliDestroyed" },
    { 13100, "lbExplode" },
    { 13101, "lbSpin" },
    { 13102, "nukeCaptureThink" },
    { 13103, "crateOtherCaptureThink" },
    { 13104, "isCapturingCrate" },
    { 13105, "crateOwnerCaptureThink" },
    { 13106, "validateOpenConditions" },
    { 13107, "killstreakCrateThink" },
    { 13108, "getKillstreakCrateIcon" },
    { 13109, "getStreakCost" },
    { 13110, "nukeCrateThink" },
    { 13111, "gtnw" },
    { 13112, "capturedNuke" },
    { 13113, "trophyCrateThink" },
    { 13114, "juggernautCrateThink" },
    { 13115, "sentryCrateThink" },
    { 13116, "trapNullFunc" },
    { 13117, "trapCrateThink" },
    { 13118, "bomb" },
    { 13119, "detonateTrap" },
    { 13120, "deleteCrate" },
    { 13121, "sentryUseTracker" },
    { 13122, "giveLocalTrophy" },
    { 13123, "activeTrophy" },
    { 13124, "trophyAmmo" },
    { 13125, "hijackNotify" },
    { 13126, "refillAmmo" },
    { 13127, "isAirdropMarker" },
    { 13128, "createUseEnt" },
    { 13129, "deleteUseEnt" },
    { 13130, "airdropDetonateOnStuck" },
    { 13131, "personalTrophyActive" },
    { 13132, "ospreySettings" },
    { 13133, "modelBlades" },
    { 13134, "tagHatchL" },
    { 13135, "tagHatchR" },
    { 13136, "tagDropCrates" },
    { 13137, "prompt" },
    { 13138, "air_support_locs" },
    { 13139, "tryUseEscortAirdrop" },
    { 13140, "tryUseOspreyGunner" },
    { 13141, "finishOspreyGunnerUsage" },
    { 13142, "stopSelectionWatcher" },
    { 13143, "selectDropLocation" },
    { 13144, "_beginLocationSelection" },
    { 13145, "stopLocationSelection" },
    { 13146, "showIcons" },
    { 13147, "locationObjectives" },
    { 13148, "createAirship" },
    { 13149, "ospreyType" },
    { 13150, "airshipFX" },
    { 13151, "useSupportEscortAirdrop" },
    { 13152, "useOspreyGunner" },
    { 13153, "rideGunner" },
    { 13154, "waitSetThermal" },
    { 13155, "showDefendPrompt" },
    { 13156, "escort_prompt" },
    { 13157, "airShipPitchPropsUp" },
    { 13158, "airShipPitchPropsDown" },
    { 13159, "airShipPitchHatchUp" },
    { 13160, "airShipPitchHatchDown" },
    { 13161, "getBestHeight" },
    { 13162, "bestHeight" },
    { 13163, "airshipFlyDefense" },
    { 13164, "killGuysNearCrates" },
    { 13165, "aiShootPlayer" },
    { 13166, "targetDeathWaiter" },
    { 13167, "waitForConfirmation" },
    { 13168, "airshipFlyGunner" },
    { 13169, "ospreyDropCratesLowImpulse" },
    { 13170, "ospreyDropCrates" },
    { 13171, "endRideOnAirshipDone" },
    { 13172, "match_events_fx" },
    { 13173, "matchEvents" },
    { 13174, "matchEventStarted" },
    { 13175, "getMapCenter" },
    { 13176, "getStartSpawns" },
    { 13177, "doMortar" },
    { 13178, "playSoundinSpace" },
    { 13179, "doSmoke" },
    { 13180, "doAirstrike" },
    { 13181, "doAirstrikeFlyBy" },
    { 13182, "spawnpoints" },
    { 13183, "playPlaneFx" },
    { 13184, "fx_airstrike_afterburner" },
    { 13185, "fx_airstrike_contrail" },
    { 13186, "doPavelow" },
    { 13187, "doHeliInsertion" },
    { 13188, "doOspreyInsertion" },
    { 13189, "drawfriend" },
    { 13190, "showFriendIcon" },
    { 13191, "updateFriendIconSettings" },
    { 13192, "updateFriendIcons" },
    { 13193, "processLobbyScoreboards" },
    { 13194, "setPlayerScoreboardInfo" },
    { 13195, "buildScoreboardType" },
    { 13196, "onForfeit" },
    { 13197, "forfeitInProgress" },
    { 13198, "forfeit_aborted" },
    { 13199, "forcedEnd" },
    { 13200, "forfeitWaitforAbort" },
    { 13201, "matchForfeitTimer" },
    { 13202, "matchForfeitText" },
    { 13203, "matchForfeitTimer_Internal" },
    { 13204, "default_onDeadEvent" },
    { 13205, "default_onOneLeftEvent" },
    { 13206, "getLastLivingPlayer" },
    { 13207, "default_onTimeLimit" },
    { 13208, "default_onHalfTime" },
    { 13209, "hostForcedEnd" },
    { 13210, "onDeadEvent" },
    { 13211, "oneLeftTime" },
    { 13212, "onOneLeftEvent" },
    { 13213, "getPotentialLivingPlayers" },
    { 13214, "waittillFinalKillcamDone" },
    { 13215, "timeLimitClock_Intermission" },
    { 13216, "waitForPlayers" },
    { 13217, "prematchPeriod" },
    { 13218, "hintMessage" },
    { 13219, "gameFlagSet" },
    { 13220, "updateWinStats" },
    { 13221, "updateTieStats" },
    { 13222, "updateWinLossStats" },
    { 13223, "privateMatch" },
    { 13224, "wasLastRound" },
    { 13225, "updateMatchBonusScores" },
    { 13226, "endGameUpdate" },
    { 13227, "getSPM" },
    { 13228, "matchBonus" },
    { 13229, "giveMatchBonus" },
    { 13230, "setXenonRanks" },
    { 13231, "checkTimeLimit" },
    { 13232, "getHalfTime" },
    { 13233, "onHalfTime" },
    { 13234, "onTimeLimit" },
    { 13235, "checkHalfTime" },
    { 13236, "getTimeRemaining" },
    { 13237, "scoreLimitOverride" },
    { 13238, "updateGametypeDvars" },
    { 13239, "matchStartTimerPC" },
    { 13240, "prematchPeriodEnd" },
    { 13241, "matchStartTimer_Internal" },
    { 13242, "matchStartTimerSkip" },
    { 13243, "halftimeSubCaption" },
    { 13244, "halftimeType" },
    { 13245, "checkRoundSwitch" },
    { 13246, "roundSwitch" },
    { 13247, "timePaused" },
    { 13248, "freeGameplayHudElems" },
    { 13249, "perkicon" },
    { 13250, "perkname" },
    { 13251, "lowerMessage" },
    { 13252, "lowerTimer" },
    { 13253, "getHostPlayer" },
    { 13254, "hostIdledOut" },
    { 13255, "roundEndWait" },
    { 13256, "isDoingSplash" },
    { 13257, "roundEndDoF" },
    { 13258, "Callback_StartGameType" },
    { 13259, "levelFlagInit" },
    { 13260, "intermission" },
    { 13261, "onPrecacheGameType" },
    { 13262, "killstreakRewards" },
    { 13263, "alivePlayers" },
    { 13264, "activePlayers" },
    { 13265, "gameFlagInit" },
    { 13266, "halftimeRoundEndDelay" },
    { 13267, "onStartGameType" },
    { 13268, "updateWatchedDvars" },
    { 13269, "Callback_CodeEndGame" },
    { 13270, "timeLimitThread" },
    { 13271, "updateUIScoreLimit" },
    { 13272, "playTickingSound" },
    { 13273, "bombTimer" },
    { 13274, "stopTickingSound" },
    { 13275, "timeLimitClock" },
    { 13276, "timerStopped" },
    { 13277, "gameTimer" },
    { 13278, "discardTime" },
    { 13279, "timerStoppedForGameMode" },
    { 13280, "timerPauseTime" },
    { 13281, "pauseTimer" },
    { 13282, "resumeTimer" },
    { 13283, "startGame" },
    { 13284, "spawnPerFrameUpdate" },
    { 13285, "roundBegin" },
    { 13286, "waveSpawnTimer" },
    { 13287, "getBetterTeam" },
    { 13288, "rankedMatchUpdates" },
    { 13289, "displayRoundEnd" },
    { 13290, "teamOutcomeNotify" },
    { 13291, "outcomeNotify" },
    { 13292, "displayGameEnd" },
    { 13293, "displayRoundSwitch" },
    { 13294, "endGameOvertime" },
    { 13295, "endGameHalfTime" },
    { 13296, "levelFlagSet" },
    { 13297, "wasOnlyRound" },
    { 13298, "levelFlagClear" },
    { 13299, "roundEnd" },
    { 13300, "updateEndReasonText" },
    { 13301, "hitRoundLimit" },
    { 13302, "hitWinLimit" },
    { 13303, "estimatedTimeTillScoreLimit" },
    { 13304, "getScorePerMinute" },
    { 13305, "getScorePerRemaining" },
    { 13306, "giveLastOnTeamWarning" },
    { 13307, "waitTillRecoveredHealth" },
    { 13308, "processLobbyData" },
    { 13309, "incrementWeaponStat" },
    { 13310, "updateWeaponBufferedStats" },
    { 13311, "incrementAttachmentStat" },
    { 13312, "playerAffectedArray" },
    { 13313, "threadedSetWeaponStatByName" },
    { 13314, "checkForPersonalBests" },
    { 13315, "xpGains" },
    { 13316, "checkForBestWeapon" },
    { 13317, "gametypestarted" },
    { 13318, "damageCallback" },
    { 13319, "callbackHostMigration" },
    { 13320, "SetupDamageFlags" },
    { 13321, "iDFLAGS_RADIUS" },
    { 13322, "iDFLAGS_NO_ARMOR" },
    { 13323, "iDFLAGS_NO_TEAM_PROTECTION" },
    { 13324, "iDFLAGS_PASSTHRU" },
    { 13325, "SetupCallbacks" },
    { 13326, "SetDefaultCallbacks" },
    { 13327, "AbortLevel" },
    { 13328, "callbackVoid" },
    { 13329, "beginHarrier" },
    { 13330, "getCorrectHeight" },
    { 13331, "spawnDefensiveHarrier" },
    { 13332, "accel" },
    { 13333, "defendLoc" },
    { 13334, "defendLocation" },
    { 13335, "closeToGoalCheck" },
    { 13336, "engageGround" },
    { 13337, "harrierLeave" },
    { 13338, "airPlane" },
    { 13339, "harrierDelete" },
    { 13340, "harrierTimer" },
    { 13341, "randomHarrierMovement" },
    { 13342, "getNewPoint" },
    { 13343, "bestTarget" },
    { 13344, "traceNewPoint" },
    { 13345, "traceGroundPoint" },
    { 13346, "playHarrierFx" },
    { 13347, "harrier_afterburnerfx" },
    { 13348, "stopHarrierWingFx" },
    { 13349, "startHarrierWingFx" },
    { 13350, "fireOnTarget" },
    { 13351, "isReadyToFire" },
    { 13352, "acquireGroundTarget" },
    { 13353, "backToDefendLocation" },
    { 13354, "wouldCollide" },
    { 13355, "watchTargetDeath" },
    { 13356, "watchTargetLOS" },
    { 13357, "breakTarget" },
    { 13358, "harrierGetTargets" },
    { 13359, "nonTarget" },
    { 13360, "isTarget" },
    { 13361, "getBestTarget" },
    { 13362, "targetEnt" },
    { 13363, "checkForFriendlies" },
    { 13364, "playDamageEfx" },
    { 13365, "harrier_smoke" },
    { 13366, "harrierDestroyed" },
    { 13367, "harrierExplode" },
    { 13368, "harrier_deathfx" },
    { 13369, "harrierSpin" },
    { 13370, "engageVehicle" },
    { 13371, "fireOnVehicleTarget" },
    { 13372, "acquireVehicleTarget" },
    { 13373, "watchVehTargetCrash" },
    { 13374, "watchVehTargetDeath" },
    { 13375, "breakVehTarget" },
    { 13376, "evasiveManuverOne" },
    { 13377, "removeFromHeliListOnDeath" },
    { 13378, "onfirefx" },
    { 13379, "airstrikefx" },
    { 13380, "mortareffect" },
    { 13381, "bombstrike" },
    { 13382, "stealthbombfx" },
    { 13383, "planes" },
    { 13384, "dangerMaxRadius" },
    { 13385, "dangerMinRadius" },
    { 13386, "dangerForwardPush" },
    { 13387, "dangerOvalScale" },
    { 13388, "artilleryDangerCenters" },
    { 13389, "tryUseDefaultAirstrike" },
    { 13390, "tryUsePrecisionAirstrike" },
    { 13391, "tryUseSuperAirstrike" },
    { 13392, "tryUseHarrierAirstrike" },
    { 13393, "tryUseStealthAirstrike" },
    { 13394, "tryUseAirstrike" },
    { 13395, "clearProgress" },
    { 13396, "getAirstrikeDanger" },
    { 13397, "getSingleAirstrikeDanger" },
    { 13398, "pointIsInAirstrikeArea" },
    { 13399, "losRadiusDamage" },
    { 13400, "isSentry" },
    { 13401, "airStrikeDamagedEntsCount" },
    { 13402, "airStrikeDamagedEnts" },
    { 13403, "airstrikeDamageEntsThread" },
    { 13404, "airstrikeDamagedEntsIndex" },
    { 13405, "radiusArtilleryShellshock" },
    { 13406, "artilleryShellshock" },
    { 13407, "beingArtilleryShellshocked" },
    { 13408, "doBomberStrike" },
    { 13409, "bomberDropBombs" },
    { 13410, "playBombFx" },
    { 13411, "stealthBomber_killCam" },
    { 13412, "airstrikeType" },
    { 13413, "callStrike_bomb" },
    { 13414, "doPlaneStrike" },
    { 13415, "handleDeath" },
    { 13416, "addPlaneToList" },
    { 13417, "removePlaneFromList" },
    { 13418, "callStrike_bombEffect" },
    { 13419, "spawnbomb" },
    { 13420, "callStrike" },
    { 13421, "getExplodeDistance" },
    { 13422, "targetGetDist" },
    { 13423, "waitForAirstrikeCancel" },
    { 13424, "selectAirstrikeLocation" },
    { 13425, "finishAirstrikeUsage" },
    { 13426, "useAirstrike" },
    { 13427, "handleEMP" },
    { 13428, "airstrikeMadeSelectionVO" },
    { 13429, "findBoxCenter" },
    { 13430, "expandMins" },
    { 13431, "expandMaxs" },
    { 13432, "addSpawnPoints" },
    { 13433, "teamSpawnPoints" },
    { 13434, "inited" },
    { 13435, "spawnMins" },
    { 13436, "spawnMaxs" },
    { 13437, "placeSpawnPoints" },
    { 13438, "startSpawnPoints" },
    { 13439, "getSpawnpointArray" },
    { 13440, "extraspawnpoints" },
    { 13441, "expandSpawnpointBounds" },
    { 13442, "spawnPointInit" },
    { 13443, "sightTracePoint" },
    { 13444, "lastspawnedplayer" },
    { 13445, "outside" },
    { 13446, "AddAlternateSpawnpoint" },
    { 13447, "getTeamSpawnPoints" },
    { 13448, "getSpawnpoint_Final" },
    { 13449, "getBestWeightedSpawnpoint" },
    { 13450, "sights" },
    { 13451, "lastSightTraceTime" },
    { 13452, "getAllOtherPlayers" },
    { 13453, "initWeights" },
    { 13454, "getSpawnpoint_NearTeam" },
    { 13455, "numPlayersAtLastUpdate" },
    { 13456, "weightedDistSum" },
    { 13457, "distSum" },
    { 13458, "favorCloseSpawnEnt" },
    { 13459, "favorCloseSpawnScalar" },
    { 13460, "getSpawnpoint_SafeSpawn" },
    { 13461, "minDist" },
    { 13462, "safeSpawnDangerDist" },
    { 13463, "getSpawnpoint_DM" },
    { 13464, "Spawnlogic_Begin" },
    { 13465, "spawnlogic_deaths" },
    { 13466, "spawnlogic_spawnkills" },
    { 13467, "pipebombs" },
    { 13468, "tanks" },
    { 13469, "ims" },
    { 13470, "safespawns" },
    { 13471, "sightCheckCost" },
    { 13472, "watchSpawnProfile" },
    { 13473, "spawnProfile" },
    { 13474, "spawnGraphCheck" },
    { 13475, "spawnGraph" },
    { 13476, "secondfloor" },
    { 13477, "fake" },
    { 13478, "drawSpawnGraph" },
    { 13479, "setupSpawnGraphPoint" },
    { 13480, "visible" },
    { 13481, "spawnGraphLine" },
    { 13482, "loopbotspawns" },
    { 13483, "trackGrenades" },
    { 13484, "trackMissiles" },
    { 13485, "isPointVulnerable" },
    { 13486, "claymoremodelcenteroffset" },
    { 13487, "claymoreDetectionRadius" },
    { 13488, "claymoreDetectionConeAngle" },
    { 13489, "avoidWeaponDamage" },
    { 13490, "adjustSightValue" },
    { 13491, "spawnPointUpdate" },
    { 13492, "getFloatProperty" },
    { 13493, "nearbyPenalty" },
    { 13494, "attackHeightPos" },
    { 13495, "getLosPenalty" },
    { 13496, "lastMinuteSightTraces" },
    { 13497, "getRevengeSpawnPenalty" },
    { 13498, "getRevengeSpawnDistanceSq" },
    { 13499, "avoidRevengeSpawn" },
    { 13500, "avoidRevengeSpawnStage2" },
    { 13501, "avoidVisibleEnemies" },
    { 13502, "forceSpawnNearTeammates" },
    { 13503, "avoidSpawnReuse" },
    { 13504, "avoidSameSpawn" },
    { 13505, "SetupKillstreakSpawnShield" },
    { 13506, "ac130_use_duration" },
    { 13507, "ac130_num_flares" },
    { 13508, "radioForcedTransmissionQueue" },
    { 13509, "enemiesKilledInTimeWindow" },
    { 13510, "lastRadioTransmission" },
    { 13511, "HUDItem" },
    { 13512, "physicsSphereRadius" },
    { 13513, "physicsSphereForce" },
    { 13514, "weaponReloadTime" },
    { 13515, "ac130_Speed" },
    { 13516, "ac130Queue" },
    { 13517, "tryUseAC130" },
    { 13564, "crashed" },
    { 13565, "init_sounds" },
    { 13566, "add_context_sensative_dialog" },
    { 13567, "played" },
    { 13568, "sounds" },
    { 13569, "add_context_sensative_timeout" },
    { 13570, "context_sensative_dialog_timeouts" },
    { 13571, "groups" },
    { 13572, "deleteOnAC130PlayerRemoved" },
    { 13573, "setAC130Player" },
    { 13574, "incomingMissile" },
    { 13575, "playAC130Effects" },
    { 13576, "AC130_AltScene" },
    { 13577, "cameraModel" },
    { 13578, "removeAC130PlayerOnGameEnd" },
    { 13579, "removeAC130PlayerOnGameCleanup" },
    { 13580, "removeAC130PlayerOnDeath" },
    { 13581, "removeAC130PlayerOnCrash" },
    { 13582, "removeAC130PlayerOnDisconnect" },
    { 13583, "removeAC130PlayerOnChangeTeams" },
    { 13584, "removeAC130PlayerOnSpectate" },
    { 13585, "removeAC130PlayerAfterTime" },
    { 13586, "removeAC130Player" },
    { 13587, "darkScreenOverlay" },
    { 13588, "damageTracker" },
    { 13589, "ac130_spawn" },
    { 13590, "overlay_coords" },
    { 13591, "updateAimingCoords" },
    { 13592, "ac130ShellShock" },
    { 13593, "rotatePlane" },
    { 13594, "attachPlayer" },
    { 13595, "changeWeapons" },
    { 13596, "weaponFiredThread" },
    { 13597, "weaponReload" },
    { 13598, "clouds" },
    { 13599, "clouds_create" },
    { 13600, "playerWeapon" },
    { 13601, "gun_fired_and_ready_105mm" },
    { 13602, "shotFired" },
    { 13603, "shotFiredPhysicsSphere" },
    { 13604, "shotFiredDarkScreenOverlay" },
    { 13605, "add_beacon_effect" },
    { 13606, "context_Sensative_Dialog" },
    { 13607, "context_Sensative_Dialog_Guy_In_Sight" },
    { 13608, "context_Sensative_Dialog_Guy_In_Sight_Check" },
    { 13609, "context_Sensative_Dialog_Guy_Crawling" },
    { 13610, "context_Sensative_Dialog_Guy_Pain" },
    { 13611, "context_Sensative_Dialog_Secondary_Explosion_Vehicle" },
    { 13612, "enemy_killed_thread" },
    { 13613, "context_Sensative_Dialog_Kill" },
    { 13614, "context_Sensative_Dialog_Kill_Thread" },
    { 13615, "context_Sensative_Dialog_Locations" },
    { 13616, "context_Sensative_Dialog_Locations_Thread" },
    { 13617, "context_Sensative_Dialog_Locations_Add_Notify_Event" },
    { 13618, "context_Sensative_Dialog_VehicleSpawn" },
    { 13619, "context_Sensative_Dialog_VehicleDeath" },
    { 13620, "context_Sensative_Dialog_Filler" },
    { 13621, "radio_in_use" },
    { 13622, "context_Sensative_Dialog_Play_Random_Group_Sound" },
    { 13623, "context_Sensative_Dialog_Timedout" },
    { 13624, "playSoundOverRadio" },
    { 13625, "playAliasOverRadio" },
    { 13626, "debug_circle" },
    { 13627, "stingerProximityDetonate" },
    { 13628, "samProximityDetonate" },
    { 13629, "crashPlane" },
    { 13630, "angelFlarePrecache" },
    { 13631, "angel_flare" },
    { 13632, "missileRemoteLaunchVert" },
    { 13633, "missileRemoteLaunchHorz" },
    { 13634, "missileRemoteLaunchTargetDist" },
    { 13635, "remotemissile_fx" },
    { 13636, "tryUsePredatorMissile" },
    { 13637, "getBestSpawnPoint" },
    { 13638, "spawnScore" },
    { 13639, "_fire" },
    { 13640, "handleDamage" },
    { 13641, "MissileEyes" },
    { 13642, "delayedFOFOverlay" },
    { 13643, "staticEffect" },
    { 13644, "archive" },
    { 13645, "Player_CleanupOnTeamChange" },
    { 13646, "Rocket_CleanupOnDeath" },
    { 13647, "Player_CleanupOnGameEnded" },
    { 13648, "radarViewTime" },
    { 13649, "uavBlockTime" },
    { 13650, "uav_fx" },
    { 13651, "killstreakSetupFuncs" },
    { 13652, "UAVRig" },
    { 13653, "activeUAVs" },
    { 13654, "activeCounterUAVs" },
    { 13655, "rotateUAVRig" },
    { 13656, "launchUAV" },
    { 13657, "timeToAdd" },
    { 13658, "spawnFxDelay" },
    { 13659, "monitorUAVStrike" },
    { 13660, "showLazeMessage" },
    { 13661, "waitForLazeDiscard" },
    { 13662, "waitForLazedTarget" },
    { 13663, "waitFxEntDie" },
    { 13664, "waittill_notify_or_timeout_hostmigration_pause" },
    { 13665, "updateUAVModelVisibility" },
    { 13666, "tryUseUAV" },
    { 13667, "tryUseUAVSupport" },
    { 13668, "tryUseDoubleUAV" },
    { 13669, "tryUseTripleUAV" },
    { 13670, "tryUseCounterUAV" },
    { 13671, "UAVStrikeSetup" },
    { 13672, "usedStrikeUAV" },
    { 13673, "tryUseUAVStrike" },
    { 13674, "tryUseDirectionalUAV" },
    { 13675, "useUAV" },
    { 13676, "UAVTracker" },
    { 13677, "_getRadarStrength" },
    { 13678, "updateTeamUAVStatus" },
    { 13679, "updatePlayersUAVStatus" },
    { 13680, "blockPlayerUAV" },
    { 13681, "updateTeamUAVType" },
    { 13682, "usePlayerUAV" },
    { 13683, "setTeamRadarWrapper" },
    { 13684, "addUAVModel" },
    { 13685, "removeUAVModel" },
    { 13686, "addActiveUAV" },
    { 13687, "addActiveCounterUAV" },
    { 13688, "removeActiveUAV" },
    { 13689, "removeActiveCounterUAV" },
    { 13690, "tryUseLbFlock" },
    { 13691, "selectLbStrikeLocation" },
    { 13692, "littlebirdMadeSelectionVO" },
    { 13693, "finishLbStrikeUsage" },
    { 13695, "getFlightPath" },
    { 13696, "doLbStrike" },
    { 13697, "spawnAttackLittleBird" },
    { 13698, "killCount" },
    { 13699, "mgTurret1" },
    { 13700, "monitorKills" },
    { 13701, "startLbFiring1" },
    { 13702, "flock_handleDamage" },
    { 13703, "trail_fx" },
    { 13704, "removeLittlebird" },
    { 13705, "lbStrike" },
    { 13706, "heliGuardSettings" },
    { 13707, "weaponInfo" },
    { 13708, "weaponModelLeft" },
    { 13709, "weaponModelRight" },
    { 13710, "weaponTagLeft" },
    { 13711, "weaponTagRight" },
    { 13712, "sentryMode" },
    { 13713, "tryUseLBSupport" },
    { 13714, "littlebirdGuard" },
    { 13715, "air_node_mesh" },
    { 13716, "createLBGuard" },
    { 13717, "followSpeed" },
    { 13718, "heliGuardType" },
    { 13719, "targettingRadius" },
    { 13720, "attract_strength" },
    { 13721, "attract_range" },
    { 13722, "hasDodged" },
    { 13723, "lbSupport_lightFX" },
    { 13724, "startLBSupport" },
    { 13725, "lbSupport_followPlayer" },
    { 13726, "lbSupport_moveToPlayer" },
    { 13727, "inTransit" },
    { 13728, "lbSupport_watchDeath" },
    { 13729, "lbSupport_watchTimeout" },
    { 13730, "lbSupport_watchOwnerLoss" },
    { 13731, "lbSupport_watchOwnerDamage" },
    { 13732, "lbSupport_watchRoundEnd" },
    { 13733, "lbSupport_leave" },
    { 13734, "lbSupport_handleDamage" },
    { 13735, "marker" },
    { 13736, "lbSupport_watchSAMProximity" },
    { 13737, "lbSupport_getClosestStartNode" },
    { 13738, "air_start_nodes" },
    { 13739, "lbSupport_getClosestNode" },
    { 13740, "lbSupport_getClosestLinkedNode" },
    { 13741, "neighbors" },
    { 13742, "lbSupport_arrayContains" },
    { 13743, "lbSupport_getLinkedStructs" },
    { 13744, "lbSupport_setAirStartNodes" },
    { 13745, "lbSupport_setAirNodeMesh" },
    { 13746, "lbSupport_attackTargets" },
    { 13747, "lbSupport_burstFireStart" },
    { 13748, "lbSupport_burstFireStop" },
    { 13749, "empTimeout" },
    { 13750, "EMP_Use" },
    { 13751, "EMP_JamTeam" },
    { 13752, "EMP_JamPlayers" },
    { 13753, "empPlayerFFADisconnect" },
    { 13754, "empEffects" },
    { 13755, "empEffect" },
    { 13756, "EMP_TeamTracker" },
    { 13757, "EMP_PlayerTracker" },
    { 13758, "destroyActiveVehicles" },
    { 13759, "nukeTimer" },
    { 13760, "cancelMode" },
    { 13761, "teamNukeEMPed" },
    { 13762, "nukeEmpTimeout" },
    { 13763, "nukeInfo" },
    { 13764, "tryUseNuke" },
    { 13765, "delaythread_nuke" },
    { 13766, "doNuke" },
    { 13767, "cancelNukeOnDeath" },
    { 13768, "nukeSoundIncoming" },
    { 13769, "nukeSoundExplosion" },
    { 13770, "nukeEffects" },
    { 13771, "nukeEffect" },
    { 13772, "nukeAftermathEffect" },
    { 13773, "nukeSlowMo" },
    { 13774, "nukeVision" },
    { 13775, "nukeVisionInProgress" },
    { 13776, "nukeDeath" },
    { 13777, "nukeEarthquake" },
    { 13778, "nuke_EMPJam" },
    { 13779, "nuke_EMPTeamTracker" },
    { 13780, "update_ui_timers" },
    { 13781, "remote_mortar_fx" },
    { 13782, "tryUseRemoteMortar" },
    { 13783, "startRemoteMortar" },
    { 13784, "spawnRemote" },
    { 13785, "lookCenter" },
    { 13786, "remoteRide" },
    { 13787, "remoteTargeting" },
    { 13788, "remoteFiring" },
    { 13789, "firingReaper" },
    { 13790, "remoteZoom" },
    { 13791, "zoomed" },
    { 13792, "remoteMissileDistance" },
    { 13793, "remoteMissileLife" },
    { 13794, "remoteEndRide" },
    { 13795, "handleTimeout" },
    { 13796, "handleOwnerChangeTeam" },
    { 13797, "handleOwnerDisconnect" },
    { 13798, "removeRemote" },
    { 13799, "remoteLeave" },
    { 13800, "boxSettings" },
    { 13801, "capturingString" },
    { 13802, "eventString" },
    { 13803, "splashName" },
    { 13804, "shaderName" },
    { 13805, "lifeSpan" },
    { 13806, "xp" },
    { 13807, "tryUseDeployableVest" },
    { 13808, "beginDeployableViaMarker" },
    { 13809, "watchMarkerUsage" },
    { 13810, "watchMarker" },
    { 13811, "takeWeaponOnStuck" },
    { 13812, "beginMarkerTracking" },
    { 13813, "markerActivate" },
    { 13814, "isMarker" },
    { 13815, "createBoxForPlayer" },
    { 13816, "boxType" },
    { 13817, "box_setActive" },
    { 13818, "isUsable" },
    { 13819, "box_playerConnected" },
    { 13820, "box_playerJoinedTeam" },
    { 13821, "box_setInactive" },
    { 13822, "box_handleDamage" },
    { 13823, "box_handleDeath" },
    { 13824, "box_handleOwnerDisconnect" },
    { 13825, "boxThink" },
    { 13826, "doubleDip" },
    { 13827, "boxCaptureThink" },
    { 13828, "isFriendlyToBox" },
    { 13829, "box_timeOut" },
    { 13830, "box_ModelTeamUpdater" },
    { 13831, "boxParams" },
    { 13832, "disableWhenJuggernaut" },
    { 13833, "imsSettings" },
    { 13834, "modelBombSquad" },
    { 13835, "placeString" },
    { 13836, "cannotPlaceString" },
    { 13837, "attacks" },
    { 13838, "modelExplosive1" },
    { 13839, "modelExplosive2" },
    { 13840, "modelExplosive3" },
    { 13841, "modelExplosive4" },
    { 13842, "modelLid1" },
    { 13843, "modelLid2" },
    { 13844, "modelLid3" },
    { 13845, "modelLid4" },
    { 13846, "tagExplosive1" },
    { 13847, "tagExplosive2" },
    { 13848, "tagExplosive3" },
    { 13849, "tagExplosive4" },
    { 13850, "tagLid1" },
    { 13851, "tagLid2" },
    { 13852, "tagLid3" },
    { 13853, "tagLid4" },
    { 13854, "tryUseIMS" },
    { 13855, "imsList" },
    { 13856, "giveIMS" },
    { 13857, "setCarryingIMS" },
    { 13858, "createIMSForPlayer" },
    { 13859, "imsType" },
    { 13860, "createIMS" },
    { 13861, "lid1" },
    { 13862, "lid2" },
    { 13863, "lid3" },
    { 13864, "lid4" },
    { 13865, "explosive1" },
    { 13866, "explosive2" },
    { 13867, "explosive3" },
    { 13868, "explosive4" },
    { 13869, "ims_createBombSquadModel" },
    { 13870, "ims_handleDamage" },
    { 13871, "ims_handleDeath" },
    { 13872, "ims_handleUse" },
    { 13873, "ims_handleOwnerDisconnect" },
    { 13874, "ims_setPlaced" },
    { 13875, "ims_setCancelled" },
    { 13876, "ims_setCarried" },
    { 13877, "updateIMSPlacement" },
    { 13878, "ims_onCarrierDeath" },
    { 13879, "ims_onCarrierDisconnect" },
    { 13880, "ims_onGameEnded" },
    { 13881, "ims_setActive" },
    { 13882, "attackTrigger" },
    { 13883, "attackMoveTime" },
    { 13884, "ims_playerConnected" },
    { 13885, "ims_playerJoinedTeam" },
    { 13886, "ims_blinky_light" },
    { 13887, "ims_setInactive" },
    { 13888, "isFriendlyToIMS" },
    { 13889, "ims_attackTargets" },
    { 13890, "fire_sensor" },
    { 13891, "ims_timeOut" },
    { 13892, "addToIMSList" },
    { 13893, "removeFromIMSList" },
    { 13894, "ims_hideAllParts" },
    { 13895, "ims_showAllParts" },
    { 13896, "turretSettings" },
    { 13897, "hintEnter" },
    { 13898, "hintExit" },
    { 13899, "laptopInfo" },
    { 13900, "remoteInfo" },
    { 13901, "tryUseRemoteMGTurret" },
    { 13902, "takeKillstreakWeapons" },
    { 13903, "takeKillstreakWeaponIfNoDupe" },
    { 13904, "tryUseRemoteTurret" },
    { 13905, "setCarryingTurret" },
    { 13906, "restoreWeaponClipAmmo" },
    { 13907, "restoreWeaponStockAmmo" },
    { 13908, "waitRestoreWeapons" },
    { 13909, "turret_setPlaced" },
    { 13910, "turret_setCancelled" },
    { 13911, "turret_setCarried" },
    { 13912, "updateTurretPlacement" },
    { 13913, "turret_onCarrierDeath" },
    { 13914, "turret_onCarrierDisconnect" },
    { 13915, "turret_onCarrierChangedTeam" },
    { 13916, "turret_onGameEnded" },
    { 13917, "createTurretForPlayer" },
    { 13918, "stunned" },
    { 13919, "stunnedTime" },
    { 13920, "damageFade" },
    { 13921, "turret_setActive" },
    { 13922, "remoteTurretList" },
    { 13923, "startUsingRemoteTurret" },
    { 13924, "stopUsingRemoteTurret" },
    { 13925, "using_remote_turret_when_died" },
    { 13926, "watchOwnerMessageOnDeath" },
    { 13927, "watchEnterAndExit" },
    { 13928, "enter_message_deleted" },
    { 13929, "turret_blinky_light" },
    { 13930, "turret_setInactive" },
    { 13931, "clearRideIntro" },
    { 13932, "turret_handleOwnerDisconnect" },
    { 13933, "turret_timeOut" },
    { 13934, "turret_handleDeath" },
    { 13935, "target_ent" },
    { 13936, "turret_handleDamage" },
    { 13937, "turret_incrementDamageFade" },
    { 13938, "turret_watchLowHealth" },
    { 13939, "turret_stun" },
    { 13940, "tankSettings" },
    { 13941, "mgTurretInfo" },
    { 13942, "glTurretInfo" },
    { 13943, "modelMGTurret" },
    { 13944, "stringPlace" },
    { 13945, "stringCannotPlace" },
    { 13946, "remote_tank_armor_bulletdamage" },
    { 13947, "tryUseRemoteTank" },
    { 13948, "giveTank" },
    { 13949, "createTankForPlayer" },
    { 13950, "tankType" },
    { 13951, "setCarryingTank" },
    { 13952, "tank_setCarried" },
    { 13953, "updateTankPlacement" },
    { 13954, "tank_onCarrierDeath" },
    { 13955, "tank_onCarrierDisconnect" },
    { 13956, "tank_onGameEnded" },
    { 13957, "tank_setCancelled" },
    { 13958, "tank_setPlaced" },
    { 13959, "tank_giveWeaponOnPlaced" },
    { 13960, "createTank" },
    { 13961, "tank" },
    { 13962, "tank_setActive" },
    { 13963, "startUsingTank" },
    { 13964, "tank_blinkyLightAntenna" },
    { 13965, "tank_blinkyLightCamera" },
    { 13966, "tank_setInactive" },
    { 13967, "tank_freezeBuffer" },
    { 13968, "tank_handleDisconnect" },
    { 13969, "tank_handleChangeTeams" },
    { 13970, "tank_handleTimeout" },
    { 13971, "tank_handleDeath" },
    { 13972, "tank_incrementDamageFade" },
    { 13973, "tank_watchLowHealth" },
    { 13974, "tank_handleDamage" },
    { 13975, "tank_turret_handleDamage" },
    { 13976, "tank_WatchFiring" },
    { 13977, "tank_FireMissiles" },
    { 13978, "tank_Earthquake" },
    { 13979, "addToUGVList" },
    { 13980, "removeFromUGVList" },
    { 13981, "tank_playerExit" },
    { 13982, "killstreakWeapons" },
    { 13983, "killstreakWeildWeapons" },
    { 13984, "killstreakChainingWeapons" },
    { 13985, "killstreakRoundDelay" },
    { 13986, "initKillstreakData" },
    { 13987, "curDefValue" },
    { 13988, "spUpdateTotal" },
    { 13989, "earnedStreakLevel" },
    { 13990, "adrenaline" },
    { 13991, "initPlayerKillstreaks" },
    { 13992, "earned" },
    { 13993, "awardxp" },
    { 13994, "kID" },
    { 13995, "isGimme" },
    { 13996, "isSpecialist" },
    { 13997, "nextSlot" },
    { 13998, "updateStreakCount" },
    { 13999, "previousAdrenaline" },
    { 14000, "resetStreakCount" },
    { 14001, "getNextStreakName" },
    { 14002, "getMaxStreakCost" },
    { 14003, "updateStreakSlots" },
    { 14004, "killstreakIndexWeapon" },
    { 14005, "waitForChangeTeam" },
    { 14006, "isRideKillstreak" },
    { 14007, "isCarryKillstreak" },
    { 14008, "deadlyKillstreak" },
    { 14009, "killstreakUsePressed" },
    { 14010, "selectingLocation" },
    { 14011, "useHardpoint" },
    { 14012, "updateSpecialistKillstreaks" },
    { 14013, "getFirstPrimaryWeapon" },
    { 14014, "killstreakUseWaiter" },
    { 14015, "lastKillStreak" },
    { 14016, "waitTakeKillstreakWeapon" },
    { 14017, "shouldSwitchWeaponPostKillstreak" },
    { 14018, "finishDeathWaiter" },
    { 14019, "checkStreakReward" },
    { 14020, "killstreakEarned" },
    { 14021, "firstKillstreakEarned" },
    { 14022, "earnKillstreak" },
    { 14023, "isPerkUpgraded" },
    { 14024, "giveKillstreakWeapon" },
    { 14025, "_setActionSlot" },
    { 14026, "isAssaultKillstreak" },
    { 14027, "isSupportKillstreak" },
    { 14028, "isSpecialistKillstreak" },
    { 14029, "getKillstreakHint" },
    { 14030, "getKillstreakInformEnemy" },
    { 14031, "getKillstreakSound" },
    { 14032, "getKillstreakDialog" },
    { 14033, "getKillstreakIcon" },
    { 14034, "getKillstreakDpadIcon" },
    { 14035, "getKillstreakIndex" },
    { 14036, "giveOwnedKillstreakItem" },
    { 14037, "initRideKillstreak_internal" },
    { 14038, "isNuked" },
    { 14039, "giveSelectedKillstreakItem" },
    { 14040, "showSelectedStreakHint" },
    { 14041, "leaderSound" },
    { 14042, "leaderSoundGroup" },
    { 14043, "getKillstreakCount" },
    { 14044, "shuffleKillstreaksUp" },
    { 14045, "shuffleKillstreaksDown" },
    { 14046, "streakSelectUpTracker" },
    { 14047, "streakSelectDownTracker" },
    { 14048, "streakNotifyTracker" },
    { 14049, "adrenalineInfo" },
    { 14050, "giveAllPerks" },
    { 14051, "setAdrenaline" },
    { 14052, "notifyOverlay" },
    { 14053, "promotionSplashNotify" },
    { 14054, "weaponPromotionSplashNotify" },
    { 14055, "textGlowColor" },
    { 14056, "iconOverlay" },
    { 14058, "playerCardPlayer" },
    { 14059, "resetOnDeath" },
    { 14060, "resetNotify" },
    { 14061, "hintMessageDeathThink" },
    { 14062, "lowerMessageThink" },
    { 14063, "lowerMessages" },
    { 14064, "outcomeOverlay" },
    { 14065, "matchOutcomeNotify" },
    { 14066, "resetOutcomeNotify" },
    { 14067, "resetTeamOutcomeNotify" },
    { 14068, "updateOutcome" },
    { 14069, "canShowSplash" },
    { 14070, "missionCallbacks" },
    { 14071, "createPerkMap" },
    { 14072, "perkMap" },
    { 14073, "mayProcessChallenges" },
    { 14074, "patientZeroName" },
    { 14075, "infected" },
    { 14076, "plague" },
    { 14077, "monitorScavengerPickup" },
    { 14078, "monitorStreakReward" },
    { 14079, "monitorBlastShieldSurvival" },
    { 14080, "monitorTacInsertionsDestroyed" },
    { 14081, "monitorFinalStandSurvival" },
    { 14082, "initMissionData" },
    { 14083, "registerMissionCallback" },
    { 14084, "ch_assists" },
    { 14085, "ch_hardpoints" },
    { 14086, "hardpointType" },
    { 14087, "ch_vehicle_kills" },
    { 14088, "victim" },
    { 14089, "ch_vehicle_killed" },
    { 14090, "clearIDShortly" },
    { 14091, "explosiveKills" },
    { 14092, "MGKill" },
    { 14093, "endMGStreakWhenLeaveMG" },
    { 14094, "endMGStreak" },
    { 14095, "killedBestEnemyPlayer" },
    { 14096, "isHighestScoringPlayer" },
    { 14097, "ch_kills" },
    { 14098, "isCACSecondaryWeapon" },
    { 14099, "brinkOfDeathKillStreak" },
    { 14100, "isStrStart" },
    { 14101, "getWeaponAttachments" },
    { 14102, "anglesOnDeath" },
    { 14103, "anglesOnKill" },
    { 14104, "isBuffUnlockedForWeapon" },
    { 14105, "holdingBreath" },
    { 14106, "adsTime" },
    { 14107, "isPlanting" },
    { 14108, "isDefusing" },
    { 14109, "isBombCarrier" },
    { 14110, "dd" },
    { 14111, "lastPrimaryWeaponSwapTime" },
    { 14112, "lastFlashedTime" },
    { 14113, "lastConcussedTime" },
    { 14114, "lastSprintEndTime" },
    { 14115, "ch_bulletDamageCommon" },
    { 14116, "victimOnGround" },
    { 14117, "attackerOnGround" },
    { 14118, "attackerStance" },
    { 14119, "ch_roundplayed" },
    { 14120, "place" },
    { 14121, "ch_roundwin" },
    { 14122, "attackerInLastStand" },
    { 14123, "waitAndProcessPlayerKilledCallback" },
    { 14124, "processingKilledChallenges" },
    { 14125, "doMissionCallback" },
    { 14126, "monitorSprintDistance" },
    { 14127, "sprintDistThisSprint" },
    { 14128, "monitorSingleSprintDistance" },
    { 14129, "monitorSprintTime" },
    { 14130, "monitorFallDistance" },
    { 14131, "lastManSD" },
    { 14132, "monitorBombUse" },
    { 14133, "monitorLiveTime" },
    { 14134, "survivalistChallenge" },
    { 14135, "monitorStreaks" },
    { 14136, "monitorMisc" },
    { 14137, "monitorMiscSingle" },
    { 14138, "monitorMiscCallback" },
    { 14139, "healthRegenerationStreak" },
    { 14140, "resetBrinkOfDeathKillStreakShortly" },
    { 14141, "playerDied" },
    { 14142, "isAtBrinkOfDeath" },
    { 14143, "getMarksmanUnlockAttachment" },
    { 14144, "getWeaponAttachment" },
    { 14145, "masteryChallengeProcess" },
    { 14146, "isAttachmentUnlocked" },
    { 14147, "buildChallegeInfo" },
    { 14148, "monitorProcessChallenge" },
    { 14149, "monitorKillstreakProgress" },
    { 14150, "monitorKilledKillstreak" },
    { 14151, "playerHasAmmo" },
    { 14152, "monitorADSTime" },
    { 14153, "monitorHoldBreath" },
    { 14154, "monitorMantle" },
    { 14155, "mantling" },
    { 14156, "monitorWeaponSwap" },
    { 14157, "monitorFlashbang" },
    { 14158, "monitorConcussion" },
    { 14159, "monitorMineTriggering" },
    { 14160, "waitDelayMineTime" },
    { 14161, "weaponRankTable" },
    { 14162, "maxPrestige" },
    { 14163, "patientZeroWaiter" },
    { 14164, "isRegisteredEvent" },
    { 14165, "getScoreInfoLabel" },
    { 14166, "getWeaponRankInfoMinXP" },
    { 14167, "getWeaponRankInfoXPAmt" },
    { 14168, "getWeaponRankInfoMaxXp" },
    { 14169, "getRankInfoLevel" },
    { 14170, "xpUpdateTotal" },
    { 14171, "bonusUpdateTotal" },
    { 14172, "hud_xpPointsPopup" },
    { 14173, "hud_xpEventPopup" },
    { 14174, "prestigeDoubleXp" },
    { 14175, "prestigeDoubleWeaponXp" },
    { 14176, "setGamesPlayed" },
    { 14177, "roundUp" },
    { 14178, "weaponShouldGetXP" },
    { 14179, "levelFlag" },
    { 14180, "updateWeaponRankAnnounceHUD" },
    { 14181, "createXpPointsPopup" },
    { 14182, "createXpEventPopup" },
    { 14183, "removeRankHUD" },
    { 14184, "levelForExperience" },
    { 14185, "weaponLevelForExperience" },
    { 14186, "getCurrentWeaponXP" },
    { 14187, "getWeaponRankForXp" },
    { 14188, "getPrestigeLevel" },
    { 14189, "getWeaponRankXP" },
    { 14190, "getWeaponMaxRankXP" },
    { 14191, "isWeaponMaxRank" },
    { 14192, "incRankXP" },
    { 14193, "isCheater" },
    { 14194, "getRestXPAward" },
    { 14195, "isLastRestXPAward" },
    { 14196, "syncXPStat" },
    { 14197, "isWeaponChallenge" },
    { 14198, "isValidPrimary" },
    { 14199, "isValidSecondary" },
    { 14200, "classMap" },
    { 14201, "defaultClass" },
    { 14202, "logClassChoice" },
    { 14203, "cac_getWeaponCamo" },
    { 14204, "cac_getWeaponReticle" },
    { 14205, "recipe_getKillstreak" },
    { 14206, "table_getWeaponCamo" },
    { 14207, "table_getWeaponReticle" },
    { 14208, "table_getTeamPerk" },
    { 14209, "loadoutFakePerks" },
    { 14210, "getLoadoutStreakTypeFromStreakType" },
    { 14211, "recipeClassApplyJuggernaut" },
    { 14212, "classTableNum" },
    { 14213, "loadoutPrimaryCamo" },
    { 14214, "loadoutSecondaryCamo" },
    { 14215, "loadoutPrimaryReticle" },
    { 14216, "loadoutSecondaryReticle" },
    { 14217, "_clearPerks" },
    { 14218, "_detachAll" },
    { 14219, "loadoutAllPerks" },
    { 14220, "loadoutPerk1" },
    { 14221, "loadoutPerk2" },
    { 14222, "loadoutPerk3" },
    { 14223, "loadoutPerkEquipment" },
    { 14224, "tryAttach" },
    { 14225, "tryDetach" },
    { 14226, "buildWeaponName" },
    { 14227, "letterToNumber" },
    { 14228, "getAttachmentType" },
    { 14229, "buildWeaponNameCamo" },
    { 14230, "buildWeaponNameReticle" },
    { 14231, "makeLettersToNumbers" },
    { 14232, "setKillstreaks" },
    { 14233, "replenishLoadout" },
    { 14234, "classGrenades" },
    { 14235, "onPlayerConnecting" },
    { 14236, "fadeAway" },
    { 14237, "getPerkForClass" },
    { 14238, "classHasPerk" },
    { 14239, "isValidAttachment" },
    { 14240, "isValidWeaponBuff" },
    { 14241, "isWeaponBuffUnlocked" },
    { 14242, "isValidCamo" },
    { 14243, "isValidReticle" },
    { 14244, "isCamoUnlocked" },
    { 14245, "isValidEquipment" },
    { 14246, "isValidOffhand" },
    { 14247, "isValidPerk1" },
    { 14248, "isValidPerk2" },
    { 14249, "isValidPerk3" },
    { 14250, "isValidDeathStreak" },
    { 14251, "isValidWeapon" },
    { 14252, "weaponRefs" },
    { 14253, "isValidKillstreak" },
    { 14254, "persistentDataInfo" },
    { 14255, "bufferedStats" },
    { 14256, "bufferedChildStats" },
    { 14257, "statAddChild" },
    { 14258, "statGetChildBuffered" },
    { 14259, "updateBufferedStats" },
    { 14260, "writeBufferedStats" },
    { 14261, "uploadGlobalStatCounters" },
    { 14262, "exploder_sound" },
    { 14263, "streakMsg" },
    { 14264, "endSelectionOnEMP" },
    { 14265, "endSelectionOnAction" },
    { 14266, "endSelectionOnEndGame" },
    { 14267, "getPlant" },
    { 14268, "orientToNormal" },
    { 14269, "deletePlacedEntity" },
    { 14270, "sortLowerMessages" },
    { 14271, "addLowerMessage" },
    { 14272, "addTime" },
    { 14273, "showTimer" },
    { 14274, "shouldFade" },
    { 14275, "fadeToAlpha" },
    { 14276, "fadeToAlphaTime" },
    { 14277, "removeLowerMessage" },
    { 14278, "getLowerMessage" },
    { 14279, "updateLowerMessage" },
    { 14280, "clearOnDeath" },
    { 14281, "clearAfterFade" },
    { 14282, "printOnTeam" },
    { 14283, "printBoldOnTeam" },
    { 14284, "printBoldOnTeamArg" },
    { 14285, "printOnTeamArg" },
    { 14286, "printOnPlayers" },
    { 14287, "printAndSoundOnEveryone" },
    { 14288, "printAndSoundOnTeam" },
    { 14289, "printAndSoundOnPlayer" },
    { 14290, "_playLocalSound" },
    { 14291, "dvarIntValue" },
    { 14292, "dvarFloatValue" },
    { 14293, "updatePersRatioBuffered" },
    { 14294, "lastSlowProcessFrame" },
    { 14295, "isExcluded" },
    { 14296, "leaderDialogBothTeams" },
    { 14297, "playLeaderDialogOnPlayer" },
    { 14298, "leaderDialogLocalSound" },
    { 14299, "setObjectiveText" },
    { 14300, "setObjectiveScoreText" },
    { 14301, "setObjectiveHintText" },
    { 14302, "getObjectiveText" },
    { 14303, "getObjectiveScoreText" },
    { 14304, "getMinutesPassed" },
    { 14305, "getValueInRange" },
    { 14306, "waitForTimeOrNotifies" },
    { 14307, "logXPGains" },
    { 14308, "registerRoundSwitchDvar" },
    { 14309, "roundswitchDvar" },
    { 14310, "roundswitchMin" },
    { 14311, "roundswitchMax" },
    { 14312, "registerRoundLimitDvar" },
    { 14313, "registerWinLimitDvar" },
    { 14314, "registerScoreLimitDvar" },
    { 14315, "registerTimeLimitDvar" },
    { 14316, "registerHalfTimeDvar" },
    { 14317, "registerNumLivesDvar" },
    { 14318, "setOverTimeLimitDvar" },
    { 14319, "getDvarVec" },
    { 14320, "_takeWeaponsExcept" },
    { 14321, "saveData" },
    { 14322, "offhandClass" },
    { 14323, "actionSlots" },
    { 14324, "weapons" },
    { 14325, "clipAmmoR" },
    { 14326, "clipAmmoL" },
    { 14327, "stockAmmo" },
    { 14328, "script_saveData" },
    { 14329, "restoreData" },
    { 14330, "isFloat" },
    { 14331, "registerWatchDvarInt" },
    { 14332, "registerWatchDvarFloat" },
    { 14333, "registerWatchDvar" },
    { 14334, "setOverrideWatchDvar" },
    { 14335, "onlyRoundOverride" },
    { 14336, "hitScoreLimit" },
    { 14337, "getRoundsWon" },
    { 14338, "objectiveBased" },
    { 14339, "bombexploded" },
    { 14340, "ddTimeToAdd" },
    { 14341, "inOvertime" },
    { 14342, "getAverageOrigin" },
    { 14343, "getLivingPlayers" },
    { 14344, "getRemoteName" },
    { 14345, "queues" },
    { 14346, "_setPerk" },
    { 14347, "_setExtraPerks" },
    { 14348, "quickSort" },
    { 14349, "quickSortMid" },
    { 14350, "swap" },
    { 14351, "onlinegame" },
    { 14352, "endSceneOnDeath" },
    { 14353, "giveCombatHigh" },
    { 14354, "arrayInsertion" },
    { 14355, "getProperty" },
    { 14356, "statusMenu" },
    { 14357, "_statusMenu" }, // a typo by IW should be statusMenu
    { 14358, "streakShouldChain" },
    { 14359, "fixAkimboString" },
    { 14360, "roundDecimalPlaces" },
    { 14361, "makeTeamUsable" },
    { 14362, "_updateTeamUsable" },
    { 14363, "_updateEnemyUsable" },
    { 14364, "gameFlagClear" },
    { 14365, "isPrimaryDamage" },
    { 14366, "levelFlags" },
    { 14367, "levelFlagWait" },
    { 14368, "levelFlagWaitOpen" },
    { 14369, "killTrigger" },
    { 14370, "setCommonRulesFromMatchRulesData" },
    { 14371, "matchRules_allowCustomClasses" },
    { 14372, "GetMatchRulesSpecialClass" },
    { 14374, "audio" },
    { 14375, "init_reverb" },
    { 14376, "add_reverb" },
    { 14377, "reverb_settings" },
    { 14378, "is_roomtype_valid" },
    { 14379, "apply_reverb" },
    { 14380, "init_whizby" },
    { 14381, "whizby_settings" },
    { 14382, "set_whizby_radius" },
    { 14383, "set_whizby_spread" },
    { 14384, "apply_whizby" },
    { 14385, "radial_button_definitions" },
    { 14386, "radial_init" },
    { 14387, "radial_button_group" },
    { 14388, "pos_angle" },
    { 14389, "end_angle" },
    { 14390, "start_angle" },
    { 14391, "debug_toggle" },
    { 14392, "crib_debug" },
    { 14393, "observer" },
    { 14394, "return_hud" },
    { 14395, "readyPlayer" },
    { 14396, "get_right_stick_angle" },
    { 14397, "rs_angle" },
    { 14398, "newRadialButtonGroup" },
    { 14399, "radial_button_group_info" },
    { 14400, "newRadialButton" },
    { 14401, "font_size" },
    { 14402, "font_color" },
    { 14403, "action_func" },
    { 14404, "radius_pos" },
    { 14405, "updateSelectedButton" },
    { 14406, "radial_button_current_group" },
    { 14407, "active_button" },
    { 14408, "watchSelectButtonPress" },
    { 14409, "watchBackButtonPress" },
    { 14410, "sort_buttons_by_angle" },
    { 14411, "button_switch" },
    { 14412, "draw_radial_buttons" },
    { 14413, "draw_radial_button" },
    { 14414, "zoom_to_radial_menu" },
    { 14415, "radial_button_previous_group" },
    { 14416, "getRadialAngleFromEnt" },
    { 14417, "radial_angle_to_vector" },
    { 14418, "getMidAngle" },
    { 14419, "isInRange" },
    { 14420, "action_back" },
    { 14421, "action_weapons_primary" },
    { 14422, "action_weapons_secondary" },
    { 14423, "action_gears" },
    { 14424, "action_killstreak" },
    { 14425, "action_leaderboards" },
    { 14426, "view_path_setup" },
    { 14427, "view_paths" },
    { 14428, "build_path_by_targetname" },
    { 14429, "go_path_by_targetname" },
    { 14430, "dummy_mover" },
    { 14431, "go_path_by_targetname_reverse" },
    { 14432, "travel_view_fx" },
    { 14433, "blur_sine" },
    { 14434, "force_player_angles" },
    { 14435, "shotgunSetup" },
    { 14436, "tryUseAutoShotgun" },
    { 14437, "thumperSetup" },
    { 14438, "tryUseThumper" },
    { 14439, "saveWeaponAmmoOnDeath" },
    { 14440, "removeWeaponOnOutOfAmmo" },
    { 14441, "spawnArmor" },
    { 14442, "standardSpeed" },
    { 14443, "deleteOnZ" },
    { 14444, "useTank" },
    { 14445, "tryUseTank" },
    { 14446, "tankInUse" },
    { 14447, "tankSpawner" },
    { 14448, "startTank" },
    { 14449, "nodes" },
    { 14450, "objID" },
    { 14451, "timeLastFired" },
    { 14452, "neutralTarget" },
    { 14453, "waitForChangeTeams" },
    { 14454, "waitForDisco" },
    { 14455, "setDirection" },
    { 14456, "setEngagementSpeed" },
    { 14457, "changingDirection" },
    { 14458, "speedType" },
    { 14459, "setMiniEngagementSpeed" },
    { 14460, "setStandardSpeed" },
    { 14461, "setEvadeSpeed" },
    { 14462, "setDangerSpeed" },
    { 14463, "stopToReverse" },
    { 14464, "stopToForward" },
    { 14465, "checkDanger" },
    { 14466, "numEnemiesClose" },
    { 14467, "tankUpdate" },
    { 14468, "graphNodes" },
    { 14469, "endNode" },
    { 14470, "tankDamageMonitor" },
    { 14471, "forcedTarget" },
    { 14472, "handleThreat" },
    { 14473, "tankCover" },
    { 14474, "handlePossibleThreat" },
    { 14475, "relativeAngle" },
    { 14476, "watchForThreat" },
    { 14477, "checkOwner" },
    { 14478, "modifyDamage" },
    { 14479, "destroyTank" },
    { 14480, "tankFire" },
    { 14481, "onHitPitchClamp" },
    { 14482, "waitForTurretReady" },
    { 14483, "tankGetTargets" },
    { 14484, "harrier" },
    { 14485, "acquireTarget" },
    { 14486, "setNoTarget" },
    { 14487, "explicitAbandonTarget" },
    { 14488, "badTarget" },
    { 14489, "badTargetReset" },
    { 14490, "removeTarget" },
    { 14491, "lastLostTime" },
    { 14492, "isVehicleTarget" },
    { 14493, "turretSightTrace" },
    { 14494, "isMiniTarget" },
    { 14495, "tankGetMiniTargets" },
    { 14496, "getBestMiniTarget" },
    { 14497, "acquireMiniTarget" },
    { 14498, "bestMiniTarget" },
    { 14499, "fireMiniOnTarget" },
    { 14500, "watchMiniTargetDeath" },
    { 14501, "watchMiniTargetDistance" },
    { 14502, "watchMiniTargetThreat" },
    { 14503, "explicitAbandonMiniTarget" },
    { 14504, "addToTankList" },
    { 14505, "removeFromTankList" },
    { 14506, "getNodeNearEnemies" },
    { 14507, "dist" },
    { 14508, "setupPaths" },
    { 14509, "branchNodes" },
    { 14510, "next" },
    { 14511, "length" },
    { 14512, "graphId" },
    { 14513, "getRandomBranchNode" },
    { 14514, "links" },
    { 14515, "linkDirs" },
    { 14516, "getNextNodeForEndNode" },
    { 14517, "g" },
    { 14518, "otherDir" },
    { 14519, "handleBranchNode" },
    { 14520, "linkStartNodes" },
    { 14521, "handleCapNode" },
    { 14522, "nodeTracker" },
    { 14523, "forwardGraphId" },
    { 14524, "reverseGraphId" },
    { 14525, "forceTrigger" },
    { 14526, "getForwardGraphNode" },
    { 14527, "getReverseGraphNode" },
    { 14528, "getNextNode" },
    { 14529, "getPrevNode" },
    { 14530, "initNodeGraph" },
    { 14531, "linkInfos" },
    { 14532, "linkLengths" },
    { 14533, "addLinkNode" },
    { 14534, "toGraphNode" },
    { 14535, "toGraphId" },
    { 14536, "direction" },
    { 14537, "startNode" },
    { 14538, "generatePath" },
    { 14539, "openList" },
    { 14540, "closedList" },
    { 14541, "h" },
    { 14542, "f" },
    { 14543, "parentNode" },
    { 14544, "addToOpenList" },
    { 14545, "openListID" },
    { 14546, "closedListID" },
    { 14547, "addToClosedList" },
    { 14548, "getHValue" },
    { 14549, "getGValue" },
    { 14550, "drawPath" },
    { 14551, "drawGraph" },
    { 14552, "drawLink" },
    { 14553, "debugPrintLn2" },
    { 14554, "debugPrint3D" },
    { 14555, "drawTankGraphIds" },
    { 14556, "tankExplode" },
    { 14557, "tankFlash" },
    { 14558, "tankDust1" },
    { 14559, "tankDust2" },
    { 14560, "ground_support_locs" },
    { 14561, "tryUseMobileMortar" },
    { 14562, "mobileMortar" },
    { 14563, "selectEntranceLocation" },
    { 14564, "createMobileMortar" },
    { 14565, "playersAttacked" },
    { 14566, "lastTarget" },
    { 14567, "lowX" },
    { 14568, "highX" },
    { 14569, "lowY" },
    { 14570, "highY" },
    { 14571, "moveToPosition" },
    { 14572, "fxEnt" },
    { 14573, "findTarget" },
    { 14574, "findRandomTarget" },
    { 14575, "mortarAttack" },
    { 14576, "fireMortar" },
    { 14577, "watchProjectileOnMiniMap" },
    { 14578, "mortarRecoil" },
    { 14579, "watchProximity" },
    { 14580, "watchDamage" },
    { 14581, "a10_fx" },
    { 14582, "a10MaxHealth" },
    { 14583, "a10Speed" },
    { 14584, "a10SpeedReduction" },
    { 14585, "a10StartPointOffset" },
    { 14586, "a10ImpactFXDelay" },
    { 14587, "a10Damage" },
    { 14588, "a10DamageRadius" },
    { 14589, "a10DamageDelay" },
    { 14590, "a10BulletRainDelay" },
    { 14591, "a10BulletImpactsDelay" },
    { 14592, "a10EarthquakeMagnitude" },
    { 14593, "a10EarthquakeDuration" },
    { 14594, "a10EarthquakeDelay" },
    { 14595, "a10DirtEffectRadius" },
    { 14596, "a10ShootingGroundSoundDelay" },
    { 14597, "a10StartPositionScalar" },
    { 14598, "a10SupportSetup" },
    { 14599, "usedUavA10" },
    { 14600, "tryUseA10Strike" },
    { 14601, "selectA10StrikeLocation" },
    { 14602, "finishA10StrikeUsage" },
    { 14603, "callA10Strike" },
    { 14604, "doA10Strike" },
    { 14605, "a10" },
    { 14606, "a10StartMove" },
    { 14607, "initialDelay" },
    { 14608, "startPoint" },
    { 14609, "attackPoint" },
    { 14610, "endPoint" },
    { 14611, "a10PlayEngineFX" },
    { 14612, "spawnA10" },
    { 14613, "fakeA10" },
    { 14614, "startA10Shooting" },
    { 14615, "a10ShootingPos" },
    { 14616, "playBulletRain" },
    { 14617, "manageShootingLoopSound" },
    { 14618, "manageShootingGroundSound" },
    { 14619, "a10Earthquake" },
    { 14620, "a10Destroyed" },
    { 14621, "a10Explode" },
    { 14622, "removeA10" },
    { 14623, "tryUseTeamAmmoRefill" },
    { 14624, "giveTeamAmmoRefill" },
    { 14625, "showSpawnpoint" },
    { 14626, "showSpawnpoints" },
    { 14627, "print3DUntilNotified" },
    { 14628, "lineUntilNotified" },
    { 14629, "hideSpawnpoints" },
    { 14630, "updateReflectionProbe" },
    { 14631, "reflectionProbeButtons" },
    { 14632, "gotoNextSpawn" },
    { 14633, "gotoPrevSpawn" },
    { 14634, "endGameOnTimeLimit" },
    { 14635, "playersLookingForSafeSpawn" },
    { 14636, "registerDvars" },
    { 14637, "blank" },
    { 14638, "testMenu" },
    { 14639, "testShock" },
    { 14640, "fakeLag" },
    { 14641, "spawn_all" },
    { 14642, "spawn_axis_start" },
    { 14643, "spawn_allies_start" },
    { 14644, "flagBaseFXid" },
    { 14645, "bestSpawnFlag" },
    { 14646, "nearbyspawns" },
    { 14647, "domFlags" },
    { 14648, "lastStatus" },
    { 14649, "baseeffectforward" },
    { 14650, "baseeffectright" },
    { 14651, "baseeffectpos" },
    { 14652, "useObj" },
    { 14653, "adjflags" },
    { 14654, "getUnownedFlagNearestStart" },
    { 14655, "didStatusNotify" },
    { 14656, "statusDialog" },
    { 14657, "resetFlagBaseEffect" },
    { 14658, "baseeffect" },
    { 14659, "captureTime" },
    { 14660, "giveFlagCaptureXP" },
    { 14661, "delayedLeaderDialog" },
    { 14662, "delayedLeaderDialogBothTeams" },
    { 14663, "updateDomScores" },
    { 14664, "getOwnedDomFlags" },
    { 14665, "getTeamFlagCount" },
    { 14666, "getFlagTeam" },
    { 14667, "getBoundaryFlags" },
    { 14668, "getBoundaryFlagSpawns" },
    { 14669, "getSpawnsBoundingFlag" },
    { 14670, "getOwnedAndBoundingFlagSpawns" },
    { 14671, "getOwnedFlagSpawns" },
    { 14672, "flagSetup" },
    { 14673, "descriptor" },
    { 14674, "updateCPM" },
    { 14675, "CPM" },
    { 14676, "numCaps" },
    { 14677, "getCapXPScale" },
    { 14678, "multiBomb" },
    { 14679, "bombPlanted" },
    { 14680, "sd_loadout" },
    { 14681, "checkAllowSpectating" },
    { 14682, "sd_endGame" },
    { 14683, "bombZones" },
    { 14684, "killCamEntNum" },
    { 14685, "bombDefused" },
    { 14686, "plantTime" },
    { 14687, "defuseTime" },
    { 14688, "removeBombZoneC" },
    { 14689, "relatedBrushModel" },
    { 14690, "bombs" },
    { 14691, "sdBomb" },
    { 14692, "exploderIndex" },
    { 14693, "bombDefuseTrig" },
    { 14694, "otherBombZones" },
    { 14695, "setupKillCamEnt" },
    { 14696, "sdBombModel" },
    { 14697, "onUsePlantObject" },
    { 14698, "bombOwner" },
    { 14699, "bombPlantedTime" },
    { 14700, "applyBombCarrierClass" },
    { 14701, "removeBombCarrierClass" },
    { 14702, "onUseDefuseObject" },
    { 14703, "tickingObject" },
    { 14704, "BombTimerWait" },
    { 14705, "handleHostMigration" },
    { 14706, "setSpecialLoadout" },
    { 14707, "spawn_axis" },
    { 14708, "spawn_axis_planted" },
    { 14709, "spawn_allies" },
    { 14710, "spawn_allies_planted" },
    { 14711, "otSpawned" },
    { 14712, "sab_loadouts" },
    { 14713, "printOTHint" },
    { 14714, "hotPotato" },
    { 14715, "scoreMode" },
    { 14716, "sabotage" },
    { 14717, "sabBomb" },
    { 14718, "getClosestSite" },
    { 14719, "distanceToSite" },
    { 14720, "scoreThread" },
    { 14721, "bombDistance" },
    { 14722, "createBombZone" },
    { 14723, "abandonmentThink" },
    { 14724, "overtimeThread" },
    { 14725, "bombDistanceThread" },
    { 14726, "dangerTeam" },
    { 14727, "resetBombsite" },
    { 14728, "setUpForDefusing" },
    { 14729, "setSpecialLoadouts" },
    { 14730, "doPrematch" },
    { 14731, "hqAutoDestroyTime" },
    { 14732, "hqSpawnTime" },
    { 14733, "kothMode" },
    { 14734, "destroyTime" },
    { 14735, "delayPlayer" },
    { 14736, "spawnDelay" },
    { 14737, "extraDelay" },
    { 14738, "proMode" },
    { 14739, "iconoffset" },
    { 14740, "updateObjectiveHintMessages" },
    { 14741, "getRespawnDelay" },
    { 14742, "radioObject" },
    { 14743, "hqDestroyTime" },
    { 14744, "objectiveHintPrepareHQ" },
    { 14745, "objectiveHintCaptureHQ" },
    { 14746, "objectiveHintDestroyHQ" },
    { 14747, "objectiveHintDefendHQ" },
    { 14748, "HQMainLoop" },
    { 14749, "hqRevealTime" },
    { 14750, "timerDisplay" },
    { 14751, "gameobject" },
    { 14752, "trigorigin" },
    { 14753, "hqDestroyedByTimer" },
    { 14754, "hideTimerDisplayOnGameEnd" },
    { 14755, "forceSpawnTeam" },
    { 14756, "onRadioCapture" },
    { 14757, "onRadioDestroy" },
    { 14758, "DestroyHQAfterTime" },
    { 14759, "awardHQPoints" },
    { 14760, "nearSpawns" },
    { 14761, "outerSpawns" },
    { 14762, "SetupRadios" },
    { 14763, "trig" },
    { 14764, "radios" },
    { 14765, "prevradio" },
    { 14766, "prevradio2" },
    { 14767, "makeRadioActive" },
    { 14768, "makeRadioInactive" },
    { 14769, "setUpNearbySpawns" },
    { 14770, "distsq" },
    { 14771, "PickRadioToSpawn" },
    { 14772, "flagReturnTime" },
    { 14773, "usingObj" },
    { 14774, "oneflag_ctf" },
    { 14775, "flagModel" },
    { 14776, "icon2D" },
    { 14777, "iconCapture3D" },
    { 14778, "iconCapture2D" },
    { 14779, "iconDefend3D" },
    { 14780, "iconDefend2D" },
    { 14781, "iconTarget3D" },
    { 14782, "iconTarget2D" },
    { 14783, "teamFlags" },
    { 14784, "capZones" },
    { 14785, "flagCaptured" },
    { 14786, "createTeamFlag" },
    { 14787, "createCapZone" },
    { 14788, "returnFlag" },
    { 14789, "attachFlag" },
    { 14790, "detachFlag" },
    { 14791, "ctf_loadouts" },
    { 14792, "ctf" },
    { 14793, "iconEscort3D" },
    { 14794, "iconEscort2D" },
    { 14795, "iconKill3D" },
    { 14796, "iconKill2D" },
    { 14797, "iconCaptureFlag3D" },
    { 14798, "iconCaptureFlag2D" },
    { 14799, "iconDefendFlag3D" },
    { 14800, "iconDefendFlag2D" },
    { 14801, "iconReturnFlag3D" },
    { 14802, "iconReturnFlag2D" },
    { 14803, "iconWaitForFlag3D" },
    { 14804, "iconWaitForFlag2D" },
    { 14805, "friendlyFlagStatusIcon" },
    { 14806, "friendlyFlagStatusText" },
    { 14807, "enemyFlagStatusIcon" },
    { 14808, "enemyFlagStatusText" },
    { 14809, "returnAfterTime" },
    { 14810, "applyFlagCarrierClass" },
    { 14811, "removeFlagCarrierClass" },
    { 14812, "ctfPro" },
    { 14813, "iconFlagBase2D" },
    { 14814, "iconFlagBase3D" },
    { 14815, "createTeamFlags" },
    { 14816, "atHome" },
    { 14817, "createCapZones" },
    { 14818, "zoneHeadIcon" },
    { 14819, "zoneMapIcon" },
    { 14820, "cappedFlag" },
    { 14821, "giveScore" },
    { 14822, "matchRules_initialAmmo" },
    { 14823, "matchRules_rewardAmmo" },
    { 14824, "oic_firstSpawn" },
    { 14825, "oic_rewardAmmo" },
    { 14826, "oic_loadouts" },
    { 14827, "waitLoadoutDone" },
    { 14828, "oic_gun" },
    { 14829, "waitGiveAmmo" },
    { 14830, "giveAmmo" },
    { 14831, "watchElimination" },
    { 14832, "setGun" },
    { 14833, "matchRules_dropTime" },
    { 14834, "matchRules_zoneSwitchTime" },
    { 14835, "grnd_fx" },
    { 14836, "grnd_centerLoc" },
    { 14838, "initFirstZone" },
    { 14839, "zonesCycling" },
    { 14840, "grnd_dropZones" },
    { 14841, "grnd_initialIndex" },
    { 14842, "grnd_zone" },
    { 14843, "initZones" },
    { 14844, "grnd_zones" },
    { 14845, "inGrindZone" },
    { 14846, "setPlayerMessages" },
    { 14847, "inGrindZonePoints" },
    { 14849, "grndHeadIcon" },
    { 14850, "grndObjId" },
    { 14851, "getNextZone" },
    { 14852, "distToZone" },
    { 14853, "cycleZones" },
    { 14854, "grndTracking" },
    { 14855, "locationScoring" },
    { 14856, "randomDrops" },
    { 14857, "getBestPlayer" },
    { 14858, "getDropZoneCrateType" },
    { 14859, "hideHudElementOnGameEnd" },
    { 14860, "createZones" },
    { 14861, "matchRules_enemyFlagRadar" },
    { 14862, "tdef" },
    { 14863, "tdef_loadouts" },
    { 14864, "tdef_flagTime" },
    { 14865, "portable_radar" },
    { 14866, "currentCarrier" },
    { 14867, "currentTeam" },
    { 14868, "watchForEndGame" },
    { 14869, "createFlag" },
    { 14870, "flagAttachRadar" },
    { 14871, "getFlagRadarOwner" },
    { 14872, "flagRadarMover" },
    { 14873, "flagWatchRadarOwnerLost" },
    { 14874, "conf_fx" },
    { 14875, "dogtags" },
    { 14876, "spawnDogTags" },
    { 14877, "victimTeam" },
    { 14878, "showToTeam" },
    { 14879, "bounce" },
    { 14880, "clearOnVictimDisconnect" },
    { 14881, "matchRules_numInitialInfected" },
    { 14882, "infect_timerDisplay" },
    { 14883, "infect_choseFirstInfected" },
    { 14884, "infect_choosingFirstInfected" },
    { 14885, "infect_firstSpawn" },
    { 14886, "isInitialInfected" },
    { 14887, "infect_loadouts" },
    { 14888, "chooseFirstInfected" },
    { 14889, "infect_isBeingChosen" },
    { 14890, "setInitialToNormalInfected" },
    { 14891, "updateTeamScores" },
    { 14894, "matchRules_respawnNewJugg" },
    { 14895, "matchRules_showJuggWorldIcon" },
    { 14896, "jugg_juggernaut" },
    { 14897, "jugg_choosingJugg" },
    { 14898, "jugg_timerDisplay" },
    { 14899, "chooseInitialJugg" },
    { 14900, "jugg_juggScore" },
    { 14901, "jugg_firstSpawn" },
    { 14902, "jugg_loadouts" },
    { 14903, "jugg_headIcon" },
    { 14904, "resetJugg" },
    { 14905, "giveJuggLoadout" },
    { 14906, "updateJuggScores" },
    { 14907, "gun_firstSpawn" },
    { 14908, "gunGameGunIndex" },
    { 14909, "gunGamePrevGunIndex" },
    { 14910, "gun_loadouts" },
    { 14911, "giveNextGun" },
    { 14912, "gun_guns" },
    { 14914, "refillSingleCountAmmo" },
    { 14915, "initGunHUD" },
    { 14917, "updateGunHUD" },
    { 14919, "hideOnGameEnd" },
    { 14920, "setGuns" },
    { 14921, "matchRules_juggSwitchTime" },
    { 14922, "jugg_available" },
    { 14923, "jugg_attackers" },
    { 14924, "jugg_currJugg" },
    { 14925, "tjugg_timerDisplay" },
    { 14926, "jugg_alligience" },
    { 14927, "isJuggModeJuggernaut" },
    { 14928, "tjugg_loadouts" },
    { 14929, "nextJuggTimeout" },
    { 14930, "respawnOldJugg" },
    { 14931, "rewardTeammateProximity" },
    { 14932, "logAttackers" },
    { 14933, "resetJuggLoadoutOnDisconnect" },
    { 14934, "getBestTeammate" },
    { 14935, "precacheFlag" },
    { 14936, "arenaTimeFlagWaiter" },
    { 14937, "arenaFlagWaiter" },
    { 14938, "arenaFlag" },
    { 14939, "setupDomFlag" },
    { 14940, "arena_endGame" },
    { 14941, "bombsPlanted" },
    { 14942, "ddBombModel" },
    { 14943, "spawn_defenders" },
    { 14944, "spawn_defenders_a" },
    { 14945, "spawn_defenders_b" },
    { 14946, "spawn_attackers" },
    { 14947, "spawn_attackers_a" },
    { 14948, "spawn_attackers_b" },
    { 14949, "spawn_defenders_start" },
    { 14950, "spawn_attackers_start" },
    { 14951, "aPlanted" },
    { 14952, "bPlanted" },
    { 14953, "waitToProcess" },
    { 14954, "dd_endGame" },
    { 14955, "verifyBombzones" },
    { 14956, "ddBomb" },
    { 14957, "onUseObject" },
    { 14958, "resetBombZone" },
    { 14959, "defusing" },
    { 14960, "timePauseStart" },
    { 14961, "destroyedObject" },
    { 14962, "bombHandler" },
    { 14963, "playDemolitionTickingSound" },
    { 14964, "waitTime" },
    { 14965, "setBombTimerDvar" },
    { 14966, "dropBombModel" },
    { 14967, "restartTimer" },
    { 14968, "skipWait" },
    { 14969, "isVip" },
    { 14970, "vip_endGame" },
    { 14971, "vipSelection" },
    { 14972, "setupVip" },
    { 14973, "extractionZone" },
    { 14974, "setVIPUse" },
    { 14975, "handleTimer" },
    { 14976, "extractionTime" },
    { 14977, "forceVIPSpawn" },
    { 14978, "gtnw_endGame" },
    { 14979, "useBar" },
    { 14980, "useBarText" },
    { 14981, "setupNukeSite" },
    { 14982, "nukeSite" },
    { 14983, "endGameTime" },
    { 14984, "scoreCounter" },
    { 14985, "activateNuke" },
    { 14986, "setUseBarScore" },
    { 14987, "updateHudElems" },
    { 14988, "AAMissileLaunchVert" },
    { 14989, "AAMissileLaunchHorz" },
    { 14990, "AAMissileLaunchTargetDist" },
    { 14991, "tryUseAAMissile" },
    { 14992, "getTargets" },
    { 14993, "aa_missile_fire" },
    { 14994, "AAMmissileLaunchTargetDist" }, // this is a misspelling in original script
    { 14995, "teamAirDenied" },
    { 14996, "tryUseAAStrike" },
    { 14997, "cycleTargets" },
    { 14998, "findTargets" },
    { 14999, "earlyAbortWatcher" },
    { 15000, "airDeniedPlayer" },
    { 15001, "finishAAStrike" },
    { 15002, "fireAtTarget" },
    { 15003, "AASoundManager" },
    { 15006, "tweakfile" },
    { 15007, "damagetype" },
    { 15008, "audio_settings" },
    { 15107, "meleeingPlayer" },
    { 17433, "attachmentMap" },
    { 17434, "checkRoundWin" },
    { 25827, "_unk_field_ID25827" }, // was introduced in an IW patch, used in _destructible.gsc
}};

struct __init__
{
    __init__()
    {
        static bool init = false;
        if (init) return;
        init = true;

        opcode_map.reserve(opcode_list.size());
        opcode_map_rev.reserve(opcode_list.size());
        function_map.reserve(function_list.size());
        function_map_rev.reserve(function_list.size());
        method_map.reserve(method_list.size());
        method_map_rev.reserve(method_list.size());
        file_map.reserve(file_list.size());
        file_map_rev.reserve(file_list.size());
        token_map.reserve(token_list.size());
        token_map_rev.reserve(token_list.size());

        for (const auto& entry : opcode_list)
        {
            opcode_map.insert({ entry.first, entry.second });
            opcode_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : function_list)
        {
            function_map.insert({ entry.first, entry.second });
            function_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : method_list)
        {
            method_map.insert({ entry.first, entry.second });
            method_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : file_list)
        {
            file_map.insert({ entry.first, entry.second });
            file_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : token_list)
        {
            token_map.insert({ entry.first, entry.second });
            token_map_rev.insert({ utils::string::to_lower(entry.second), entry.first });
        }
    }
};

__init__ _;

} // namespace xsk::gsc::iw5

#ifdef _MSC_VER
#pragma warning(pop)
#endif

/* unmaped files */

// "character/character_mp_ally_juggernaut"
// "xmodelalias\\alias_us_army_heads"
// "character\\mp_character_us_army_assault_a"
// "character\\mp_character_us_army_assault_b"
// "character\\mp_character_us_army_assault_c"
// "character\\mp_character_us_army_lmg"
// "character\\mp_character_us_army_lmg_b"
// "character\\mp_character_us_army_lmg_c"
// "character\\mp_character_us_army_shotgun"
// "character\\mp_character_us_army_shotgun_b"
// "character\\mp_character_us_army_shotgun_c"
// "xmodelalias\\alias_us_army_riot"
// "character\\mp_character_us_army_riot"
// "character\\mp_character_us_army_smg"
// "character\\mp_character_us_army_smg_b"
// "character\\mp_character_us_army_smg_c"
// "character\\mp_character_us_army_sniper"
// "xmodelalias\\alias_opforce_arctic_heads"
// "character\\mp_character_opforce_arctic_assault_a"
// "character\\mp_character_opforce_arctic_assault_b"
// "character\\mp_character_opforce_arctic_assault_c"
// "character\\mp_character_opforce_arctic_lmg"
// "character\\mp_character_opforce_arctic_lmg_b"
// "character\\mp_character_opforce_arctic_lmg_c"
// "character\\mp_character_opforce_arctic_shotgun"
// "character\\mp_character_opforce_arctic_shotgun_b"
// "character\\mp_character_opforce_arctic_shotgun_c"
// "character\\mp_character_opforce_arctic_smg"
// "character\\mp_character_opforce_arctic_smg_b"
// "character\\mp_character_opforce_arctic_smg_c"
// "character\\mp_character_op_arctic_sniper"
// "character\\mp_character_op_arctic_riot"
// "xmodelalias\\alias_seal_udt_heads"
// "character\\mp_character_seal_udt_assault_a"
// "character\\mp_character_seal_udt_lmg"
// "character\\mp_character_seal_udt_assault_b"
// "character\\mp_character_seal_udt_smg"
// "character\\mp_character_seal_udt_sniper"
// "character\\mp_character_udt_riot"
// "xmodelalias\\alias_opforce_arab_heads"
// "character\\mp_character_composite_assault_a"
// "character\\mp_character_composite_lmg"
// "character\\mp_character_composite_shotgun"
// "character\\mp_character_composite_smg"
// "character\\mp_character_op_arab_sniper"
// "character\\mp_character_op_arab_riot"
// "xmodelalias\\alias_airborne_heads"
// "character\\mp_character_airborne_assault_a"
// "character\\mp_character_airborne_assault_b"
// "character\\mp_character_airborne_assault_c"
// "character\\mp_character_airborne_lmg"
// "character\\mp_character_airborne_lmg_b"
// "character\\mp_character_airborne_lmg_c"
// "character\\mp_character_airborne_shotgun"
// "character\\mp_character_airborne_shotgun_b"
// "character\\mp_character_airborne_shotgun_c"
// "character\\mp_character_airborne_smg"
// "character\\mp_character_airborne_smg_b"
// "character\\mp_character_airborne_smg_c"
// "character\\mp_character_op_airborne_sniper"
// "character\\mp_character_op_airborne_riot"
// "xmodelalias\\alias_opforce_militia_heads_blk"
// "xmodelalias\\alias_opforce_militia_heads_wht"
// "character\\mp_character_militia_assault_aa_blk"
// "character\\mp_character_militia_assault_aa_wht"
// "character\\mp_character_militia_assault_ab_blk"
// "character\\mp_character_militia_assault_ac_blk"
// "character\\mp_character_militia_lmg_aa_blk"
// "character\\mp_character_militia_lmg_ab_blk"
// "character\\mp_character_militia_lmg_ac_blk"
// "character\\mp_character_militia_shotgun_aa_blk"
// "character\\mp_character_militia_shotgun_aa_wht"
// "character\\mp_character_militia_smg_aa_blk"
// "character\\mp_character_militia_smg_aa_wht"
// "character\\mp_character_militia_smg_ab_blk"
// "character\\mp_character_militia_smg_ac_blk"
// "character\\mp_character_op_militia_sniper"
// "character\\mp_character_op_militia_riot"
// "xmodelalias\\alias_tf141_heads_arctic"
// "character\\mp_character_tf_141_arctic_assault_a"
// "character\\mp_character_tf_141_arctic_assault_b"
// "character\\mp_character_tf_141_arctic_lmg"
// "character\\mp_character_tf_141_arctic_shotgun"
// "character\\mp_character_tf_141_arctic_smg"
// "character\\mp_character_tf141_arctic_sniper"
// "character\\mp_character_tf141_arctic_riot"
// "xmodelalias\\alias_tf141_heads_desert"
// "character\\mp_character_tf_141_desert_assault_a"
// "character\\mp_character_tf_141_desert_assault_b"
// "character\\mp_character_tf_141_desert_lmg"
// "character\\mp_character_tf_141_desert_shotgun"
// "character\\mp_character_tf_141_desert_smg"
// "character\\mp_character_tf141_desert_sniper"
// "character\\mp_character_tf141_desert_riot"
// "xmodelalias\\alias_tf141_heads_forest"
// "character\\mp_character_tf_141_forest_assault_a"
// "character\\mp_character_tf_141_forest_assault_b"
// "character\\mp_character_tf_141_forest_lmg"
// "character\\mp_character_tf_141_forest_shotgun"
// "character\\mp_character_tf_141_forest_smg"
// "character\\mp_character_tf141_forest_sniper"
// "character\\mp_character_tf141_forest_riot"
// "xmodelalias\\alias_delta_heads"
// "xmodelalias\\alias_delta_heads_longsleeves"
// "character\\mp_character_delta_urban_assault_a"
// "character\\mp_character_delta_urban_assault_b"
// "character\\mp_character_delta_urban_lmg_a"
// "character\\mp_character_delta_urban_lmg_b"
// "character\\mp_character_delta_urban_shotgun"
// "character\\mp_character_delta_urban_smg_a"
// "character\\mp_character_delta_urban_smg_b"
// "character\\mp_character_delta_urban_smg_c"
// "character\\mp_character_delta_urban_riot"
// "mptype\\mptype_us_army_assault"
// "mptype\\mptype_us_army_lmg"
// "mptype\\mptype_us_army_shotgun"
// "mptype\\mptype_us_army_riot"
// "mptype\\mptype_us_army_smg"
// "mptype\\mptype_us_army_sniper"
// "mptype\\mptype_opforce_arctic_assault"
// "mptype\\mptype_opforce_arctic_lmg"
// "mptype\\mptype_opforce_arctic_shotgun"
// "mptype\\mptype_opforce_arctic_smg"
// "mptype\\mptype_opforce_arctic_sniper"
// "mptype\\mptype_opforce_arctic_riot"
// "mptype\\mptype_seal_udt_assault"
// "mptype\\mptype_seal_udt_lmg"
// "mptype\\mptype_seal_udt_shotgun"
// "mptype\\mptype_seal_udt_smg"
// "mptype\\mptype_seal_udt_sniper"
// "mptype\\mptype_seal_udt_riot"
// "mptype\\mptype_opforce_comp_assault"
// "mptype\\mptype_opforce_comp_lmg"
// "mptype\\mptype_opforce_comp_shotgun"
// "mptype\\mptype_opforce_comp_smg"
// "mptype\\mptype_opforce_comp_sniper"
// "mptype\\mptype_opforce_comp_riot"
// "character\\mp_character_composite_sniper"
// "mptype\\mptype_opforce_airborne_assault"
// "mptype\\mptype_opforce_airborne_lmg"
// "mptype\\mptype_opforce_airborne_shotgun"
// "mptype\\mptype_opforce_airborne_smg"
// "mptype\\mptype_opforce_airborne_sniper"
// "mptype\\mptype_opforce_airborne_riot"
// "mptype\\mptype_opforce_militia_assault"
// "mptype\\mptype_opforce_militia_lmg"
// "mptype\\mptype_opforce_militia_shotgun"
// "mptype\\mptype_opforce_militia_smg"
// "mptype\\mptype_opforce_militia_sniper"
// "mptype\\mptype_opforce_militia_riot"
// "mptype\\mptype_tf141_arctic_assault"
// "mptype\\mptype_tf141_arctic_lmg"
// "mptype\\mptype_tf141_arctic_shotgun"
// "mptype\\mptype_tf141_arctic_smg"
// "mptype\\mptype_tf141_arctic_sniper"
// "mptype\\mptype_tf141_arctic_riot"
// "mptype\\mptype_tf141_desert_assault"
// "mptype\\mptype_tf141_desert_lmg"
// "mptype\\mptype_tf141_desert_shotgun"
// "mptype\\mptype_tf141_desert_smg"
// "mptype\\mptype_tf141_desert_sniper"
// "mptype\\mptype_tf141_desert_riot"
// "mptype\\mptype_tf141_forest_assault"
// "mptype\\mptype_tf141_forest_lmg"
// "mptype\\mptype_tf141_forest_shotgun"
// "mptype\\mptype_tf141_forest_smg"
// "mptype\\mptype_tf141_forest_sniper"
// "mptype\\mptype_tf141_forest_riot"
// "mptype\\mptype_delta_ucp_assault"
// "mptype\\mptype_delta_ucp_lmg"
// "mptype\\mptype_delta_ucp_shotgun"
// "mptype\\mptype_delta_ucp_smg"
// "mptype\\mptype_delta_ucp_sniper"
// "mptype\\mptype_delta_ucp_riot"
// "character\\mp_character_delta_urban_assault_c"
// "vehicle_scripts\\_empty"
// "maps\\_breach"
// "maps\\_briefing"
// "maps\\_cagedchickens"
// "maps\\_carry_ai"
// "maps\\_deadbody"
// "maps\\_drone_ai"
// "maps\\_drone_civilian"
// "maps\\_flare"
// "maps\\_flashbang"
// "maps\\_float"
// "maps\\_heli_ride"
// "maps\\_hiding_door_anims"
// "maps\\_hud_weapons"
// "maps\\_inventory"
// "maps\\_leak"
// "maps\\_menus"
// "maps\\_shellshock"
// "maps\\_mortar"
// "maps\\_overheat"
// "maps\\_props"
// "maps\\_radiation"
// "maps\\_sea"
// "maps\\_stealth_utility"
// "maps\\_stealth_shared_utilities"
// "maps\\_stealth_animation_funcs"
// "maps\\_stealth_threat_enemy"
// "maps\\_stealth_visibility_enemy"
// "maps\\_stealth_visibility_friendly"
// "maps\\_stealth_behavior_friendly"
// "maps\\_stealth_behavior_enemy"
// "maps\\_stealth_corpse_enemy"
// "maps\\_stealth_corpse_system"
// "maps\\_stealth_event_enemy"
// "maps\\_stealth_color_friendly"
// "maps\\_stealth_accuracy_friendly"
// "maps\\_stealth_smartstance_friendly"
// "maps\\_stealth_visibility_system"
// "maps\\_stealth_vehicle"
// "maps\\_treeburst"
// "maps\\_vehicle_missile"
// "maps\\_vehicle_spline"
// "maps\\_vehicledrive"
// "maps\\_weather"
// "maps\\_wood"
// "maps\\_audio_presets_music"
// "maps\\_shg_common"
// "xmodelalias\\alias_so_martyrdom_smg_bodies"
// "character\\character_so_martyrdom"
// "xmodelalias\\alias_africa_militia_heads_a"
// "xmodelalias\\alias_africa_militia_hats_a"
// "xmodelalias\\alias_africa_militia_hats_b"
// "xmodelalias\\alias_africa_militia_hats_c"
// "character\\character_africa_militia_smg_a"
// "character\\character_africa_militia_smg_b"
// "character\\character_africa_militia_smg_c"
// "xmodelalias\\alias_so_regular_smg_heads"
// "character\\character_so_regular_smg"
// "maps\\_rambo"
// "xmodelalias\\alias_so_veteran_ar_heads"
// "character\\character_so_veteran_ar"
// "character\\character_sp_juggernaut"
// "maps\\_juggernaut"
// "character\\character_so_juggernaut_mid"
// "maps\\_riotshield"
// "character\\character_so_juggernaut_explosive"
// "character\\character_so_juggernaut_headshot"
// "xmodelalias\\alias_chemwar_russian_heads_masked"
// "character\\character_chemwar_russian_assault_a"
// "character\\character_chemwar_russian_assault_m_b"
// "character\\character_chemwar_russian_assault_m_c"
// "character\\character_chemwar_russian_assault_m_d"
// "character\\character_chemwar_russian_assault_m_e"
// "character\\character_chemwar_russian_assault_aa"
// "character\\character_chemwar_russian_assault_m_bb"
// "character\\character_chemwar_russian_assault_m_cc"
// "character\\character_chemwar_russian_assault_m_dd"
// "character\\character_chemwar_russian_assault_m_ee"
// "xmodelalias\\alias_so_russian_naval_bodies"
// "xmodelalias\\alias_russian_naval_heads"
// "character\\character_so_russian_naval_assault"
// "xmodelalias\\alias_chemwar_russian_heads_so"
// "character\\character_chemwar_m_d_so"
// "character\\character_chemwar_m_dd_so"
// "xmodelalias\\alias_so_hardened_ar_heads"
// "character\\character_so_hardened_ar"
// "character\\character_gign_paris_smg"
// "character\\character_gign_paris_assault"
// "character\\character_delta_elite_assault_aa"
// "character\\character_delta_elite_assault_ab"
// "character\\character_delta_elite_assault_ba"
// "character\\character_delta_elite_assault_bb"
// "character\\character_sp_german_sheperd_dog"
// "maps\\_utility_joec"
// "maps\\_squad_enemies"
// "maps\\_sp_killstreaks"
// "maps\\_remotemissile_utility"
// "maps\\_remotemissile"
// "maps\\_sp_airdrop"
// "maps\\_so_survival_perks"
// "common_scripts\\_sentry"
// "maps\\_sp_airstrike"
// "maps\\_so_survival_code"
// "maps\\_chopperboss"
// "maps\\_so_survival_loot"
// "maps\\_so_survival_ai"
// "vehicle_scripts\\_mi17_noai"
// "vehicle_scripts\\_littlebird"
// "vehicle_scripts\\_ucav"
// "vehicle_scripts\\_mi17"
// "vehicle_scripts\\_blackhawk"
// "maps\\_air_support_strobe"
// "maps\\_so_survival"
// "maps\\_so_survival_armory"
// "maps\\_so_survival_challenge"
// "maps\\_so_survival_dialog"
// "vehicle_scripts\\_littlebird_player"
// "maps\\_shg_fx"
// "maps\\createart\\ny_manhattan_fog"
// "maps\\createart\\ny_manhattan_art"
// SCR_OPAQUE_FILE3_maps_cinematic_setups_nym_hind_finale = 0x295,
// SCR_OPAQUE_FILE3_maps_cinematic_setups_nym_sewer_exit = 0x296,
// "maps\\_c4"
// "xmodelalias\\alias_russian_military_manhattan_heads"
// "character\\character_opforce_manhattan_assault_a"
// "character\\character_russian_military_rpg_a"
// "character\\character_opforce_manhattan_shgn_a"
// "character\\character_opforce_manhattan_lmg_a"
// "xmodelalias\\alias_rangers_heads"
// "character\\character_rangers_bdu_assault_a"
// "xmodelalias\\alias_air_crew_heads"
// "character\\character_air_crew_manhattan"
// "character\\character_opforce_manhattan_rpg_a"
// "character\\character_delta_elite_smg_a"
// "character\\character_delta_elite_smg_b"
// "maps\\_minigun_viewmodel"
// "maps\\_minigun"
// "maps\\createfx\\ny_manhattan_fx"
// "vehicle_scripts\\_gaz_dshk"
// "vehicle_scripts\\_hind"
// "vehicle_scripts\\_ss_n_12"
// "maps\\animated_models\\fence_tarp_108x76_med_01"
// "maps\\animated_models\\fence_tarp_132x82_med_01"
// "maps\\_predator2"
// "maps\\_xm25"
// "vehicle_scripts\\_blackhawk_minigun"
// "vehicle_scripts\\_stryker50cal"
// "character\\character_opforce_rushwood_assault_a"
// "maps\\createart\\so_survival_mp_dome_art"
// "character\\character_so_juggernaut_lite"
// "xmodelalias\\alias_russian_naval_bodies"
// "vehicle_scripts\\_submarine_sdv"
// "vehicle_scripts\\_super_dvora"
// "vehicle_scripts\\_zodiac"
// "maps\\_ocean"
// "vehicle_scripts\\_zubr"
// "character\\character_sp_opforce_e"
// "character\\character_sp_opforce_f"
// "character\\character_sp_opforce_derik"
// "character\\character_opforce_rushwood_ass_dust_a"
// "character\\character_opforce_rushwood_smg_dust_a"
// "maps\\createart\\berlin_fog"
// "xmodelalias\\alias_russian_military_gasmask_heads"
// "character\\character_opforce_paris_gasmask"
// "maps\\createart\\so_jeep_paris_b_fog"
// "maps\\createart\\so_jeep_paris_b_art"
// "vehicle_scripts\\_mig29"
// "vehicle_scripts\\_t72"
// "vehicle_scripts\\_btr80"
// "maps\\createfx\\prague_fx"
// "maps\\_stealth_anims"
// "maps\\_idle_lean_smoke"
// "maps\\_idle_sleep"
// "maps\\_idle_smoke_balcony"
// "vehicle_scripts\\_technical_aa"
// "vehicle_scripts\\_technical_payback"
// "maps\\animated_models\\highrise_fencetarp_03"
// "maps\\createart\\so_littlebird_payback_fog"
// "vehicle_scripts\\_uk_delivery_truck"
// "vehicle_scripts\\_uk_utility_truck"
// "maps\\createart\\london_art"
// "maps\\_nightvision"
// "maps\\createart\\so_timetrial_london_art"
// "character\\character_fso_vest_nopacks_alt"
// "maps\\createart\\hijack_art"
// "character\\character_africa_militia_assault_a"
// "character\\character_africa_militia_assault_b"
// "character\\character_africa_militia_rpg_a"
// "maps\\_stinger"
// "vehicle_scripts\\_cobra"
// "character\\character_opforce_rescue_assault_b"
// "character\\character_hero_delta_sandman"
// "character\\character_delta_elite_snow_assault_aa"
// "vehicle_scripts\\_bm21"
// "xmodelalias\\alias_henchmen_heads"
// "maps\\_javelin"
// "common_scripts\\_destructible_types_anim_prop_radar_maz"
// "character\\character_hero_africa_price"
// "maps\\_slowmo_breach_payback"
// "maps\\_underwater_debris"
// "character\\character_hero_delta_sandman_udt"
// "vehicle_scripts\\_russian_torpedo"
// "maps\\_credits"
// "maps\\createart\\paris_a_art"
//
